/*
**
** File: fm.c -- software implementation of Yamaha FM sound generator
**
** Copyright (C) 2001, 2002, 2003 Jarek Burczynski (bujar at mame dot net)
** Copyright (C) 1998 Tatsuyuki Satoh , MultiArcadeMachineEmulator development
**
** Version 1.4 (final beta)
**
*/

#include "app.h"
//#include "fm.h"
//#include "fmint.h"
//#include "fmtables.h"

//#include "fmtables.h"

/* sustain level table (3dB per step) */
/* bit0, bit1, bit2, bit3, bit4, bit5, bit6 */
/* 1,    2,    4,    8,    16,   32,   64   (value)*/
/* 0.75, 1.5,  3,    6,    12,   24,   48   (dB)*/

/* 0 - 15: 0, 3, 6, 9,12,15,18,21,24,27,30,33,36,39,42,93 (dB)*/
#define SC(db) (UINT32) ( db * (4.0/ENV_STEP) )
const UINT32 sl_table[16] ={
 SC( 0),SC( 1),SC( 2),SC(3 ),SC(4 ),SC(5 ),SC(6 ),SC( 7),
 SC( 8),SC( 9),SC(10),SC(11),SC(12),SC(13),SC(14),SC(31)
};
#undef SC

const UINT8 eg_inc[19*RATE_STEPS] ={

/*cycle:0 1  2 3  4 5  6 7*/

/* 0 */ 0,1, 0,1, 0,1, 0,1, /* rates 00..11 0 (increment by 0 or 1) */
/* 1 */ 0,1, 0,1, 1,1, 0,1, /* rates 00..11 1 */
/* 2 */ 0,1, 1,1, 0,1, 1,1, /* rates 00..11 2 */
/* 3 */ 0,1, 1,1, 1,1, 1,1, /* rates 00..11 3 */

/* 4 */ 1,1, 1,1, 1,1, 1,1, /* rate 12 0 (increment by 1) */
/* 5 */ 1,1, 1,2, 1,1, 1,2, /* rate 12 1 */
/* 6 */ 1,2, 1,2, 1,2, 1,2, /* rate 12 2 */
/* 7 */ 1,2, 2,2, 1,2, 2,2, /* rate 12 3 */

/* 8 */ 2,2, 2,2, 2,2, 2,2, /* rate 13 0 (increment by 2) */
/* 9 */ 2,2, 2,4, 2,2, 2,4, /* rate 13 1 */
/*10 */ 2,4, 2,4, 2,4, 2,4, /* rate 13 2 */
/*11 */ 2,4, 4,4, 2,4, 4,4, /* rate 13 3 */

/*12 */ 4,4, 4,4, 4,4, 4,4, /* rate 14 0 (increment by 4) */
/*13 */ 4,4, 4,8, 4,4, 4,8, /* rate 14 1 */
/*14 */ 4,8, 4,8, 4,8, 4,8, /* rate 14 2 */
/*15 */ 4,8, 8,8, 4,8, 8,8, /* rate 14 3 */

/*16 */ 8,8, 8,8, 8,8, 8,8, /* rates 15 0, 15 1, 15 2, 15 3 (increment by 8) */
/*17 */ 16,16,16,16,16,16,16,16, /* rates 15 2, 15 3 for attack */
/*18 */ 0,0, 0,0, 0,0, 0,0, /* infinity rates for attack and decay(s) */
};


#define O(a) (a*RATE_STEPS)

/*note that there is no O(17) in this table - it's directly in the code */
const UINT8 eg_rate_select[32+64+32] ={	/* Envelope Generator rates (32 + 64 rates + 32 RKS) */
/* 32 infinite time rates */
O(18),O(18),O(18),O(18),O(18),O(18),O(18),O(18),
O(18),O(18),O(18),O(18),O(18),O(18),O(18),O(18),
O(18),O(18),O(18),O(18),O(18),O(18),O(18),O(18),
O(18),O(18),O(18),O(18),O(18),O(18),O(18),O(18),

/* rates 00-11 */
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),
O( 0),O( 1),O( 2),O( 3),

/* rate 12 */
O( 4),O( 5),O( 6),O( 7),

/* rate 13 */
O( 8),O( 9),O(10),O(11),

/* rate 14 */
O(12),O(13),O(14),O(15),

/* rate 15 */
O(16),O(16),O(16),O(16),

/* 32 dummy rates (same as 15 3) */
O(16),O(16),O(16),O(16),O(16),O(16),O(16),O(16),
O(16),O(16),O(16),O(16),O(16),O(16),O(16),O(16),
O(16),O(16),O(16),O(16),O(16),O(16),O(16),O(16),
O(16),O(16),O(16),O(16),O(16),O(16),O(16),O(16)

};
#undef O

/*rate  0,    1,    2,   3,   4,   5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15*/
/*shift 11,   10,   9,   8,   7,   6,  5,  4,  3,  2, 1,  0,  0,  0,  0,  0 */
/*mask  2047, 1023, 511, 255, 127, 63, 31, 15, 7,  3, 1,  0,  0,  0,  0,  0 */

#define O(a) (a*1)
const UINT8 eg_rate_shift[32+64+32] ={	/* Envelope Generator counter shifts (32 + 64 rates + 32 RKS) */
/* 32 infinite time rates */
O(0),O(0),O(0),O(0),O(0),O(0),O(0),O(0),
O(0),O(0),O(0),O(0),O(0),O(0),O(0),O(0),
O(0),O(0),O(0),O(0),O(0),O(0),O(0),O(0),
O(0),O(0),O(0),O(0),O(0),O(0),O(0),O(0),

/* rates 00-11 */
O(11),O(11),O(11),O(11),
O(10),O(10),O(10),O(10),
O( 9),O( 9),O( 9),O( 9),
O( 8),O( 8),O( 8),O( 8),
O( 7),O( 7),O( 7),O( 7),
O( 6),O( 6),O( 6),O( 6),
O( 5),O( 5),O( 5),O( 5),
O( 4),O( 4),O( 4),O( 4),
O( 3),O( 3),O( 3),O( 3),
O( 2),O( 2),O( 2),O( 2),
O( 1),O( 1),O( 1),O( 1),
O( 0),O( 0),O( 0),O( 0),

/* rate 12 */
O( 0),O( 0),O( 0),O( 0),

/* rate 13 */
O( 0),O( 0),O( 0),O( 0),

/* rate 14 */
O( 0),O( 0),O( 0),O( 0),

/* rate 15 */
O( 0),O( 0),O( 0),O( 0),

/* 32 dummy rates (same as 15 3) */
O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),
O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),
O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),
O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0),O( 0)

};
#undef O

const UINT8 dt_tab[4 * 32] ={
/* this is YM2151 and YM2612 phase increment data (in 10.10 fixed point format)*/
/* FD=0 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* FD=1 */
	0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2,
	2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 7, 8, 8, 8, 8,
/* FD=2 */
	1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5,
	5, 6, 6, 7, 8, 8, 9,10,11,12,13,14,16,16,16,16,
/* FD=3 */
	2, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 7,
	8 , 8, 9,10,11,12,13,14,16,17,19,20,22,22,22,22
};


/* OPN key frequency number -> key code follow table */
/* fnum higher 4bit -> keycode lower 2bit */
const UINT8 opn_fktable[16]  = {0,0,0,0,0,0,0,1,2,3,3,3,3,3,3,3};


/* 8 LFO speed parameters */
/* each value represents number of samples that one LFO level will last for */
const UINT32 lfo_samples_per_step[8]  = {108, 77, 71, 67, 62, 44, 8, 5};



/*There are 4 different LFO AM depths available, they are:
  0 dB,	1.4 dB,	5.9 dB, 11.8 dB
  Here is how it is generated (in EG steps):

  11.8 dB =	0, 2, 4, 6, 8, 10,12,14,16...126,126,124,122,120,118,....4,2,0
   5.9 dB =	0, 1, 2, 3, 4, 5, 6, 7, 8....63, 63, 62, 61, 60, 59,.....2,1,0
   1.4 dB =	0, 0, 0, 0, 1, 1, 1, 1, 2,...15, 15, 15, 15, 14, 14,.....0,0,0

  (1.4 dB is loosing precision as you can see)

  It's implemented as generator from 0..126 with step 2 then a shift
  right N times, where N is:
	8 for 0 dB
	3 for 1.4 dB
	1 for 5.9 dB
	0 for 11.8 dB
*/
const UINT8 lfo_ams_depth_shift[4]  = {8, 3, 1, 0};



/*There are 8 different LFO PM depths available, they are:
  0, 3.4, 6.7, 10, 14, 20, 40, 80 (cents)

  Modulation level at each depth depends on F-NUMBER bits: 4,5,6,7,8,9,10
  (bits 8,9,10 = FNUM MSB from OCT/FNUM register)

  Here we store only first quarter (positive one) of full waveform.
  Full table (lfo_pm_table) containing all 128 waveforms is build
  at run (init) time.

  One value in table below represents 4 (four) basic LFO steps
  (1 PM step = 4 AM steps).

  For example:
   at LFO SPEED=0 (which is 108 samples per basic LFO step)
   one value from "lfo_pm_output" table lasts for 432 consecutive
   samples (4*108=432) and one full LFO waveform cycle lasts for 13824
   samples (32*432=13824; 32 because we store only a quarter of whole
            waveform in the table below)
*/
const UINT8 lfo_pm_output[7*8][8] ={ /* 7 bits meaningful (of F-NUMBER), 8 LFO output levels per one depth (out of 32), 8 LFO depths */
/* FNUM BIT 4: 000 0001xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 2 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 3 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 4 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 5 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 6 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 7 */ {0,   0,   0,   0,   1,   1,   1,   1},

/* FNUM BIT 5: 000 0010xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 2 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 3 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 4 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 5 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 6 */ {0,   0,   0,   0,   1,   1,   1,   1},
/* DEPTH 7 */ {0,   0,   1,   1,   2,   2,   2,   3},

/* FNUM BIT 6: 000 0100xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 2 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 3 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 4 */ {0,   0,   0,   0,   0,   0,   0,   1},
/* DEPTH 5 */ {0,   0,   0,   0,   1,   1,   1,   1},
/* DEPTH 6 */ {0,   0,   1,   1,   2,   2,   2,   3},
/* DEPTH 7 */ {0,   0,   2,   3,   4,   4,   5,   6},

/* FNUM BIT 7: 000 1000xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 2 */ {0,   0,   0,   0,   0,   0,   1,   1},
/* DEPTH 3 */ {0,   0,   0,   0,   1,   1,   1,   1},
/* DEPTH 4 */ {0,   0,   0,   1,   1,   1,   1,   2},
/* DEPTH 5 */ {0,   0,   1,   1,   2,   2,   2,   3},
/* DEPTH 6 */ {0,   0,   2,   3,   4,   4,   5,   6},
/* DEPTH 7 */ {0,   0,   4,   6,   8,   8, 0xa, 0xc},

/* FNUM BIT 8: 001 0000xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   1,   1,   1,   1},
/* DEPTH 2 */ {0,   0,   0,   1,   1,   1,   2,   2},
/* DEPTH 3 */ {0,   0,   1,   1,   2,   2,   3,   3},
/* DEPTH 4 */ {0,   0,   1,   2,   2,   2,   3,   4},
/* DEPTH 5 */ {0,   0,   2,   3,   4,   4,   5,   6},
/* DEPTH 6 */ {0,   0,   4,   6,   8,   8, 0xa, 0xc},
/* DEPTH 7 */ {0,   0,   8, 0xc,0x10,0x10,0x14,0x18},

/* FNUM BIT 9: 010 0000xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   2,   2,   2,   2},
/* DEPTH 2 */ {0,   0,   0,   2,   2,   2,   4,   4},
/* DEPTH 3 */ {0,   0,   2,   2,   4,   4,   6,   6},
/* DEPTH 4 */ {0,   0,   2,   4,   4,   4,   6,   8},
/* DEPTH 5 */ {0,   0,   4,   6,   8,   8, 0xa, 0xc},
/* DEPTH 6 */ {0,   0,   8, 0xc,0x10,0x10,0x14,0x18},
/* DEPTH 7 */ {0,   0,0x10,0x18,0x20,0x20,0x28,0x30},

/* FNUM BIT10: 100 0000xxxx */
/* DEPTH 0 */ {0,   0,   0,   0,   0,   0,   0,   0},
/* DEPTH 1 */ {0,   0,   0,   0,   4,   4,   4,   4},
/* DEPTH 2 */ {0,   0,   0,   4,   4,   4,   8,   8},
/* DEPTH 3 */ {0,   0,   4,   4,   8,   8, 0xc, 0xc},
/* DEPTH 4 */ {0,   0,   4,   8,   8,   8, 0xc,0x10},
/* DEPTH 5 */ {0,   0,   8, 0xc,0x10,0x10,0x14,0x18},
/* DEPTH 6 */ {0,   0,0x10,0x18,0x20,0x20,0x28,0x30},
/* DEPTH 7 */ {0,   0,0x20,0x30,0x40,0x40,0x50,0x60},

};

const signed short tl_tab [256]  = {
	0x1FE8, 0x1FD4, 0x1FBC, 0x1FA8, 0x1F90, 0x1F7C, 0x1F68, 0x1F50,
	0x1F3C, 0x1F24, 0x1F10, 0x1EFC, 0x1EE4, 0x1ED0, 0x1EB8, 0x1EA4,
	0x1E90, 0x1E7C, 0x1E64, 0x1E50, 0x1E3C, 0x1E28, 0x1E10, 0x1DFC,
	0x1DE8, 0x1DD4, 0x1DC0, 0x1DA8, 0x1D94, 0x1D80, 0x1D6C, 0x1D58,
	0x1D44, 0x1D30, 0x1D1C, 0x1D08, 0x1CF4, 0x1CE0, 0x1CCC, 0x1CB8,
	0x1CA4, 0x1C90, 0x1C7C, 0x1C68, 0x1C54, 0x1C40, 0x1C2C, 0x1C18,
	0x1C08, 0x1BF4, 0x1BE0, 0x1BCC, 0x1BB8, 0x1BA4, 0x1B94, 0x1B80,
	0x1B6C, 0x1B58, 0x1B48, 0x1B34, 0x1B20, 0x1B10, 0x1AFC, 0x1AE8,
	0x1AD4, 0x1AC4, 0x1AB0, 0x1AA0, 0x1A8C, 0x1A78, 0x1A68, 0x1A54,
	0x1A44, 0x1A30, 0x1A20, 0x1A0C, 0x19FC, 0x19E8, 0x19D8, 0x19C4,
	0x19B4, 0x19A0, 0x1990, 0x197C, 0x196C, 0x195C, 0x1948, 0x1938,
	0x1924, 0x1914, 0x1904, 0x18F0, 0x18E0, 0x18D0, 0x18C0, 0x18AC,
	0x189C, 0x188C, 0x1878, 0x1868, 0x1858, 0x1848, 0x1838, 0x1824,
	0x1814, 0x1804, 0x17F4, 0x17E4, 0x17D4, 0x17C0, 0x17B0, 0x17A0,
	0x1790, 0x1780, 0x1770, 0x1760, 0x1750, 0x1740, 0x1730, 0x1720,
	0x1710, 0x1700, 0x16F0, 0x16E0, 0x16D0, 0x16C0, 0x16B0, 0x16A0,
	0x1690, 0x1680, 0x1670, 0x1664, 0x1654, 0x1644, 0x1634, 0x1624,
	0x1614, 0x1604, 0x15F8, 0x15E8, 0x15D8, 0x15C8, 0x15BC, 0x15AC,
	0x159C, 0x158C, 0x1580, 0x1570, 0x1560, 0x1550, 0x1544, 0x1534,
	0x1524, 0x1518, 0x1508, 0x14F8, 0x14EC, 0x14DC, 0x14D0, 0x14C0,
	0x14B0, 0x14A4, 0x1494, 0x1488, 0x1478, 0x146C, 0x145C, 0x1450,
	0x1440, 0x1430, 0x1424, 0x1418, 0x1408, 0x13FC, 0x13EC, 0x13E0,
	0x13D0, 0x13C4, 0x13B4, 0x13A8, 0x139C, 0x138C, 0x1380, 0x1370,
	0x1364, 0x1358, 0x1348, 0x133C, 0x1330, 0x1320, 0x1314, 0x1308,
	0x12F8, 0x12EC, 0x12E0, 0x12D4, 0x12C4, 0x12B8, 0x12AC, 0x12A0,
	0x1290, 0x1284, 0x1278, 0x126C, 0x1260, 0x1250, 0x1244, 0x1238,
	0x122C, 0x1220, 0x1214, 0x1208, 0x11F8, 0x11EC, 0x11E0, 0x11D4,
	0x11C8, 0x11BC, 0x11B0, 0x11A4, 0x1198, 0x118C, 0x1180, 0x1174,
	0x1168, 0x115C, 0x1150, 0x1144, 0x1138, 0x112C, 0x1120, 0x1114,
	0x1108, 0x10FC, 0x10F0, 0x10E4, 0x10D8, 0x10CC, 0x10C0, 0x10B4,
	0x10A8, 0x10A0, 0x1094, 0x1088, 0x107C, 0x1070, 0x1064, 0x1058,
	0x1050, 0x1044, 0x1038, 0x102C, 0x1020, 0x1018, 0x100C, 0x1000
};

const signed char lfo_pm_table [32768]  = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02, 0x03, 0x03, 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x03, 0x03, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x03, 0x03, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x03, 0x03, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x03, 0x03, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x03, 0x04, 0x04, 0x03, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x04, 0x04, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFC, 0xFC, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x04, 0x04, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFC, 0xFC, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x04, 0x04, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFC, 0xFC, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x04, 0x04, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFC, 0xFC, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFC, 0xFC, 0xFB, 0xFA, 0xFA, 0xFB, 0xFC, 0xFC, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x05, 0x05, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFB, 0xFB, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x05, 0x05, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFB, 0xFB, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x05, 0x05, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFB, 0xFB, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFE, 0xFE, 0xFE, 0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03, 0x03, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFD, 0xFE, 0xFE, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03, 0x05, 0x05, 0x03, 0x02, 0x02, 0x02, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFE, 0xFE, 0xFE, 0xFD, 0xFB, 0xFB, 0xFD, 0xFE, 0xFE, 0xFE, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x03, 0x05, 0x05, 0x06, 0x07, 0x07, 0x06, 0x05, 0x05, 0x03, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFD, 0xFB, 0xFB, 0xFA, 0xF9, 0xF9, 0xFA, 0xFB, 0xFB, 0xFD, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x06, 0x06, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFA, 0xFA, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x06, 0x06, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFA, 0xFA, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x06, 0x06, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFA, 0xFA, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x06, 0x06, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xFA, 0xFA, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x06, 0x06, 0x07, 0x09, 0x09, 0x07, 0x06, 0x06, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xFA, 0xFA, 0xF9, 0xF7, 0xF7, 0xF9, 0xFA, 0xFA, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x07, 0x07, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xF9, 0xF9, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x07, 0x07, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xF9, 0xF9, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x07, 0x07, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xF9, 0xF9, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x03, 0x03, 0x03, 0x03, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFD, 0xFD, 0xFD, 0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x01, 0x03, 0x03, 0x04, 0x04, 0x04, 0x04, 0x03, 0x03, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFF, 0xFD, 0xFD, 0xFC, 0xFC, 0xFC, 0xFC, 0xFD, 0xFD, 0xFF, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x03, 0x03, 0x03, 0x04, 0x07, 0x07, 0x04, 0x03, 0x03, 0x03, 0x01, 0x00, 0x00,
	0x00, 0x00, 0xFF, 0xFD, 0xFD, 0xFD, 0xFC, 0xF9, 0xF9, 0xFC, 0xFD, 0xFD, 0xFD, 0xFF, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x04, 0x07, 0x07, 0x08, 0x0A, 0x0A, 0x08, 0x07, 0x07, 0x04, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFC, 0xF9, 0xF9, 0xF8, 0xF6, 0xF6, 0xF8, 0xF9, 0xF9, 0xFC, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x08, 0x08, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF8, 0xF8, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x20, 0x20, 0x28, 0x30, 0x30, 0x28, 0x20, 0x20, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xE0, 0xE0, 0xD8, 0xD0, 0xD0, 0xD8, 0xE0, 0xE0, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x08, 0x08, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF8, 0xF8, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x21, 0x21, 0x29, 0x31, 0x31, 0x29, 0x21, 0x21, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xDF, 0xDF, 0xD7, 0xCF, 0xCF, 0xD7, 0xDF, 0xDF, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x08, 0x08, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF8, 0xF8, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x22, 0x22, 0x2A, 0x33, 0x33, 0x2A, 0x22, 0x22, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDE, 0xDE, 0xD6, 0xCD, 0xCD, 0xD6, 0xDE, 0xDE, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x08, 0x08, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF8, 0xF8, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x08, 0x08, 0x0A, 0x0C, 0x0C, 0x0A, 0x08, 0x08, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF8, 0xF8, 0xF6, 0xF4, 0xF4, 0xF6, 0xF8, 0xF8, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x23, 0x23, 0x2B, 0x34, 0x34, 0x2B, 0x23, 0x23, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDD, 0xDD, 0xD5, 0xCC, 0xCC, 0xD5, 0xDD, 0xDD, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x09, 0x09, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF7, 0xF7, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x24, 0x24, 0x2D, 0x36, 0x36, 0x2D, 0x24, 0x24, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDC, 0xDC, 0xD3, 0xCA, 0xCA, 0xD3, 0xDC, 0xDC, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x09, 0x09, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF7, 0xF7, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x25, 0x25, 0x2E, 0x37, 0x37, 0x2E, 0x25, 0x25, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDB, 0xDB, 0xD2, 0xC9, 0xC9, 0xD2, 0xDB, 0xDB, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x09, 0x09, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF7, 0xF7, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x26, 0x26, 0x2F, 0x39, 0x39, 0x2F, 0x26, 0x26, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xDA, 0xDA, 0xD1, 0xC7, 0xC7, 0xD1, 0xDA, 0xDA, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x04, 0x04, 0x04, 0x04, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFC, 0xFC, 0xFC, 0xFC, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x04, 0x04, 0x06, 0x06, 0x06, 0x06, 0x04, 0x04, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFC, 0xFC, 0xFA, 0xFA, 0xFA, 0xFA, 0xFC, 0xFC, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x04, 0x04, 0x04, 0x06, 0x09, 0x09, 0x06, 0x04, 0x04, 0x04, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFC, 0xFC, 0xFC, 0xFA, 0xF7, 0xF7, 0xFA, 0xFC, 0xFC, 0xFC, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x06, 0x09, 0x09, 0x0B, 0x0D, 0x0D, 0x0B, 0x09, 0x09, 0x06, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFA, 0xF7, 0xF7, 0xF5, 0xF3, 0xF3, 0xF5, 0xF7, 0xF7, 0xFA, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x27, 0x27, 0x30, 0x3A, 0x3A, 0x30, 0x27, 0x27, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xD9, 0xD9, 0xD0, 0xC6, 0xC6, 0xD0, 0xD9, 0xD9, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0A, 0x0A, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF6, 0xF6, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x28, 0x28, 0x32, 0x3C, 0x3C, 0x32, 0x28, 0x28, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD8, 0xD8, 0xCE, 0xC4, 0xC4, 0xCE, 0xD8, 0xD8, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0A, 0x0A, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF6, 0xF6, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x29, 0x29, 0x33, 0x3D, 0x3D, 0x33, 0x29, 0x29, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD7, 0xD7, 0xCD, 0xC3, 0xC3, 0xCD, 0xD7, 0xD7, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0A, 0x0A, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF6, 0xF6, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2A, 0x2A, 0x34, 0x3F, 0x3F, 0x34, 0x2A, 0x2A, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD6, 0xD6, 0xCC, 0xC1, 0xC1, 0xCC, 0xD6, 0xD6, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0A, 0x0A, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF6, 0xF6, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0A, 0x0A, 0x0C, 0x0F, 0x0F, 0x0C, 0x0A, 0x0A, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF6, 0xF6, 0xF4, 0xF1, 0xF1, 0xF4, 0xF6, 0xF6, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2B, 0x2B, 0x35, 0x40, 0x40, 0x35, 0x2B, 0x2B, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD5, 0xD5, 0xCB, 0xC0, 0xC0, 0xCB, 0xD5, 0xD5, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0B, 0x0B, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF5, 0xF5, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2C, 0x2C, 0x37, 0x42, 0x42, 0x37, 0x2C, 0x2C, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD4, 0xD4, 0xC9, 0xBE, 0xBE, 0xC9, 0xD4, 0xD4, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0B, 0x0B, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF5, 0xF5, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2D, 0x2D, 0x38, 0x43, 0x43, 0x38, 0x2D, 0x2D, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD3, 0xD3, 0xC8, 0xBD, 0xBD, 0xC8, 0xD3, 0xD3, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0B, 0x0B, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF5, 0xF5, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2E, 0x2E, 0x39, 0x45, 0x45, 0x39, 0x2E, 0x2E, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD2, 0xD2, 0xC7, 0xBB, 0xBB, 0xC7, 0xD2, 0xD2, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x05, 0x05, 0x05, 0x05, 0x02, 0x02, 0x02, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFE, 0xFE, 0xFE, 0xFB, 0xFB, 0xFB, 0xFB, 0xFE, 0xFE, 0xFE, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x02, 0x05, 0x05, 0x07, 0x07, 0x07, 0x07, 0x05, 0x05, 0x02, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFE, 0xFB, 0xFB, 0xF9, 0xF9, 0xF9, 0xF9, 0xFB, 0xFB, 0xFE, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x05, 0x05, 0x05, 0x07, 0x0B, 0x0B, 0x07, 0x05, 0x05, 0x05, 0x02, 0x00, 0x00,
	0x00, 0x00, 0xFE, 0xFB, 0xFB, 0xFB, 0xF9, 0xF5, 0xF5, 0xF9, 0xFB, 0xFB, 0xFB, 0xFE, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x07, 0x0B, 0x0B, 0x0D, 0x10, 0x10, 0x0D, 0x0B, 0x0B, 0x07, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF9, 0xF5, 0xF5, 0xF3, 0xF0, 0xF0, 0xF3, 0xF5, 0xF5, 0xF9, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2F, 0x2F, 0x3A, 0x46, 0x46, 0x3A, 0x2F, 0x2F, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD1, 0xD1, 0xC6, 0xBA, 0xBA, 0xC6, 0xD1, 0xD1, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0C, 0x0C, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF4, 0xF4, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x30, 0x30, 0x3C, 0x48, 0x48, 0x3C, 0x30, 0x30, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xD0, 0xD0, 0xC4, 0xB8, 0xB8, 0xC4, 0xD0, 0xD0, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0C, 0x0C, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF4, 0xF4, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x31, 0x31, 0x3D, 0x49, 0x49, 0x3D, 0x31, 0x31, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xCF, 0xCF, 0xC3, 0xB7, 0xB7, 0xC3, 0xCF, 0xCF, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0C, 0x0C, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF4, 0xF4, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x32, 0x32, 0x3E, 0x4B, 0x4B, 0x3E, 0x32, 0x32, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCE, 0xCE, 0xC2, 0xB5, 0xB5, 0xC2, 0xCE, 0xCE, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0C, 0x0C, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF4, 0xF4, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0C, 0x0C, 0x0F, 0x12, 0x12, 0x0F, 0x0C, 0x0C, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF4, 0xF4, 0xF1, 0xEE, 0xEE, 0xF1, 0xF4, 0xF4, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x33, 0x33, 0x3F, 0x4C, 0x4C, 0x3F, 0x33, 0x33, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCD, 0xCD, 0xC1, 0xB4, 0xB4, 0xC1, 0xCD, 0xCD, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0D, 0x0D, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF3, 0xF3, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x34, 0x34, 0x41, 0x4E, 0x4E, 0x41, 0x34, 0x34, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCC, 0xCC, 0xBF, 0xB2, 0xB2, 0xBF, 0xCC, 0xCC, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0D, 0x0D, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF3, 0xF3, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x35, 0x35, 0x42, 0x4F, 0x4F, 0x42, 0x35, 0x35, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCB, 0xCB, 0xBE, 0xB1, 0xB1, 0xBE, 0xCB, 0xCB, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0D, 0x0D, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF3, 0xF3, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x36, 0x36, 0x43, 0x51, 0x51, 0x43, 0x36, 0x36, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xCA, 0xCA, 0xBD, 0xAF, 0xAF, 0xBD, 0xCA, 0xCA, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x06, 0x06, 0x06, 0x06, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFA, 0xFA, 0xFA, 0xFA, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x06, 0x06, 0x09, 0x09, 0x09, 0x09, 0x06, 0x06, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xFA, 0xFA, 0xF7, 0xF7, 0xF7, 0xF7, 0xFA, 0xFA, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x06, 0x06, 0x06, 0x09, 0x0D, 0x0D, 0x09, 0x06, 0x06, 0x06, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFA, 0xFA, 0xFA, 0xF7, 0xF3, 0xF3, 0xF7, 0xFA, 0xFA, 0xFA, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x09, 0x0D, 0x0D, 0x10, 0x13, 0x13, 0x10, 0x0D, 0x0D, 0x09, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF7, 0xF3, 0xF3, 0xF0, 0xED, 0xED, 0xF0, 0xF3, 0xF3, 0xF7, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x37, 0x37, 0x44, 0x52, 0x52, 0x44, 0x37, 0x37, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xC9, 0xC9, 0xBC, 0xAE, 0xAE, 0xBC, 0xC9, 0xC9, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0E, 0x0E, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF2, 0xF2, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x38, 0x38, 0x46, 0x54, 0x54, 0x46, 0x38, 0x38, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC8, 0xC8, 0xBA, 0xAC, 0xAC, 0xBA, 0xC8, 0xC8, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0E, 0x0E, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF2, 0xF2, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x39, 0x39, 0x47, 0x55, 0x55, 0x47, 0x39, 0x39, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC7, 0xC7, 0xB9, 0xAB, 0xAB, 0xB9, 0xC7, 0xC7, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0E, 0x0E, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF2, 0xF2, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3A, 0x3A, 0x48, 0x57, 0x57, 0x48, 0x3A, 0x3A, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC6, 0xC6, 0xB8, 0xA9, 0xA9, 0xB8, 0xC6, 0xC6, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0E, 0x0E, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF2, 0xF2, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0E, 0x0E, 0x11, 0x15, 0x15, 0x11, 0x0E, 0x0E, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF2, 0xF2, 0xEF, 0xEB, 0xEB, 0xEF, 0xF2, 0xF2, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3B, 0x3B, 0x49, 0x58, 0x58, 0x49, 0x3B, 0x3B, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC5, 0xC5, 0xB7, 0xA8, 0xA8, 0xB7, 0xC5, 0xC5, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0F, 0x0F, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF1, 0xF1, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3C, 0x3C, 0x4B, 0x5A, 0x5A, 0x4B, 0x3C, 0x3C, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC4, 0xC4, 0xB5, 0xA6, 0xA6, 0xB5, 0xC4, 0xC4, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0F, 0x0F, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF1, 0xF1, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3D, 0x3D, 0x4C, 0x5B, 0x5B, 0x4C, 0x3D, 0x3D, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC3, 0xC3, 0xB4, 0xA5, 0xA5, 0xB4, 0xC3, 0xC3, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0F, 0x0F, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF1, 0xF1, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3E, 0x3E, 0x4D, 0x5D, 0x5D, 0x4D, 0x3E, 0x3E, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC2, 0xC2, 0xB3, 0xA3, 0xA3, 0xB3, 0xC2, 0xC2, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x03, 0x03, 0x03, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFD, 0xFD, 0xFD, 0xF9, 0xF9, 0xF9, 0xF9, 0xFD, 0xFD, 0xFD, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x03, 0x07, 0x07, 0x0A, 0x0A, 0x0A, 0x0A, 0x07, 0x07, 0x03, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xFD, 0xF9, 0xF9, 0xF6, 0xF6, 0xF6, 0xF6, 0xF9, 0xF9, 0xFD, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x03, 0x07, 0x07, 0x07, 0x0A, 0x0F, 0x0F, 0x0A, 0x07, 0x07, 0x07, 0x03, 0x00, 0x00,
	0x00, 0x00, 0xFD, 0xF9, 0xF9, 0xF9, 0xF6, 0xF1, 0xF1, 0xF6, 0xF9, 0xF9, 0xF9, 0xFD, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0A, 0x0F, 0x0F, 0x12, 0x16, 0x16, 0x12, 0x0F, 0x0F, 0x0A, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF6, 0xF1, 0xF1, 0xEE, 0xEA, 0xEA, 0xEE, 0xF1, 0xF1, 0xF6, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3F, 0x3F, 0x4E, 0x5E, 0x5E, 0x4E, 0x3F, 0x3F, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC1, 0xC1, 0xB2, 0xA2, 0xA2, 0xB2, 0xC1, 0xC1, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x10, 0x10, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xF0, 0xF0, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x20, 0x20, 0x28, 0x30, 0x30, 0x28, 0x20, 0x20, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xE0, 0xE0, 0xD8, 0xD0, 0xD0, 0xD8, 0xE0, 0xE0, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x20, 0x30, 0x40, 0x40, 0x50, 0x60, 0x60, 0x50, 0x40, 0x40, 0x30, 0x20, 0x00, 0x00,
	0x00, 0x00, 0xE0, 0xD0, 0xC0, 0xC0, 0xB0, 0xA0, 0xA0, 0xB0, 0xC0, 0xC0, 0xD0, 0xE0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x10, 0x10, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xF0, 0xF0, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x20, 0x20, 0x28, 0x30, 0x30, 0x28, 0x20, 0x20, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xE0, 0xE0, 0xD8, 0xD0, 0xD0, 0xD8, 0xE0, 0xE0, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x20, 0x30, 0x41, 0x41, 0x51, 0x61, 0x61, 0x51, 0x41, 0x41, 0x30, 0x20, 0x00, 0x00,
	0x00, 0x00, 0xE0, 0xD0, 0xBF, 0xBF, 0xAF, 0x9F, 0x9F, 0xAF, 0xBF, 0xBF, 0xD0, 0xE0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x10, 0x10, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xF0, 0xF0, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x21, 0x21, 0x29, 0x31, 0x31, 0x29, 0x21, 0x21, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xDF, 0xDF, 0xD7, 0xCF, 0xCF, 0xD7, 0xDF, 0xDF, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x21, 0x31, 0x42, 0x42, 0x52, 0x63, 0x63, 0x52, 0x42, 0x42, 0x31, 0x21, 0x00, 0x00,
	0x00, 0x00, 0xDF, 0xCF, 0xBE, 0xBE, 0xAE, 0x9D, 0x9D, 0xAE, 0xBE, 0xBE, 0xCF, 0xDF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x10, 0x10, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xF0, 0xF0, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x10, 0x10, 0x14, 0x18, 0x18, 0x14, 0x10, 0x10, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xF0, 0xF0, 0xEC, 0xE8, 0xE8, 0xEC, 0xF0, 0xF0, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x18, 0x21, 0x21, 0x29, 0x31, 0x31, 0x29, 0x21, 0x21, 0x18, 0x10, 0x00, 0x00,
	0x00, 0x00, 0xF0, 0xE8, 0xDF, 0xDF, 0xD7, 0xCF, 0xCF, 0xD7, 0xDF, 0xDF, 0xE8, 0xF0, 0x00, 0x00,
	0x00, 0x00, 0x21, 0x31, 0x43, 0x43, 0x53, 0x64, 0x64, 0x53, 0x43, 0x43, 0x31, 0x21, 0x00, 0x00,
	0x00, 0x00, 0xDF, 0xCF, 0xBD, 0xBD, 0xAD, 0x9C, 0x9C, 0xAD, 0xBD, 0xBD, 0xCF, 0xDF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x11, 0x11, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xEF, 0xEF, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x22, 0x22, 0x2A, 0x33, 0x33, 0x2A, 0x22, 0x22, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDE, 0xDE, 0xD6, 0xCD, 0xCD, 0xD6, 0xDE, 0xDE, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x22, 0x33, 0x44, 0x44, 0x55, 0x66, 0x66, 0x55, 0x44, 0x44, 0x33, 0x22, 0x00, 0x00,
	0x00, 0x00, 0xDE, 0xCD, 0xBC, 0xBC, 0xAB, 0x9A, 0x9A, 0xAB, 0xBC, 0xBC, 0xCD, 0xDE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x11, 0x11, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xEF, 0xEF, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x22, 0x22, 0x2A, 0x33, 0x33, 0x2A, 0x22, 0x22, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDE, 0xDE, 0xD6, 0xCD, 0xCD, 0xD6, 0xDE, 0xDE, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x22, 0x33, 0x45, 0x45, 0x56, 0x67, 0x67, 0x56, 0x45, 0x45, 0x33, 0x22, 0x00, 0x00,
	0x00, 0x00, 0xDE, 0xCD, 0xBB, 0xBB, 0xAA, 0x99, 0x99, 0xAA, 0xBB, 0xBB, 0xCD, 0xDE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x11, 0x11, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xEF, 0xEF, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x23, 0x23, 0x2B, 0x34, 0x34, 0x2B, 0x23, 0x23, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDD, 0xDD, 0xD5, 0xCC, 0xCC, 0xD5, 0xDD, 0xDD, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x23, 0x34, 0x46, 0x46, 0x57, 0x69, 0x69, 0x57, 0x46, 0x46, 0x34, 0x23, 0x00, 0x00,
	0x00, 0x00, 0xDD, 0xCC, 0xBA, 0xBA, 0xA9, 0x97, 0x97, 0xA9, 0xBA, 0xBA, 0xCC, 0xDD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF8, 0xF8, 0xF8, 0xF8, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x08, 0x08, 0x0C, 0x0C, 0x0C, 0x0C, 0x08, 0x08, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF8, 0xF8, 0xF4, 0xF4, 0xF4, 0xF4, 0xF8, 0xF8, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x08, 0x08, 0x08, 0x0C, 0x11, 0x11, 0x0C, 0x08, 0x08, 0x08, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF8, 0xF8, 0xF8, 0xF4, 0xEF, 0xEF, 0xF4, 0xF8, 0xF8, 0xF8, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x08, 0x0C, 0x11, 0x11, 0x15, 0x19, 0x19, 0x15, 0x11, 0x11, 0x0C, 0x08, 0x00, 0x00,
	0x00, 0x00, 0xF8, 0xF4, 0xEF, 0xEF, 0xEB, 0xE7, 0xE7, 0xEB, 0xEF, 0xEF, 0xF4, 0xF8, 0x00, 0x00,
	0x00, 0x00, 0x11, 0x19, 0x23, 0x23, 0x2B, 0x34, 0x34, 0x2B, 0x23, 0x23, 0x19, 0x11, 0x00, 0x00,
	0x00, 0x00, 0xEF, 0xE7, 0xDD, 0xDD, 0xD5, 0xCC, 0xCC, 0xD5, 0xDD, 0xDD, 0xE7, 0xEF, 0x00, 0x00,
	0x00, 0x00, 0x23, 0x34, 0x47, 0x47, 0x58, 0x6A, 0x6A, 0x58, 0x47, 0x47, 0x34, 0x23, 0x00, 0x00,
	0x00, 0x00, 0xDD, 0xCC, 0xB9, 0xB9, 0xA8, 0x96, 0x96, 0xA8, 0xB9, 0xB9, 0xCC, 0xDD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x12, 0x12, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xEE, 0xEE, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x24, 0x24, 0x2D, 0x36, 0x36, 0x2D, 0x24, 0x24, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDC, 0xDC, 0xD3, 0xCA, 0xCA, 0xD3, 0xDC, 0xDC, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x24, 0x36, 0x48, 0x48, 0x5A, 0x6C, 0x6C, 0x5A, 0x48, 0x48, 0x36, 0x24, 0x00, 0x00,
	0x00, 0x00, 0xDC, 0xCA, 0xB8, 0xB8, 0xA6, 0x94, 0x94, 0xA6, 0xB8, 0xB8, 0xCA, 0xDC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x12, 0x12, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xEE, 0xEE, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x24, 0x24, 0x2D, 0x36, 0x36, 0x2D, 0x24, 0x24, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDC, 0xDC, 0xD3, 0xCA, 0xCA, 0xD3, 0xDC, 0xDC, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x24, 0x36, 0x49, 0x49, 0x5B, 0x6D, 0x6D, 0x5B, 0x49, 0x49, 0x36, 0x24, 0x00, 0x00,
	0x00, 0x00, 0xDC, 0xCA, 0xB7, 0xB7, 0xA5, 0x93, 0x93, 0xA5, 0xB7, 0xB7, 0xCA, 0xDC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x12, 0x12, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xEE, 0xEE, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x25, 0x25, 0x2E, 0x37, 0x37, 0x2E, 0x25, 0x25, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDB, 0xDB, 0xD2, 0xC9, 0xC9, 0xD2, 0xDB, 0xDB, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x25, 0x37, 0x4A, 0x4A, 0x5C, 0x6F, 0x6F, 0x5C, 0x4A, 0x4A, 0x37, 0x25, 0x00, 0x00,
	0x00, 0x00, 0xDB, 0xC9, 0xB6, 0xB6, 0xA4, 0x91, 0x91, 0xA4, 0xB6, 0xB6, 0xC9, 0xDB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x12, 0x12, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xEE, 0xEE, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x12, 0x12, 0x16, 0x1B, 0x1B, 0x16, 0x12, 0x12, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xEE, 0xEE, 0xEA, 0xE5, 0xE5, 0xEA, 0xEE, 0xEE, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x12, 0x1B, 0x25, 0x25, 0x2E, 0x37, 0x37, 0x2E, 0x25, 0x25, 0x1B, 0x12, 0x00, 0x00,
	0x00, 0x00, 0xEE, 0xE5, 0xDB, 0xDB, 0xD2, 0xC9, 0xC9, 0xD2, 0xDB, 0xDB, 0xE5, 0xEE, 0x00, 0x00,
	0x00, 0x00, 0x25, 0x37, 0x4B, 0x4B, 0x5D, 0x70, 0x70, 0x5D, 0x4B, 0x4B, 0x37, 0x25, 0x00, 0x00,
	0x00, 0x00, 0xDB, 0xC9, 0xB5, 0xB5, 0xA3, 0x90, 0x90, 0xA3, 0xB5, 0xB5, 0xC9, 0xDB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x13, 0x13, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xED, 0xED, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x26, 0x26, 0x2F, 0x39, 0x39, 0x2F, 0x26, 0x26, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xDA, 0xDA, 0xD1, 0xC7, 0xC7, 0xD1, 0xDA, 0xDA, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x26, 0x39, 0x4C, 0x4C, 0x5F, 0x72, 0x72, 0x5F, 0x4C, 0x4C, 0x39, 0x26, 0x00, 0x00,
	0x00, 0x00, 0xDA, 0xC7, 0xB4, 0xB4, 0xA1, 0x8E, 0x8E, 0xA1, 0xB4, 0xB4, 0xC7, 0xDA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x13, 0x13, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xED, 0xED, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x26, 0x26, 0x2F, 0x39, 0x39, 0x2F, 0x26, 0x26, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xDA, 0xDA, 0xD1, 0xC7, 0xC7, 0xD1, 0xDA, 0xDA, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x26, 0x39, 0x4D, 0x4D, 0x60, 0x73, 0x73, 0x60, 0x4D, 0x4D, 0x39, 0x26, 0x00, 0x00,
	0x00, 0x00, 0xDA, 0xC7, 0xB3, 0xB3, 0xA0, 0x8D, 0x8D, 0xA0, 0xB3, 0xB3, 0xC7, 0xDA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x13, 0x13, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xED, 0xED, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x27, 0x27, 0x30, 0x3A, 0x3A, 0x30, 0x27, 0x27, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xD9, 0xD9, 0xD0, 0xC6, 0xC6, 0xD0, 0xD9, 0xD9, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x27, 0x3A, 0x4E, 0x4E, 0x61, 0x75, 0x75, 0x61, 0x4E, 0x4E, 0x3A, 0x27, 0x00, 0x00,
	0x00, 0x00, 0xD9, 0xC6, 0xB2, 0xB2, 0x9F, 0x8B, 0x8B, 0x9F, 0xB2, 0xB2, 0xC6, 0xD9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x09, 0x09, 0x09, 0x09, 0x04, 0x04, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFC, 0xFC, 0xFC, 0xF7, 0xF7, 0xF7, 0xF7, 0xFC, 0xFC, 0xFC, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x04, 0x09, 0x09, 0x0D, 0x0D, 0x0D, 0x0D, 0x09, 0x09, 0x04, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xFC, 0xF7, 0xF7, 0xF3, 0xF3, 0xF3, 0xF3, 0xF7, 0xF7, 0xFC, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x04, 0x09, 0x09, 0x09, 0x0D, 0x13, 0x13, 0x0D, 0x09, 0x09, 0x09, 0x04, 0x00, 0x00,
	0x00, 0x00, 0xFC, 0xF7, 0xF7, 0xF7, 0xF3, 0xED, 0xED, 0xF3, 0xF7, 0xF7, 0xF7, 0xFC, 0x00, 0x00,
	0x00, 0x00, 0x09, 0x0D, 0x13, 0x13, 0x17, 0x1C, 0x1C, 0x17, 0x13, 0x13, 0x0D, 0x09, 0x00, 0x00,
	0x00, 0x00, 0xF7, 0xF3, 0xED, 0xED, 0xE9, 0xE4, 0xE4, 0xE9, 0xED, 0xED, 0xF3, 0xF7, 0x00, 0x00,
	0x00, 0x00, 0x13, 0x1C, 0x27, 0x27, 0x30, 0x3A, 0x3A, 0x30, 0x27, 0x27, 0x1C, 0x13, 0x00, 0x00,
	0x00, 0x00, 0xED, 0xE4, 0xD9, 0xD9, 0xD0, 0xC6, 0xC6, 0xD0, 0xD9, 0xD9, 0xE4, 0xED, 0x00, 0x00,
	0x00, 0x00, 0x27, 0x3A, 0x4F, 0x4F, 0x62, 0x76, 0x76, 0x62, 0x4F, 0x4F, 0x3A, 0x27, 0x00, 0x00,
	0x00, 0x00, 0xD9, 0xC6, 0xB1, 0xB1, 0x9E, 0x8A, 0x8A, 0x9E, 0xB1, 0xB1, 0xC6, 0xD9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x14, 0x14, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEC, 0xEC, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x28, 0x28, 0x32, 0x3C, 0x3C, 0x32, 0x28, 0x28, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD8, 0xD8, 0xCE, 0xC4, 0xC4, 0xCE, 0xD8, 0xD8, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x28, 0x3C, 0x50, 0x50, 0x64, 0x78, 0x78, 0x64, 0x50, 0x50, 0x3C, 0x28, 0x00, 0x00,
	0x00, 0x00, 0xD8, 0xC4, 0xB0, 0xB0, 0x9C, 0x88, 0x88, 0x9C, 0xB0, 0xB0, 0xC4, 0xD8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x14, 0x14, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEC, 0xEC, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x28, 0x28, 0x32, 0x3C, 0x3C, 0x32, 0x28, 0x28, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD8, 0xD8, 0xCE, 0xC4, 0xC4, 0xCE, 0xD8, 0xD8, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x28, 0x3C, 0x51, 0x51, 0x65, 0x79, 0x79, 0x65, 0x51, 0x51, 0x3C, 0x28, 0x00, 0x00,
	0x00, 0x00, 0xD8, 0xC4, 0xAF, 0xAF, 0x9B, 0x87, 0x87, 0x9B, 0xAF, 0xAF, 0xC4, 0xD8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x14, 0x14, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEC, 0xEC, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x29, 0x29, 0x33, 0x3D, 0x3D, 0x33, 0x29, 0x29, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD7, 0xD7, 0xCD, 0xC3, 0xC3, 0xCD, 0xD7, 0xD7, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x29, 0x3D, 0x52, 0x52, 0x66, 0x7B, 0x7B, 0x66, 0x52, 0x52, 0x3D, 0x29, 0x00, 0x00,
	0x00, 0x00, 0xD7, 0xC3, 0xAE, 0xAE, 0x9A, 0x85, 0x85, 0x9A, 0xAE, 0xAE, 0xC3, 0xD7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x14, 0x14, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEC, 0xEC, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x14, 0x14, 0x19, 0x1E, 0x1E, 0x19, 0x14, 0x14, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEC, 0xEC, 0xE7, 0xE2, 0xE2, 0xE7, 0xEC, 0xEC, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x14, 0x1E, 0x29, 0x29, 0x33, 0x3D, 0x3D, 0x33, 0x29, 0x29, 0x1E, 0x14, 0x00, 0x00,
	0x00, 0x00, 0xEC, 0xE2, 0xD7, 0xD7, 0xCD, 0xC3, 0xC3, 0xCD, 0xD7, 0xD7, 0xE2, 0xEC, 0x00, 0x00,
	0x00, 0x00, 0x29, 0x3D, 0x53, 0x53, 0x67, 0x7C, 0x7C, 0x67, 0x53, 0x53, 0x3D, 0x29, 0x00, 0x00,
	0x00, 0x00, 0xD7, 0xC3, 0xAD, 0xAD, 0x99, 0x84, 0x84, 0x99, 0xAD, 0xAD, 0xC3, 0xD7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x15, 0x15, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEB, 0xEB, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2A, 0x2A, 0x34, 0x3F, 0x3F, 0x34, 0x2A, 0x2A, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD6, 0xD6, 0xCC, 0xC1, 0xC1, 0xCC, 0xD6, 0xD6, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x2A, 0x3F, 0x54, 0x54, 0x69, 0x7E, 0x7E, 0x69, 0x54, 0x54, 0x3F, 0x2A, 0x00, 0x00,
	0x00, 0x00, 0xD6, 0xC1, 0xAC, 0xAC, 0x97, 0x82, 0x82, 0x97, 0xAC, 0xAC, 0xC1, 0xD6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x15, 0x15, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEB, 0xEB, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2A, 0x2A, 0x34, 0x3F, 0x3F, 0x34, 0x2A, 0x2A, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD6, 0xD6, 0xCC, 0xC1, 0xC1, 0xCC, 0xD6, 0xD6, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x2A, 0x3F, 0x55, 0x55, 0x6A, 0x7F, 0x7F, 0x6A, 0x55, 0x55, 0x3F, 0x2A, 0x00, 0x00,
	0x00, 0x00, 0xD6, 0xC1, 0xAB, 0xAB, 0x96, 0x81, 0x81, 0x96, 0xAB, 0xAB, 0xC1, 0xD6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x15, 0x15, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEB, 0xEB, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2B, 0x2B, 0x35, 0x40, 0x40, 0x35, 0x2B, 0x2B, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD5, 0xD5, 0xCB, 0xC0, 0xC0, 0xCB, 0xD5, 0xD5, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x2B, 0x40, 0x56, 0x56, 0x6B, 0x81, 0x81, 0x6B, 0x56, 0x56, 0x40, 0x2B, 0x00, 0x00,
	0x00, 0x00, 0xD5, 0xC0, 0xAA, 0xAA, 0x95, 0x7F, 0x7F, 0x95, 0xAA, 0xAA, 0xC0, 0xD5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0A, 0x0A, 0x0A, 0x0A, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF6, 0xF6, 0xF6, 0xF6, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0A, 0x0A, 0x0F, 0x0F, 0x0F, 0x0F, 0x0A, 0x0A, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF6, 0xF6, 0xF1, 0xF1, 0xF1, 0xF1, 0xF6, 0xF6, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0A, 0x0A, 0x0A, 0x0F, 0x15, 0x15, 0x0F, 0x0A, 0x0A, 0x0A, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF6, 0xF6, 0xF6, 0xF1, 0xEB, 0xEB, 0xF1, 0xF6, 0xF6, 0xF6, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0A, 0x0F, 0x15, 0x15, 0x1A, 0x1F, 0x1F, 0x1A, 0x15, 0x15, 0x0F, 0x0A, 0x00, 0x00,
	0x00, 0x00, 0xF6, 0xF1, 0xEB, 0xEB, 0xE6, 0xE1, 0xE1, 0xE6, 0xEB, 0xEB, 0xF1, 0xF6, 0x00, 0x00,
	0x00, 0x00, 0x15, 0x1F, 0x2B, 0x2B, 0x35, 0x40, 0x40, 0x35, 0x2B, 0x2B, 0x1F, 0x15, 0x00, 0x00,
	0x00, 0x00, 0xEB, 0xE1, 0xD5, 0xD5, 0xCB, 0xC0, 0xC0, 0xCB, 0xD5, 0xD5, 0xE1, 0xEB, 0x00, 0x00,
	0x00, 0x00, 0x2B, 0x40, 0x57, 0x57, 0x6C, 0x82, 0x82, 0x6C, 0x57, 0x57, 0x40, 0x2B, 0x00, 0x00,
	0x00, 0x00, 0xD5, 0xC0, 0xA9, 0xA9, 0x94, 0x7E, 0x7E, 0x94, 0xA9, 0xA9, 0xC0, 0xD5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x16, 0x16, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xEA, 0xEA, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2C, 0x2C, 0x37, 0x42, 0x42, 0x37, 0x2C, 0x2C, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD4, 0xD4, 0xC9, 0xBE, 0xBE, 0xC9, 0xD4, 0xD4, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x2C, 0x42, 0x58, 0x58, 0x6E, 0x84, 0x84, 0x6E, 0x58, 0x58, 0x42, 0x2C, 0x00, 0x00,
	0x00, 0x00, 0xD4, 0xBE, 0xA8, 0xA8, 0x92, 0x7C, 0x7C, 0x92, 0xA8, 0xA8, 0xBE, 0xD4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x16, 0x16, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xEA, 0xEA, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2C, 0x2C, 0x37, 0x42, 0x42, 0x37, 0x2C, 0x2C, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD4, 0xD4, 0xC9, 0xBE, 0xBE, 0xC9, 0xD4, 0xD4, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x2C, 0x42, 0x59, 0x59, 0x6F, 0x85, 0x85, 0x6F, 0x59, 0x59, 0x42, 0x2C, 0x00, 0x00,
	0x00, 0x00, 0xD4, 0xBE, 0xA7, 0xA7, 0x91, 0x7B, 0x7B, 0x91, 0xA7, 0xA7, 0xBE, 0xD4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x16, 0x16, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xEA, 0xEA, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2D, 0x2D, 0x38, 0x43, 0x43, 0x38, 0x2D, 0x2D, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD3, 0xD3, 0xC8, 0xBD, 0xBD, 0xC8, 0xD3, 0xD3, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x2D, 0x43, 0x5A, 0x5A, 0x70, 0x87, 0x87, 0x70, 0x5A, 0x5A, 0x43, 0x2D, 0x00, 0x00,
	0x00, 0x00, 0xD3, 0xBD, 0xA6, 0xA6, 0x90, 0x79, 0x79, 0x90, 0xA6, 0xA6, 0xBD, 0xD3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x16, 0x16, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xEA, 0xEA, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x16, 0x16, 0x1B, 0x21, 0x21, 0x1B, 0x16, 0x16, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xEA, 0xEA, 0xE5, 0xDF, 0xDF, 0xE5, 0xEA, 0xEA, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x16, 0x21, 0x2D, 0x2D, 0x38, 0x43, 0x43, 0x38, 0x2D, 0x2D, 0x21, 0x16, 0x00, 0x00,
	0x00, 0x00, 0xEA, 0xDF, 0xD3, 0xD3, 0xC8, 0xBD, 0xBD, 0xC8, 0xD3, 0xD3, 0xDF, 0xEA, 0x00, 0x00,
	0x00, 0x00, 0x2D, 0x43, 0x5B, 0x5B, 0x71, 0x88, 0x88, 0x71, 0x5B, 0x5B, 0x43, 0x2D, 0x00, 0x00,
	0x00, 0x00, 0xD3, 0xBD, 0xA5, 0xA5, 0x8F, 0x78, 0x78, 0x8F, 0xA5, 0xA5, 0xBD, 0xD3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x17, 0x17, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xE9, 0xE9, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2E, 0x2E, 0x39, 0x45, 0x45, 0x39, 0x2E, 0x2E, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD2, 0xD2, 0xC7, 0xBB, 0xBB, 0xC7, 0xD2, 0xD2, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x2E, 0x45, 0x5C, 0x5C, 0x73, 0x8A, 0x8A, 0x73, 0x5C, 0x5C, 0x45, 0x2E, 0x00, 0x00,
	0x00, 0x00, 0xD2, 0xBB, 0xA4, 0xA4, 0x8D, 0x76, 0x76, 0x8D, 0xA4, 0xA4, 0xBB, 0xD2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x17, 0x17, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xE9, 0xE9, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2E, 0x2E, 0x39, 0x45, 0x45, 0x39, 0x2E, 0x2E, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD2, 0xD2, 0xC7, 0xBB, 0xBB, 0xC7, 0xD2, 0xD2, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x2E, 0x45, 0x5D, 0x5D, 0x74, 0x8B, 0x8B, 0x74, 0x5D, 0x5D, 0x45, 0x2E, 0x00, 0x00,
	0x00, 0x00, 0xD2, 0xBB, 0xA3, 0xA3, 0x8C, 0x75, 0x75, 0x8C, 0xA3, 0xA3, 0xBB, 0xD2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x17, 0x17, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xE9, 0xE9, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2F, 0x2F, 0x3A, 0x46, 0x46, 0x3A, 0x2F, 0x2F, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD1, 0xD1, 0xC6, 0xBA, 0xBA, 0xC6, 0xD1, 0xD1, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x2F, 0x46, 0x5E, 0x5E, 0x75, 0x8D, 0x8D, 0x75, 0x5E, 0x5E, 0x46, 0x2F, 0x00, 0x00,
	0x00, 0x00, 0xD1, 0xBA, 0xA2, 0xA2, 0x8B, 0x73, 0x73, 0x8B, 0xA2, 0xA2, 0xBA, 0xD1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x0B, 0x0B, 0x0B, 0x0B, 0x05, 0x05, 0x05, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFB, 0xFB, 0xFB, 0xF5, 0xF5, 0xF5, 0xF5, 0xFB, 0xFB, 0xFB, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x05, 0x0B, 0x0B, 0x10, 0x10, 0x10, 0x10, 0x0B, 0x0B, 0x05, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xFB, 0xF5, 0xF5, 0xF0, 0xF0, 0xF0, 0xF0, 0xF5, 0xF5, 0xFB, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x05, 0x0B, 0x0B, 0x0B, 0x10, 0x17, 0x17, 0x10, 0x0B, 0x0B, 0x0B, 0x05, 0x00, 0x00,
	0x00, 0x00, 0xFB, 0xF5, 0xF5, 0xF5, 0xF0, 0xE9, 0xE9, 0xF0, 0xF5, 0xF5, 0xF5, 0xFB, 0x00, 0x00,
	0x00, 0x00, 0x0B, 0x10, 0x17, 0x17, 0x1C, 0x22, 0x22, 0x1C, 0x17, 0x17, 0x10, 0x0B, 0x00, 0x00,
	0x00, 0x00, 0xF5, 0xF0, 0xE9, 0xE9, 0xE4, 0xDE, 0xDE, 0xE4, 0xE9, 0xE9, 0xF0, 0xF5, 0x00, 0x00,
	0x00, 0x00, 0x17, 0x22, 0x2F, 0x2F, 0x3A, 0x46, 0x46, 0x3A, 0x2F, 0x2F, 0x22, 0x17, 0x00, 0x00,
	0x00, 0x00, 0xE9, 0xDE, 0xD1, 0xD1, 0xC6, 0xBA, 0xBA, 0xC6, 0xD1, 0xD1, 0xDE, 0xE9, 0x00, 0x00,
	0x00, 0x00, 0x2F, 0x46, 0x5F, 0x5F, 0x76, 0x8E, 0x8E, 0x76, 0x5F, 0x5F, 0x46, 0x2F, 0x00, 0x00,
	0x00, 0x00, 0xD1, 0xBA, 0xA1, 0xA1, 0x8A, 0x72, 0x72, 0x8A, 0xA1, 0xA1, 0xBA, 0xD1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x18, 0x18, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE8, 0xE8, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x30, 0x30, 0x3C, 0x48, 0x48, 0x3C, 0x30, 0x30, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xD0, 0xD0, 0xC4, 0xB8, 0xB8, 0xC4, 0xD0, 0xD0, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x30, 0x48, 0x60, 0x60, 0x78, 0x90, 0x90, 0x78, 0x60, 0x60, 0x48, 0x30, 0x00, 0x00,
	0x00, 0x00, 0xD0, 0xB8, 0xA0, 0xA0, 0x88, 0x70, 0x70, 0x88, 0xA0, 0xA0, 0xB8, 0xD0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x18, 0x18, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE8, 0xE8, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x30, 0x30, 0x3C, 0x48, 0x48, 0x3C, 0x30, 0x30, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xD0, 0xD0, 0xC4, 0xB8, 0xB8, 0xC4, 0xD0, 0xD0, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x30, 0x48, 0x61, 0x61, 0x79, 0x91, 0x91, 0x79, 0x61, 0x61, 0x48, 0x30, 0x00, 0x00,
	0x00, 0x00, 0xD0, 0xB8, 0x9F, 0x9F, 0x87, 0x6F, 0x6F, 0x87, 0x9F, 0x9F, 0xB8, 0xD0, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x18, 0x18, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE8, 0xE8, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x31, 0x31, 0x3D, 0x49, 0x49, 0x3D, 0x31, 0x31, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xCF, 0xCF, 0xC3, 0xB7, 0xB7, 0xC3, 0xCF, 0xCF, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x31, 0x49, 0x62, 0x62, 0x7A, 0x93, 0x93, 0x7A, 0x62, 0x62, 0x49, 0x31, 0x00, 0x00,
	0x00, 0x00, 0xCF, 0xB7, 0x9E, 0x9E, 0x86, 0x6D, 0x6D, 0x86, 0x9E, 0x9E, 0xB7, 0xCF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x18, 0x18, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE8, 0xE8, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x18, 0x18, 0x1E, 0x24, 0x24, 0x1E, 0x18, 0x18, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE8, 0xE8, 0xE2, 0xDC, 0xDC, 0xE2, 0xE8, 0xE8, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x18, 0x24, 0x31, 0x31, 0x3D, 0x49, 0x49, 0x3D, 0x31, 0x31, 0x24, 0x18, 0x00, 0x00,
	0x00, 0x00, 0xE8, 0xDC, 0xCF, 0xCF, 0xC3, 0xB7, 0xB7, 0xC3, 0xCF, 0xCF, 0xDC, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x31, 0x49, 0x63, 0x63, 0x7B, 0x94, 0x94, 0x7B, 0x63, 0x63, 0x49, 0x31, 0x00, 0x00,
	0x00, 0x00, 0xCF, 0xB7, 0x9D, 0x9D, 0x85, 0x6C, 0x6C, 0x85, 0x9D, 0x9D, 0xB7, 0xCF, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x19, 0x19, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE7, 0xE7, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x32, 0x32, 0x3E, 0x4B, 0x4B, 0x3E, 0x32, 0x32, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCE, 0xCE, 0xC2, 0xB5, 0xB5, 0xC2, 0xCE, 0xCE, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x32, 0x4B, 0x64, 0x64, 0x7D, 0x96, 0x96, 0x7D, 0x64, 0x64, 0x4B, 0x32, 0x00, 0x00,
	0x00, 0x00, 0xCE, 0xB5, 0x9C, 0x9C, 0x83, 0x6A, 0x6A, 0x83, 0x9C, 0x9C, 0xB5, 0xCE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x19, 0x19, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE7, 0xE7, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x32, 0x32, 0x3E, 0x4B, 0x4B, 0x3E, 0x32, 0x32, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCE, 0xCE, 0xC2, 0xB5, 0xB5, 0xC2, 0xCE, 0xCE, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x32, 0x4B, 0x65, 0x65, 0x7E, 0x97, 0x97, 0x7E, 0x65, 0x65, 0x4B, 0x32, 0x00, 0x00,
	0x00, 0x00, 0xCE, 0xB5, 0x9B, 0x9B, 0x82, 0x69, 0x69, 0x82, 0x9B, 0x9B, 0xB5, 0xCE, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x19, 0x19, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE7, 0xE7, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x33, 0x33, 0x3F, 0x4C, 0x4C, 0x3F, 0x33, 0x33, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCD, 0xCD, 0xC1, 0xB4, 0xB4, 0xC1, 0xCD, 0xCD, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x33, 0x4C, 0x66, 0x66, 0x7F, 0x99, 0x99, 0x7F, 0x66, 0x66, 0x4C, 0x33, 0x00, 0x00,
	0x00, 0x00, 0xCD, 0xB4, 0x9A, 0x9A, 0x81, 0x67, 0x67, 0x81, 0x9A, 0x9A, 0xB4, 0xCD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0C, 0x0C, 0x0C, 0x0C, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF4, 0xF4, 0xF4, 0xF4, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0C, 0x0C, 0x12, 0x12, 0x12, 0x12, 0x0C, 0x0C, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF4, 0xF4, 0xEE, 0xEE, 0xEE, 0xEE, 0xF4, 0xF4, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0C, 0x0C, 0x0C, 0x12, 0x19, 0x19, 0x12, 0x0C, 0x0C, 0x0C, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF4, 0xF4, 0xF4, 0xEE, 0xE7, 0xE7, 0xEE, 0xF4, 0xF4, 0xF4, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0C, 0x12, 0x19, 0x19, 0x1F, 0x25, 0x25, 0x1F, 0x19, 0x19, 0x12, 0x0C, 0x00, 0x00,
	0x00, 0x00, 0xF4, 0xEE, 0xE7, 0xE7, 0xE1, 0xDB, 0xDB, 0xE1, 0xE7, 0xE7, 0xEE, 0xF4, 0x00, 0x00,
	0x00, 0x00, 0x19, 0x25, 0x33, 0x33, 0x3F, 0x4C, 0x4C, 0x3F, 0x33, 0x33, 0x25, 0x19, 0x00, 0x00,
	0x00, 0x00, 0xE7, 0xDB, 0xCD, 0xCD, 0xC1, 0xB4, 0xB4, 0xC1, 0xCD, 0xCD, 0xDB, 0xE7, 0x00, 0x00,
	0x00, 0x00, 0x33, 0x4C, 0x67, 0x67, 0x80, 0x9A, 0x9A, 0x80, 0x67, 0x67, 0x4C, 0x33, 0x00, 0x00,
	0x00, 0x00, 0xCD, 0xB4, 0x99, 0x99, 0x80, 0x66, 0x66, 0x80, 0x99, 0x99, 0xB4, 0xCD, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1A, 0x1A, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE6, 0xE6, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x34, 0x34, 0x41, 0x4E, 0x4E, 0x41, 0x34, 0x34, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCC, 0xCC, 0xBF, 0xB2, 0xB2, 0xBF, 0xCC, 0xCC, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x34, 0x4E, 0x68, 0x68, 0x82, 0x9C, 0x9C, 0x82, 0x68, 0x68, 0x4E, 0x34, 0x00, 0x00,
	0x00, 0x00, 0xCC, 0xB2, 0x98, 0x98, 0x7E, 0x64, 0x64, 0x7E, 0x98, 0x98, 0xB2, 0xCC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1A, 0x1A, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE6, 0xE6, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x34, 0x34, 0x41, 0x4E, 0x4E, 0x41, 0x34, 0x34, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCC, 0xCC, 0xBF, 0xB2, 0xB2, 0xBF, 0xCC, 0xCC, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x34, 0x4E, 0x69, 0x69, 0x83, 0x9D, 0x9D, 0x83, 0x69, 0x69, 0x4E, 0x34, 0x00, 0x00,
	0x00, 0x00, 0xCC, 0xB2, 0x97, 0x97, 0x7D, 0x63, 0x63, 0x7D, 0x97, 0x97, 0xB2, 0xCC, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1A, 0x1A, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE6, 0xE6, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x35, 0x35, 0x42, 0x4F, 0x4F, 0x42, 0x35, 0x35, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCB, 0xCB, 0xBE, 0xB1, 0xB1, 0xBE, 0xCB, 0xCB, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x35, 0x4F, 0x6A, 0x6A, 0x84, 0x9F, 0x9F, 0x84, 0x6A, 0x6A, 0x4F, 0x35, 0x00, 0x00,
	0x00, 0x00, 0xCB, 0xB1, 0x96, 0x96, 0x7C, 0x61, 0x61, 0x7C, 0x96, 0x96, 0xB1, 0xCB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1A, 0x1A, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE6, 0xE6, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1A, 0x1A, 0x20, 0x27, 0x27, 0x20, 0x1A, 0x1A, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE6, 0xE6, 0xE0, 0xD9, 0xD9, 0xE0, 0xE6, 0xE6, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1A, 0x27, 0x35, 0x35, 0x42, 0x4F, 0x4F, 0x42, 0x35, 0x35, 0x27, 0x1A, 0x00, 0x00,
	0x00, 0x00, 0xE6, 0xD9, 0xCB, 0xCB, 0xBE, 0xB1, 0xB1, 0xBE, 0xCB, 0xCB, 0xD9, 0xE6, 0x00, 0x00,
	0x00, 0x00, 0x35, 0x4F, 0x6B, 0x6B, 0x85, 0xA0, 0xA0, 0x85, 0x6B, 0x6B, 0x4F, 0x35, 0x00, 0x00,
	0x00, 0x00, 0xCB, 0xB1, 0x95, 0x95, 0x7B, 0x60, 0x60, 0x7B, 0x95, 0x95, 0xB1, 0xCB, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1B, 0x1B, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE5, 0xE5, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x36, 0x36, 0x43, 0x51, 0x51, 0x43, 0x36, 0x36, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xCA, 0xCA, 0xBD, 0xAF, 0xAF, 0xBD, 0xCA, 0xCA, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x36, 0x51, 0x6C, 0x6C, 0x87, 0xA2, 0xA2, 0x87, 0x6C, 0x6C, 0x51, 0x36, 0x00, 0x00,
	0x00, 0x00, 0xCA, 0xAF, 0x94, 0x94, 0x79, 0x5E, 0x5E, 0x79, 0x94, 0x94, 0xAF, 0xCA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1B, 0x1B, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE5, 0xE5, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x36, 0x36, 0x43, 0x51, 0x51, 0x43, 0x36, 0x36, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xCA, 0xCA, 0xBD, 0xAF, 0xAF, 0xBD, 0xCA, 0xCA, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x36, 0x51, 0x6D, 0x6D, 0x88, 0xA3, 0xA3, 0x88, 0x6D, 0x6D, 0x51, 0x36, 0x00, 0x00,
	0x00, 0x00, 0xCA, 0xAF, 0x93, 0x93, 0x78, 0x5D, 0x5D, 0x78, 0x93, 0x93, 0xAF, 0xCA, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1B, 0x1B, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE5, 0xE5, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x37, 0x37, 0x44, 0x52, 0x52, 0x44, 0x37, 0x37, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xC9, 0xC9, 0xBC, 0xAE, 0xAE, 0xBC, 0xC9, 0xC9, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x37, 0x52, 0x6E, 0x6E, 0x89, 0xA5, 0xA5, 0x89, 0x6E, 0x6E, 0x52, 0x37, 0x00, 0x00,
	0x00, 0x00, 0xC9, 0xAE, 0x92, 0x92, 0x77, 0x5B, 0x5B, 0x77, 0x92, 0x92, 0xAE, 0xC9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x0D, 0x0D, 0x0D, 0x0D, 0x06, 0x06, 0x06, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xFA, 0xFA, 0xFA, 0xF3, 0xF3, 0xF3, 0xF3, 0xFA, 0xFA, 0xFA, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x06, 0x0D, 0x0D, 0x13, 0x13, 0x13, 0x13, 0x0D, 0x0D, 0x06, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xFA, 0xF3, 0xF3, 0xED, 0xED, 0xED, 0xED, 0xF3, 0xF3, 0xFA, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x0D, 0x0D, 0x0D, 0x13, 0x1B, 0x1B, 0x13, 0x0D, 0x0D, 0x0D, 0x06, 0x00, 0x00,
	0x00, 0x00, 0xFA, 0xF3, 0xF3, 0xF3, 0xED, 0xE5, 0xE5, 0xED, 0xF3, 0xF3, 0xF3, 0xFA, 0x00, 0x00,
	0x00, 0x00, 0x0D, 0x13, 0x1B, 0x1B, 0x21, 0x28, 0x28, 0x21, 0x1B, 0x1B, 0x13, 0x0D, 0x00, 0x00,
	0x00, 0x00, 0xF3, 0xED, 0xE5, 0xE5, 0xDF, 0xD8, 0xD8, 0xDF, 0xE5, 0xE5, 0xED, 0xF3, 0x00, 0x00,
	0x00, 0x00, 0x1B, 0x28, 0x37, 0x37, 0x44, 0x52, 0x52, 0x44, 0x37, 0x37, 0x28, 0x1B, 0x00, 0x00,
	0x00, 0x00, 0xE5, 0xD8, 0xC9, 0xC9, 0xBC, 0xAE, 0xAE, 0xBC, 0xC9, 0xC9, 0xD8, 0xE5, 0x00, 0x00,
	0x00, 0x00, 0x37, 0x52, 0x6F, 0x6F, 0x8A, 0xA6, 0xA6, 0x8A, 0x6F, 0x6F, 0x52, 0x37, 0x00, 0x00,
	0x00, 0x00, 0xC9, 0xAE, 0x91, 0x91, 0x76, 0x5A, 0x5A, 0x76, 0x91, 0x91, 0xAE, 0xC9, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1C, 0x1C, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE4, 0xE4, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x38, 0x38, 0x46, 0x54, 0x54, 0x46, 0x38, 0x38, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC8, 0xC8, 0xBA, 0xAC, 0xAC, 0xBA, 0xC8, 0xC8, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x38, 0x54, 0x70, 0x70, 0x8C, 0xA8, 0xA8, 0x8C, 0x70, 0x70, 0x54, 0x38, 0x00, 0x00,
	0x00, 0x00, 0xC8, 0xAC, 0x90, 0x90, 0x74, 0x58, 0x58, 0x74, 0x90, 0x90, 0xAC, 0xC8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1C, 0x1C, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE4, 0xE4, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x38, 0x38, 0x46, 0x54, 0x54, 0x46, 0x38, 0x38, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC8, 0xC8, 0xBA, 0xAC, 0xAC, 0xBA, 0xC8, 0xC8, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x38, 0x54, 0x71, 0x71, 0x8D, 0xA9, 0xA9, 0x8D, 0x71, 0x71, 0x54, 0x38, 0x00, 0x00,
	0x00, 0x00, 0xC8, 0xAC, 0x8F, 0x8F, 0x73, 0x57, 0x57, 0x73, 0x8F, 0x8F, 0xAC, 0xC8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1C, 0x1C, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE4, 0xE4, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x39, 0x39, 0x47, 0x55, 0x55, 0x47, 0x39, 0x39, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC7, 0xC7, 0xB9, 0xAB, 0xAB, 0xB9, 0xC7, 0xC7, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x39, 0x55, 0x72, 0x72, 0x8E, 0xAB, 0xAB, 0x8E, 0x72, 0x72, 0x55, 0x39, 0x00, 0x00,
	0x00, 0x00, 0xC7, 0xAB, 0x8E, 0x8E, 0x72, 0x55, 0x55, 0x72, 0x8E, 0x8E, 0xAB, 0xC7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1C, 0x1C, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE4, 0xE4, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1C, 0x1C, 0x23, 0x2A, 0x2A, 0x23, 0x1C, 0x1C, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE4, 0xE4, 0xDD, 0xD6, 0xD6, 0xDD, 0xE4, 0xE4, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1C, 0x2A, 0x39, 0x39, 0x47, 0x55, 0x55, 0x47, 0x39, 0x39, 0x2A, 0x1C, 0x00, 0x00,
	0x00, 0x00, 0xE4, 0xD6, 0xC7, 0xC7, 0xB9, 0xAB, 0xAB, 0xB9, 0xC7, 0xC7, 0xD6, 0xE4, 0x00, 0x00,
	0x00, 0x00, 0x39, 0x55, 0x73, 0x73, 0x8F, 0xAC, 0xAC, 0x8F, 0x73, 0x73, 0x55, 0x39, 0x00, 0x00,
	0x00, 0x00, 0xC7, 0xAB, 0x8D, 0x8D, 0x71, 0x54, 0x54, 0x71, 0x8D, 0x8D, 0xAB, 0xC7, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1D, 0x1D, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE3, 0xE3, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3A, 0x3A, 0x48, 0x57, 0x57, 0x48, 0x3A, 0x3A, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC6, 0xC6, 0xB8, 0xA9, 0xA9, 0xB8, 0xC6, 0xC6, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x3A, 0x57, 0x74, 0x74, 0x91, 0xAE, 0xAE, 0x91, 0x74, 0x74, 0x57, 0x3A, 0x00, 0x00,
	0x00, 0x00, 0xC6, 0xA9, 0x8C, 0x8C, 0x6F, 0x52, 0x52, 0x6F, 0x8C, 0x8C, 0xA9, 0xC6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1D, 0x1D, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE3, 0xE3, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3A, 0x3A, 0x48, 0x57, 0x57, 0x48, 0x3A, 0x3A, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC6, 0xC6, 0xB8, 0xA9, 0xA9, 0xB8, 0xC6, 0xC6, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x3A, 0x57, 0x75, 0x75, 0x92, 0xAF, 0xAF, 0x92, 0x75, 0x75, 0x57, 0x3A, 0x00, 0x00,
	0x00, 0x00, 0xC6, 0xA9, 0x8B, 0x8B, 0x6E, 0x51, 0x51, 0x6E, 0x8B, 0x8B, 0xA9, 0xC6, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1D, 0x1D, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE3, 0xE3, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3B, 0x3B, 0x49, 0x58, 0x58, 0x49, 0x3B, 0x3B, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC5, 0xC5, 0xB7, 0xA8, 0xA8, 0xB7, 0xC5, 0xC5, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x3B, 0x58, 0x76, 0x76, 0x93, 0xB1, 0xB1, 0x93, 0x76, 0x76, 0x58, 0x3B, 0x00, 0x00,
	0x00, 0x00, 0xC5, 0xA8, 0x8A, 0x8A, 0x6D, 0x4F, 0x4F, 0x6D, 0x8A, 0x8A, 0xA8, 0xC5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0E, 0x0E, 0x0E, 0x0E, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF2, 0xF2, 0xF2, 0xF2, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0E, 0x0E, 0x15, 0x15, 0x15, 0x15, 0x0E, 0x0E, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF2, 0xF2, 0xEB, 0xEB, 0xEB, 0xEB, 0xF2, 0xF2, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0E, 0x0E, 0x0E, 0x15, 0x1D, 0x1D, 0x15, 0x0E, 0x0E, 0x0E, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF2, 0xF2, 0xF2, 0xEB, 0xE3, 0xE3, 0xEB, 0xF2, 0xF2, 0xF2, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0E, 0x15, 0x1D, 0x1D, 0x24, 0x2B, 0x2B, 0x24, 0x1D, 0x1D, 0x15, 0x0E, 0x00, 0x00,
	0x00, 0x00, 0xF2, 0xEB, 0xE3, 0xE3, 0xDC, 0xD5, 0xD5, 0xDC, 0xE3, 0xE3, 0xEB, 0xF2, 0x00, 0x00,
	0x00, 0x00, 0x1D, 0x2B, 0x3B, 0x3B, 0x49, 0x58, 0x58, 0x49, 0x3B, 0x3B, 0x2B, 0x1D, 0x00, 0x00,
	0x00, 0x00, 0xE3, 0xD5, 0xC5, 0xC5, 0xB7, 0xA8, 0xA8, 0xB7, 0xC5, 0xC5, 0xD5, 0xE3, 0x00, 0x00,
	0x00, 0x00, 0x3B, 0x58, 0x77, 0x77, 0x94, 0xB2, 0xB2, 0x94, 0x77, 0x77, 0x58, 0x3B, 0x00, 0x00,
	0x00, 0x00, 0xC5, 0xA8, 0x89, 0x89, 0x6C, 0x4E, 0x4E, 0x6C, 0x89, 0x89, 0xA8, 0xC5, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1E, 0x1E, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE2, 0xE2, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3C, 0x3C, 0x4B, 0x5A, 0x5A, 0x4B, 0x3C, 0x3C, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC4, 0xC4, 0xB5, 0xA6, 0xA6, 0xB5, 0xC4, 0xC4, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x3C, 0x5A, 0x78, 0x78, 0x96, 0xB4, 0xB4, 0x96, 0x78, 0x78, 0x5A, 0x3C, 0x00, 0x00,
	0x00, 0x00, 0xC4, 0xA6, 0x88, 0x88, 0x6A, 0x4C, 0x4C, 0x6A, 0x88, 0x88, 0xA6, 0xC4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1E, 0x1E, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE2, 0xE2, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3C, 0x3C, 0x4B, 0x5A, 0x5A, 0x4B, 0x3C, 0x3C, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC4, 0xC4, 0xB5, 0xA6, 0xA6, 0xB5, 0xC4, 0xC4, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x3C, 0x5A, 0x79, 0x79, 0x97, 0xB5, 0xB5, 0x97, 0x79, 0x79, 0x5A, 0x3C, 0x00, 0x00,
	0x00, 0x00, 0xC4, 0xA6, 0x87, 0x87, 0x69, 0x4B, 0x4B, 0x69, 0x87, 0x87, 0xA6, 0xC4, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1E, 0x1E, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE2, 0xE2, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3D, 0x3D, 0x4C, 0x5B, 0x5B, 0x4C, 0x3D, 0x3D, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC3, 0xC3, 0xB4, 0xA5, 0xA5, 0xB4, 0xC3, 0xC3, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x3D, 0x5B, 0x7A, 0x7A, 0x98, 0xB7, 0xB7, 0x98, 0x7A, 0x7A, 0x5B, 0x3D, 0x00, 0x00,
	0x00, 0x00, 0xC3, 0xA5, 0x86, 0x86, 0x68, 0x49, 0x49, 0x68, 0x86, 0x86, 0xA5, 0xC3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1E, 0x1E, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE2, 0xE2, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1E, 0x1E, 0x25, 0x2D, 0x2D, 0x25, 0x1E, 0x1E, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE2, 0xE2, 0xDB, 0xD3, 0xD3, 0xDB, 0xE2, 0xE2, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1E, 0x2D, 0x3D, 0x3D, 0x4C, 0x5B, 0x5B, 0x4C, 0x3D, 0x3D, 0x2D, 0x1E, 0x00, 0x00,
	0x00, 0x00, 0xE2, 0xD3, 0xC3, 0xC3, 0xB4, 0xA5, 0xA5, 0xB4, 0xC3, 0xC3, 0xD3, 0xE2, 0x00, 0x00,
	0x00, 0x00, 0x3D, 0x5B, 0x7B, 0x7B, 0x99, 0xB8, 0xB8, 0x99, 0x7B, 0x7B, 0x5B, 0x3D, 0x00, 0x00,
	0x00, 0x00, 0xC3, 0xA5, 0x85, 0x85, 0x67, 0x48, 0x48, 0x67, 0x85, 0x85, 0xA5, 0xC3, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1F, 0x1F, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE1, 0xE1, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3E, 0x3E, 0x4D, 0x5D, 0x5D, 0x4D, 0x3E, 0x3E, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC2, 0xC2, 0xB3, 0xA3, 0xA3, 0xB3, 0xC2, 0xC2, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x3E, 0x5D, 0x7C, 0x7C, 0x9B, 0xBA, 0xBA, 0x9B, 0x7C, 0x7C, 0x5D, 0x3E, 0x00, 0x00,
	0x00, 0x00, 0xC2, 0xA3, 0x84, 0x84, 0x65, 0x46, 0x46, 0x65, 0x84, 0x84, 0xA3, 0xC2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1F, 0x1F, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE1, 0xE1, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3E, 0x3E, 0x4D, 0x5D, 0x5D, 0x4D, 0x3E, 0x3E, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC2, 0xC2, 0xB3, 0xA3, 0xA3, 0xB3, 0xC2, 0xC2, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x3E, 0x5D, 0x7D, 0x7D, 0x9C, 0xBB, 0xBB, 0x9C, 0x7D, 0x7D, 0x5D, 0x3E, 0x00, 0x00,
	0x00, 0x00, 0xC2, 0xA3, 0x83, 0x83, 0x64, 0x45, 0x45, 0x64, 0x83, 0x83, 0xA3, 0xC2, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1F, 0x1F, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE1, 0xE1, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3F, 0x3F, 0x4E, 0x5E, 0x5E, 0x4E, 0x3F, 0x3F, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC1, 0xC1, 0xB2, 0xA2, 0xA2, 0xB2, 0xC1, 0xC1, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x3F, 0x5E, 0x7E, 0x7E, 0x9D, 0xBD, 0xBD, 0x9D, 0x7E, 0x7E, 0x5E, 0x3F, 0x00, 0x00,
	0x00, 0x00, 0xC1, 0xA2, 0x82, 0x82, 0x63, 0x43, 0x43, 0x63, 0x82, 0x82, 0xA2, 0xC1, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x0F, 0x0F, 0x0F, 0x0F, 0x07, 0x07, 0x07, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0xF9, 0xF9, 0xF9, 0xF1, 0xF1, 0xF1, 0xF1, 0xF9, 0xF9, 0xF9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x07, 0x0F, 0x0F, 0x16, 0x16, 0x16, 0x16, 0x0F, 0x0F, 0x07, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF9, 0xF1, 0xF1, 0xEA, 0xEA, 0xEA, 0xEA, 0xF1, 0xF1, 0xF9, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x07, 0x0F, 0x0F, 0x0F, 0x16, 0x1F, 0x1F, 0x16, 0x0F, 0x0F, 0x0F, 0x07, 0x00, 0x00,
	0x00, 0x00, 0xF9, 0xF1, 0xF1, 0xF1, 0xEA, 0xE1, 0xE1, 0xEA, 0xF1, 0xF1, 0xF1, 0xF9, 0x00, 0x00,
	0x00, 0x00, 0x0F, 0x16, 0x1F, 0x1F, 0x26, 0x2E, 0x2E, 0x26, 0x1F, 0x1F, 0x16, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0xF1, 0xEA, 0xE1, 0xE1, 0xDA, 0xD2, 0xD2, 0xDA, 0xE1, 0xE1, 0xEA, 0xF1, 0x00, 0x00,
	0x00, 0x00, 0x1F, 0x2E, 0x3F, 0x3F, 0x4E, 0x5E, 0x5E, 0x4E, 0x3F, 0x3F, 0x2E, 0x1F, 0x00, 0x00,
	0x00, 0x00, 0xE1, 0xD2, 0xC1, 0xC1, 0xB2, 0xA2, 0xA2, 0xB2, 0xC1, 0xC1, 0xD2, 0xE1, 0x00, 0x00,
	0x00, 0x00, 0x3F, 0x5E, 0x7F, 0x7F, 0x9E, 0xBE, 0xBE, 0x9E, 0x7F, 0x7F, 0x5E, 0x3F, 0x00, 0x00,
	0x00, 0x00, 0xC1, 0xA2, 0x81, 0x81, 0x62, 0x42, 0x42, 0x62, 0x81, 0x81, 0xA2, 0xC1, 0x00, 0x00
};

const unsigned short sin_tab [256]  = {
	0x10B2, 0x0D86, 0x0C0E, 0x0B16, 0x0A5C, 0x09C8, 0x094C, 0x08E2, 0x0886, 0x0834, 0x07EA, 0x07A6, 0x076A, 0x0730, 0x06FC, 0x06CA,
	0x069C, 0x0672, 0x0648, 0x0622, 0x05FE, 0x05DA, 0x05B8, 0x059A, 0x057A, 0x055E, 0x0540, 0x0526, 0x050C, 0x04F2, 0x04DA, 0x04C2,
	0x04AC, 0x0496, 0x0480, 0x046C, 0x0458, 0x0444, 0x0430, 0x041E, 0x040C, 0x03FA, 0x03EA, 0x03D8, 0x03C8, 0x03B8, 0x03A8, 0x039A,
	0x038A, 0x037C, 0x036E, 0x0360, 0x0352, 0x0344, 0x0336, 0x032A, 0x031E, 0x0310, 0x0304, 0x02F8, 0x02EE, 0x02E2, 0x02D6, 0x02CC,
	0x02C0, 0x02B6, 0x02AA, 0x02A0, 0x0296, 0x028C, 0x0282, 0x0278, 0x026E, 0x0266, 0x025C, 0x0252, 0x024A, 0x0242, 0x0238, 0x0230,
	0x0228, 0x021E, 0x0216, 0x020E, 0x0206, 0x01FE, 0x01F6, 0x01F0, 0x01E8, 0x01E0, 0x01D8, 0x01D2, 0x01CA, 0x01C4, 0x01BC, 0x01B6,
	0x01AE, 0x01A8, 0x01A2, 0x019A, 0x0194, 0x018E, 0x0188, 0x0182, 0x017C, 0x0176, 0x0170, 0x016A, 0x0164, 0x015E, 0x0158, 0x0152,
	0x014E, 0x0148, 0x0142, 0x013E, 0x0138, 0x0132, 0x012E, 0x0128, 0x0124, 0x011E, 0x011A, 0x0114, 0x0110, 0x010C, 0x0106, 0x0102,
	0x00FE, 0x00FA, 0x00F4, 0x00F0, 0x00EC, 0x00E8, 0x00E4, 0x00E0, 0x00DC, 0x00D8, 0x00D4, 0x00D0, 0x00CC, 0x00C8, 0x00C4, 0x00C0,
	0x00BC, 0x00B8, 0x00B6, 0x00B2, 0x00AE, 0x00AA, 0x00A6, 0x00A4, 0x00A0, 0x009C, 0x009A, 0x0096, 0x0094, 0x0090, 0x008C, 0x008A,
	0x0086, 0x0084, 0x0080, 0x007E, 0x007C, 0x0078, 0x0076, 0x0072, 0x0070, 0x006E, 0x006A, 0x0068, 0x0066, 0x0062, 0x0060, 0x005E,
	0x005C, 0x005A, 0x0056, 0x0054, 0x0052, 0x0050, 0x004E, 0x004C, 0x004A, 0x0048, 0x0046, 0x0044, 0x0042, 0x0040, 0x003E, 0x003C,
	0x003A, 0x0038, 0x0036, 0x0034, 0x0032, 0x0030, 0x002E, 0x002E, 0x002C, 0x002A, 0x0028, 0x0028, 0x0026, 0x0024, 0x0022, 0x0022,
	0x0020, 0x001E, 0x001E, 0x001C, 0x001A, 0x001A, 0x0018, 0x0018, 0x0016, 0x0014, 0x0014, 0x0012, 0x0012, 0x0010, 0x0010, 0x000E,
	0x000E, 0x000E, 0x000C, 0x000C, 0x000A, 0x000A, 0x000A, 0x0008, 0x0008, 0x0008, 0x0006, 0x0006, 0x0006, 0x0004, 0x0004, 0x0004,
	0x0004, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
};

unsigned int OPN_fn_table [4096]  = {
	0x0000000, 0x00019C4, 0x0003389, 0x0004D4E, 0x0006713, 0x00080D7, 0x0009A9C, 0x000B461,
	0x000CE26, 0x000E7EA, 0x00101AF, 0x0011B74, 0x0013539, 0x0014EFE, 0x00168C2, 0x0018287,
	0x0019C4C, 0x001B611, 0x001CFD5, 0x001E99A, 0x002035F, 0x0021D24, 0x00236E9, 0x00250AD,
	0x0026A72, 0x0028437, 0x0029DFC, 0x002B7C0, 0x002D185, 0x002EB4A, 0x003050F, 0x0031ED4,
	0x0033898, 0x003525D, 0x0036C22, 0x00385E7, 0x0039FAB, 0x003B970, 0x003D335, 0x003ECFA,
	0x00406BF, 0x0042083, 0x0043A48, 0x004540D, 0x0046DD2, 0x0048796, 0x004A15B, 0x004BB20,
	0x004D4E5, 0x004EEAA, 0x005086E, 0x0052233, 0x0053BF8, 0x00555BD, 0x0056F81, 0x0058946,
	0x005A30B, 0x005BCD0, 0x005D695, 0x005F059, 0x0060A1E, 0x00623E3, 0x0063DA8, 0x006576C,
	0x0067131, 0x0068AF6, 0x006A4BB, 0x006BE80, 0x006D844, 0x006F209, 0x0070BCE, 0x0072593,
	0x0073F57, 0x007591C, 0x00772E1, 0x0078CA6, 0x007A66A, 0x007C02F, 0x007D9F4, 0x007F3B9,
	0x0080D7E, 0x0082742, 0x0084107, 0x0085ACC, 0x0087491, 0x0088E55, 0x008A81A, 0x008C1DF,
	0x008DBA4, 0x008F569, 0x0090F2D, 0x00928F2, 0x00942B7, 0x0095C7C, 0x0097640, 0x0099005,
	0x009A9CA, 0x009C38F, 0x009DD54, 0x009F718, 0x00A10DD, 0x00A2AA2, 0x00A4467, 0x00A5E2B,
	0x00A77F0, 0x00A91B5, 0x00AAB7A, 0x00AC53F, 0x00ADF03, 0x00AF8C8, 0x00B128D, 0x00B2C52,
	0x00B4616, 0x00B5FDB, 0x00B79A0, 0x00B9365, 0x00BAD2A, 0x00BC6EE, 0x00BE0B3, 0x00BFA78,
	0x00C143D, 0x00C2E01, 0x00C47C6, 0x00C618B, 0x00C7B50, 0x00C9515, 0x00CAED9, 0x00CC89E,
	0x00CE263, 0x00CFC28, 0x00D15EC, 0x00D2FB1, 0x00D4976, 0x00D633B, 0x00D7D00, 0x00D96C4,
	0x00DB089, 0x00DCA4E, 0x00DE413, 0x00DFDD7, 0x00E179C, 0x00E3161, 0x00E4B26, 0x00E64EA,
	0x00E7EAF, 0x00E9874, 0x00EB239, 0x00ECBFE, 0x00EE5C2, 0x00EFF87, 0x00F194C, 0x00F3311,
	0x00F4CD5, 0x00F669A, 0x00F805F, 0x00F9A24, 0x00FB3E9, 0x00FCDAD, 0x00FE772, 0x0100137,
	0x0101AFC, 0x01034C0, 0x0104E85, 0x010684A, 0x010820F, 0x0109BD4, 0x010B598, 0x010CF5D,
	0x010E922, 0x01102E7, 0x0111CAB, 0x0113670, 0x0115035, 0x01169FA, 0x01183BF, 0x0119D83,
	0x011B748, 0x011D10D, 0x011EAD2, 0x0120496, 0x0121E5B, 0x0123820, 0x01251E5, 0x0126BAA,
	0x012856E, 0x0129F33, 0x012B8F8, 0x012D2BD, 0x012EC81, 0x0130646, 0x013200B, 0x01339D0,
	0x0135395, 0x0136D59, 0x013871E, 0x013A0E3, 0x013BAA8, 0x013D46C, 0x013EE31, 0x01407F6,
	0x01421BB, 0x0143B80, 0x0145544, 0x0146F09, 0x01488CE, 0x014A293, 0x014BC57, 0x014D61C,
	0x014EFE1, 0x01509A6, 0x015236A, 0x0153D2F, 0x01556F4, 0x01570B9, 0x0158A7E, 0x015A442,
	0x015BE07, 0x015D7CC, 0x015F191, 0x0160B55, 0x016251A, 0x0163EDF, 0x01658A4, 0x0167269,
	0x0168C2D, 0x016A5F2, 0x016BFB7, 0x016D97C, 0x016F340, 0x0170D05, 0x01726CA, 0x017408F,
	0x0175A54, 0x0177418, 0x0178DDD, 0x017A7A2, 0x017C167, 0x017DB2B, 0x017F4F0, 0x0180EB5,
	0x018287A, 0x018423F, 0x0185C03, 0x01875C8, 0x0188F8D, 0x018A952, 0x018C316, 0x018DCDB,
	0x018F6A0, 0x0191065, 0x0192A2A, 0x01943EE, 0x0195DB3, 0x0197778, 0x019913D, 0x019AB01,
	0x019C4C6, 0x019DE8B, 0x019F850, 0x01A1215, 0x01A2BD9, 0x01A459E, 0x01A5F63, 0x01A7928,
	0x01A92EC, 0x01AACB1, 0x01AC676, 0x01AE03B, 0x01AFA00, 0x01B13C4, 0x01B2D89, 0x01B474E,
	0x01B6113, 0x01B7AD7, 0x01B949C, 0x01BAE61, 0x01BC826, 0x01BE1EA, 0x01BFBAF, 0x01C1574,
	0x01C2F39, 0x01C48FE, 0x01C62C2, 0x01C7C87, 0x01C964C, 0x01CB011, 0x01CC9D5, 0x01CE39A,
	0x01CFD5F, 0x01D1724, 0x01D30E9, 0x01D4AAD, 0x01D6472, 0x01D7E37, 0x01D97FC, 0x01DB1C0,
	0x01DCB85, 0x01DE54A, 0x01DFF0F, 0x01E18D4, 0x01E3298, 0x01E4C5D, 0x01E6622, 0x01E7FE7,
	0x01E99AB, 0x01EB370, 0x01ECD35, 0x01EE6FA, 0x01F00BF, 0x01F1A83, 0x01F3448, 0x01F4E0D,
	0x01F67D2, 0x01F8196, 0x01F9B5B, 0x01FB520, 0x01FCEE5, 0x01FE8AA, 0x020026E, 0x0201C33,
	0x02035F8, 0x0204FBD, 0x0206981, 0x0208346, 0x0209D0B, 0x020B6D0, 0x020D095, 0x020EA59,
	0x021041E, 0x0211DE3, 0x02137A8, 0x021516C, 0x0216B31, 0x02184F6, 0x0219EBB, 0x021B880,
	0x021D244, 0x021EC09, 0x02205CE, 0x0221F93, 0x0223957, 0x022531C, 0x0226CE1, 0x02286A6,
	0x022A06A, 0x022BA2F, 0x022D3F4, 0x022EDB9, 0x023077E, 0x0232142, 0x0233B07, 0x02354CC,
	0x0236E91, 0x0238855, 0x023A21A, 0x023BBDF, 0x023D5A4, 0x023EF69, 0x024092D, 0x02422F2,
	0x0243CB7, 0x024567C, 0x0247040, 0x0248A05, 0x024A3CA, 0x024BD8F, 0x024D754, 0x024F118,
	0x0250ADD, 0x02524A2, 0x0253E67, 0x025582B, 0x02571F0, 0x0258BB5, 0x025A57A, 0x025BF3F,
	0x025D903, 0x025F2C8, 0x0260C8D, 0x0262652, 0x0264016, 0x02659DB, 0x02673A0, 0x0268D65,
	0x026A72A, 0x026C0EE, 0x026DAB3, 0x026F478, 0x0270E3D, 0x0272801, 0x02741C6, 0x0275B8B,
	0x0277550, 0x0278F15, 0x027A8D9, 0x027C29E, 0x027DC63, 0x027F628, 0x0280FEC, 0x02829B1,
	0x0284376, 0x0285D3B, 0x0287700, 0x02890C4, 0x028AA89, 0x028C44E, 0x028DE13, 0x028F7D7,
	0x029119C, 0x0292B61, 0x0294526, 0x0295EEA, 0x02978AF, 0x0299274, 0x029AC39, 0x029C5FE,
	0x029DFC2, 0x029F987, 0x02A134C, 0x02A2D11, 0x02A46D5, 0x02A609A, 0x02A7A5F, 0x02A9424,
	0x02AADE9, 0x02AC7AD, 0x02AE172, 0x02AFB37, 0x02B14FC, 0x02B2EC0, 0x02B4885, 0x02B624A,
	0x02B7C0F, 0x02B95D4, 0x02BAF98, 0x02BC95D, 0x02BE322, 0x02BFCE7, 0x02C16AB, 0x02C3070,
	0x02C4A35, 0x02C63FA, 0x02C7DBF, 0x02C9783, 0x02CB148, 0x02CCB0D, 0x02CE4D2, 0x02CFE96,
	0x02D185B, 0x02D3220, 0x02D4BE5, 0x02D65AA, 0x02D7F6E, 0x02D9933, 0x02DB2F8, 0x02DCCBD,
	0x02DE681, 0x02E0046, 0x02E1A0B, 0x02E33D0, 0x02E4D95, 0x02E6759, 0x02E811E, 0x02E9AE3,
	0x02EB4A8, 0x02ECE6C, 0x02EE831, 0x02F01F6, 0x02F1BBB, 0x02F3580, 0x02F4F44, 0x02F6909,
	0x02F82CE, 0x02F9C93, 0x02FB657, 0x02FD01C, 0x02FE9E1, 0x03003A6, 0x0301D6A, 0x030372F,
	0x03050F4, 0x0306AB9, 0x030847E, 0x0309E42, 0x030B807, 0x030D1CC, 0x030EB91, 0x0310555,
	0x0311F1A, 0x03138DF, 0x03152A4, 0x0316C69, 0x031862D, 0x0319FF2, 0x031B9B7, 0x031D37C,
	0x031ED40, 0x0320705, 0x03220CA, 0x0323A8F, 0x0325454, 0x0326E18, 0x03287DD, 0x032A1A2,
	0x032BB67, 0x032D52B, 0x032EEF0, 0x03308B5, 0x033227A, 0x0333C3F, 0x0335603, 0x0336FC8,
	0x033898D, 0x033A352, 0x033BD16, 0x033D6DB, 0x033F0A0, 0x0340A65, 0x034242A, 0x0343DEE,
	0x03457B3, 0x0347178, 0x0348B3D, 0x034A501, 0x034BEC6, 0x034D88B, 0x034F250, 0x0350C15,
	0x03525D9, 0x0353F9E, 0x0355963, 0x0357328, 0x0358CEC, 0x035A6B1, 0x035C076, 0x035DA3B,
	0x035F400, 0x0360DC4, 0x0362789, 0x036414E, 0x0365B13, 0x03674D7, 0x0368E9C, 0x036A861,
	0x036C226, 0x036DBEA, 0x036F5AF, 0x0370F74, 0x0372939, 0x03742FE, 0x0375CC2, 0x0377687,
	0x037904C, 0x037AA11, 0x037C3D5, 0x037DD9A, 0x037F75F, 0x0381124, 0x0382AE9, 0x03844AD,
	0x0385E72, 0x0387837, 0x03891FC, 0x038ABC0, 0x038C585, 0x038DF4A, 0x038F90F, 0x03912D4,
	0x0392C98, 0x039465D, 0x0396022, 0x03979E7, 0x03993AB, 0x039AD70, 0x039C735, 0x039E0FA,
	0x039FABF, 0x03A1483, 0x03A2E48, 0x03A480D, 0x03A61D2, 0x03A7B96, 0x03A955B, 0x03AAF20,
	0x03AC8E5, 0x03AE2AA, 0x03AFC6E, 0x03B1633, 0x03B2FF8, 0x03B49BD, 0x03B6381, 0x03B7D46,
	0x03B970B, 0x03BB0D0, 0x03BCA95, 0x03BE459, 0x03BFE1E, 0x03C17E3, 0x03C31A8, 0x03C4B6C,
	0x03C6531, 0x03C7EF6, 0x03C98BB, 0x03CB280, 0x03CCC44, 0x03CE609, 0x03CFFCE, 0x03D1993,
	0x03D3357, 0x03D4D1C, 0x03D66E1, 0x03D80A6, 0x03D9A6B, 0x03DB42F, 0x03DCDF4, 0x03DE7B9,
	0x03E017E, 0x03E1B42, 0x03E3507, 0x03E4ECC, 0x03E6891, 0x03E8255, 0x03E9C1A, 0x03EB5DF,
	0x03ECFA4, 0x03EE969, 0x03F032D, 0x03F1CF2, 0x03F36B7, 0x03F507C, 0x03F6A40, 0x03F8405,
	0x03F9DCA, 0x03FB78F, 0x03FD154, 0x03FEB18, 0x04004DD, 0x0401EA2, 0x0403867, 0x040522B,
	0x0406BF0, 0x04085B5, 0x0409F7A, 0x040B93F, 0x040D303, 0x040ECC8, 0x041068D, 0x0412052,
	0x0413A16, 0x04153DB, 0x0416DA0, 0x0418765, 0x041A12A, 0x041BAEE, 0x041D4B3, 0x041EE78,
	0x042083D, 0x0422201, 0x0423BC6, 0x042558B, 0x0426F50, 0x0428915, 0x042A2D9, 0x042BC9E,
	0x042D663, 0x042F028, 0x04309EC, 0x04323B1, 0x0433D76, 0x043573B, 0x0437100, 0x0438AC4,
	0x043A489, 0x043BE4E, 0x043D813, 0x043F1D7, 0x0440B9C, 0x0442561, 0x0443F26, 0x04458EB,
	0x04472AF, 0x0448C74, 0x044A639, 0x044BFFE, 0x044D9C2, 0x044F387, 0x0450D4C, 0x0452711,
	0x04540D5, 0x0455A9A, 0x045745F, 0x0458E24, 0x045A7E9, 0x045C1AD, 0x045DB72, 0x045F537,
	0x0460EFC, 0x04628C0, 0x0464285, 0x0465C4A, 0x046760F, 0x0468FD4, 0x046A998, 0x046C35D,
	0x046DD22, 0x046F6E7, 0x04710AB, 0x0472A70, 0x0474435, 0x0475DFA, 0x04777BF, 0x0479183,
	0x047AB48, 0x047C50D, 0x047DED2, 0x047F896, 0x048125B, 0x0482C20, 0x04845E5, 0x0485FAA,
	0x048796E, 0x0489333, 0x048ACF8, 0x048C6BD, 0x048E081, 0x048FA46, 0x049140B, 0x0492DD0,
	0x0494795, 0x0496159, 0x0497B1E, 0x04994E3, 0x049AEA8, 0x049C86C, 0x049E231, 0x049FBF6,
	0x04A15BB, 0x04A2F80, 0x04A4944, 0x04A6309, 0x04A7CCE, 0x04A9693, 0x04AB057, 0x04ACA1C,
	0x04AE3E1, 0x04AFDA6, 0x04B176B, 0x04B312F, 0x04B4AF4, 0x04B64B9, 0x04B7E7E, 0x04B9842,
	0x04BB207, 0x04BCBCC, 0x04BE591, 0x04BFF55, 0x04C191A, 0x04C32DF, 0x04C4CA4, 0x04C6669,
	0x04C802D, 0x04C99F2, 0x04CB3B7, 0x04CCD7C, 0x04CE740, 0x04D0105, 0x04D1ACA, 0x04D348F,
	0x04D4E54, 0x04D6818, 0x04D81DD, 0x04D9BA2, 0x04DB567, 0x04DCF2B, 0x04DE8F0, 0x04E02B5,
	0x04E1C7A, 0x04E363F, 0x04E5003, 0x04E69C8, 0x04E838D, 0x04E9D52, 0x04EB716, 0x04ED0DB,
	0x04EEAA0, 0x04F0465, 0x04F1E2A, 0x04F37EE, 0x04F51B3, 0x04F6B78, 0x04F853D, 0x04F9F01,
	0x04FB8C6, 0x04FD28B, 0x04FEC50, 0x0500615, 0x0501FD9, 0x050399E, 0x0505363, 0x0506D28,
	0x05086EC, 0x050A0B1, 0x050BA76, 0x050D43B, 0x050EE00, 0x05107C4, 0x0512189, 0x0513B4E,
	0x0515513, 0x0516ED7, 0x051889C, 0x051A261, 0x051BC26, 0x051D5EB, 0x051EFAF, 0x0520974,
	0x0522339, 0x0523CFE, 0x05256C2, 0x0527087, 0x0528A4C, 0x052A411, 0x052BDD5, 0x052D79A,
	0x052F15F, 0x0530B24, 0x05324E9, 0x0533EAD, 0x0535872, 0x0537237, 0x0538BFC, 0x053A5C0,
	0x053BF85, 0x053D94A, 0x053F30F, 0x0540CD4, 0x0542698, 0x054405D, 0x0545A22, 0x05473E7,
	0x0548DAB, 0x054A770, 0x054C135, 0x054DAFA, 0x054F4BF, 0x0550E83, 0x0552848, 0x055420D,
	0x0555BD2, 0x0557596, 0x0558F5B, 0x055A920, 0x055C2E5, 0x055DCAA, 0x055F66E, 0x0561033,
	0x05629F8, 0x05643BD, 0x0565D81, 0x0567746, 0x056910B, 0x056AAD0, 0x056C495, 0x056DE59,
	0x056F81E, 0x05711E3, 0x0572BA8, 0x057456C, 0x0575F31, 0x05778F6, 0x05792BB, 0x057AC80,
	0x057C644, 0x057E009, 0x057F9CE, 0x0581393, 0x0582D57, 0x058471C, 0x05860E1, 0x0587AA6,
	0x058946B, 0x058AE2F, 0x058C7F4, 0x058E1B9, 0x058FB7E, 0x0591542, 0x0592F07, 0x05948CC,
	0x0596291, 0x0597C55, 0x059961A, 0x059AFDF, 0x059C9A4, 0x059E369, 0x059FD2D, 0x05A16F2,
	0x05A30B7, 0x05A4A7C, 0x05A6440, 0x05A7E05, 0x05A97CA, 0x05AB18F, 0x05ACB54, 0x05AE518,
	0x05AFEDD, 0x05B18A2, 0x05B3267, 0x05B4C2B, 0x05B65F0, 0x05B7FB5, 0x05B997A, 0x05BB33F,
	0x05BCD03, 0x05BE6C8, 0x05C008D, 0x05C1A52, 0x05C3416, 0x05C4DDB, 0x05C67A0, 0x05C8165,
	0x05C9B2A, 0x05CB4EE, 0x05CCEB3, 0x05CE878, 0x05D023D, 0x05D1C01, 0x05D35C6, 0x05D4F8B,
	0x05D6950, 0x05D8315, 0x05D9CD9, 0x05DB69E, 0x05DD063, 0x05DEA28, 0x05E03EC, 0x05E1DB1,
	0x05E3776, 0x05E513B, 0x05E6B00, 0x05E84C4, 0x05E9E89, 0x05EB84E, 0x05ED213, 0x05EEBD7,
	0x05F059C, 0x05F1F61, 0x05F3926, 0x05F52EB, 0x05F6CAF, 0x05F8674, 0x05FA039, 0x05FB9FE,
	0x05FD3C2, 0x05FED87, 0x060074C, 0x0602111, 0x0603AD5, 0x060549A, 0x0606E5F, 0x0608824,
	0x060A1E9, 0x060BBAD, 0x060D572, 0x060EF37, 0x06108FC, 0x06122C0, 0x0613C85, 0x061564A,
	0x061700F, 0x06189D4, 0x061A398, 0x061BD5D, 0x061D722, 0x061F0E7, 0x0620AAB, 0x0622470,
	0x0623E35, 0x06257FA, 0x06271BF, 0x0628B83, 0x062A548, 0x062BF0D, 0x062D8D2, 0x062F296,
	0x0630C5B, 0x0632620, 0x0633FE5, 0x06359AA, 0x063736E, 0x0638D33, 0x063A6F8, 0x063C0BD,
	0x063DA81, 0x063F446, 0x0640E0B, 0x06427D0, 0x0644195, 0x0645B59, 0x064751E, 0x0648EE3,
	0x064A8A8, 0x064C26C, 0x064DC31, 0x064F5F6, 0x0650FBB, 0x0652980, 0x0654344, 0x0655D09,
	0x06576CE, 0x0659093, 0x065AA57, 0x065C41C, 0x065DDE1, 0x065F7A6, 0x066116B, 0x0662B2F,
	0x06644F4, 0x0665EB9, 0x066787E, 0x0669242, 0x066AC07, 0x066C5CC, 0x066DF91, 0x066F955,
	0x067131A, 0x0672CDF, 0x06746A4, 0x0676069, 0x0677A2D, 0x06793F2, 0x067ADB7, 0x067C77C,
	0x067E140, 0x067FB05, 0x06814CA, 0x0682E8F, 0x0684854, 0x0686218, 0x0687BDD, 0x06895A2,
	0x068AF67, 0x068C92B, 0x068E2F0, 0x068FCB5, 0x069167A, 0x069303F, 0x0694A03, 0x06963C8,
	0x0697D8D, 0x0699752, 0x069B116, 0x069CADB, 0x069E4A0, 0x069FE65, 0x06A182A, 0x06A31EE,
	0x06A4BB3, 0x06A6578, 0x06A7F3D, 0x06A9901, 0x06AB2C6, 0x06ACC8B, 0x06AE650, 0x06B0015,
	0x06B19D9, 0x06B339E, 0x06B4D63, 0x06B6728, 0x06B80EC, 0x06B9AB1, 0x06BB476, 0x06BCE3B,
	0x06BE800, 0x06C01C4, 0x06C1B89, 0x06C354E, 0x06C4F13, 0x06C68D7, 0x06C829C, 0x06C9C61,
	0x06CB626, 0x06CCFEB, 0x06CE9AF, 0x06D0374, 0x06D1D39, 0x06D36FE, 0x06D50C2, 0x06D6A87,
	0x06D844C, 0x06D9E11, 0x06DB7D5, 0x06DD19A, 0x06DEB5F, 0x06E0524, 0x06E1EE9, 0x06E38AD,
	0x06E5272, 0x06E6C37, 0x06E85FC, 0x06E9FC0, 0x06EB985, 0x06ED34A, 0x06EED0F, 0x06F06D4,
	0x06F2098, 0x06F3A5D, 0x06F5422, 0x06F6DE7, 0x06F87AB, 0x06FA170, 0x06FBB35, 0x06FD4FA,
	0x06FEEBF, 0x0700883, 0x0702248, 0x0703C0D, 0x07055D2, 0x0706F96, 0x070895B, 0x070A320,
	0x070BCE5, 0x070D6AA, 0x070F06E, 0x0710A33, 0x07123F8, 0x0713DBD, 0x0715781, 0x0717146,
	0x0718B0B, 0x071A4D0, 0x071BE95, 0x071D859, 0x071F21E, 0x0720BE3, 0x07225A8, 0x0723F6C,
	0x0725931, 0x07272F6, 0x0728CBB, 0x072A680, 0x072C044, 0x072DA09, 0x072F3CE, 0x0730D93,
	0x0732757, 0x073411C, 0x0735AE1, 0x07374A6, 0x0738E6B, 0x073A82F, 0x073C1F4, 0x073DBB9,
	0x073F57E, 0x0740F42, 0x0742907, 0x07442CC, 0x0745C91, 0x0747655, 0x074901A, 0x074A9DF,
	0x074C3A4, 0x074DD69, 0x074F72D, 0x07510F2, 0x0752AB7, 0x075447C, 0x0755E40, 0x0757805,
	0x07591CA, 0x075AB8F, 0x075C554, 0x075DF18, 0x075F8DD, 0x07612A2, 0x0762C67, 0x076462B,
	0x0765FF0, 0x07679B5, 0x076937A, 0x076AD3F, 0x076C703, 0x076E0C8, 0x076FA8D, 0x0771452,
	0x0772E16, 0x07747DB, 0x07761A0, 0x0777B65, 0x077952A, 0x077AEEE, 0x077C8B3, 0x077E278,
	0x077FC3D, 0x0781601, 0x0782FC6, 0x078498B, 0x0786350, 0x0787D15, 0x07896D9, 0x078B09E,
	0x078CA63, 0x078E428, 0x078FDEC, 0x07917B1, 0x0793176, 0x0794B3B, 0x0796500, 0x0797EC4,
	0x0799889, 0x079B24E, 0x079CC13, 0x079E5D7, 0x079FF9C, 0x07A1961, 0x07A3326, 0x07A4CEB,
	0x07A66AF, 0x07A8074, 0x07A9A39, 0x07AB3FE, 0x07ACDC2, 0x07AE787, 0x07B014C, 0x07B1B11,
	0x07B34D6, 0x07B4E9A, 0x07B685F, 0x07B8224, 0x07B9BE9, 0x07BB5AD, 0x07BCF72, 0x07BE937,
	0x07C02FC, 0x07C1CC0, 0x07C3685, 0x07C504A, 0x07C6A0F, 0x07C83D4, 0x07C9D98, 0x07CB75D,
	0x07CD122, 0x07CEAE7, 0x07D04AB, 0x07D1E70, 0x07D3835, 0x07D51FA, 0x07D6BBF, 0x07D8583,
	0x07D9F48, 0x07DB90D, 0x07DD2D2, 0x07DEC96, 0x07E065B, 0x07E2020, 0x07E39E5, 0x07E53AA,
	0x07E6D6E, 0x07E8733, 0x07EA0F8, 0x07EBABD, 0x07ED481, 0x07EEE46, 0x07F080B, 0x07F21D0,
	0x07F3B95, 0x07F5559, 0x07F6F1E, 0x07F88E3, 0x07FA2A8, 0x07FBC6C, 0x07FD631, 0x07FEFF6,
	0x08009BB, 0x0802380, 0x0803D44, 0x0805709, 0x08070CE, 0x0808A93, 0x080A457, 0x080BE1C,
	0x080D7E1, 0x080F1A6, 0x0810B6B, 0x081252F, 0x0813EF4, 0x08158B9, 0x081727E, 0x0818C42,
	0x081A607, 0x081BFCC, 0x081D991, 0x081F356, 0x0820D1A, 0x08226DF, 0x08240A4, 0x0825A69,
	0x082742D, 0x0828DF2, 0x082A7B7, 0x082C17C, 0x082DB40, 0x082F505, 0x0830ECA, 0x083288F,
	0x0834254, 0x0835C18, 0x08375DD, 0x0838FA2, 0x083A967, 0x083C32B, 0x083DCF0, 0x083F6B5,
	0x084107A, 0x0842A3F, 0x0844403, 0x0845DC8, 0x084778D, 0x0849152, 0x084AB16, 0x084C4DB,
	0x084DEA0, 0x084F865, 0x085122A, 0x0852BEE, 0x08545B3, 0x0855F78, 0x085793D, 0x0859301,
	0x085ACC6, 0x085C68B, 0x085E050, 0x085FA15, 0x08613D9, 0x0862D9E, 0x0864763, 0x0866128,
	0x0867AEC, 0x08694B1, 0x086AE76, 0x086C83B, 0x086E200, 0x086FBC4, 0x0871589, 0x0872F4E,
	0x0874913, 0x08762D7, 0x0877C9C, 0x0879661, 0x087B026, 0x087C9EB, 0x087E3AF, 0x087FD74,
	0x0881739, 0x08830FE, 0x0884AC2, 0x0886487, 0x0887E4C, 0x0889811, 0x088B1D6, 0x088CB9A,
	0x088E55F, 0x088FF24, 0x08918E9, 0x08932AD, 0x0894C72, 0x0896637, 0x0897FFC, 0x08999C0,
	0x089B385, 0x089CD4A, 0x089E70F, 0x08A00D4, 0x08A1A98, 0x08A345D, 0x08A4E22, 0x08A67E7,
	0x08A81AB, 0x08A9B70, 0x08AB535, 0x08ACEFA, 0x08AE8BF, 0x08B0283, 0x08B1C48, 0x08B360D,
	0x08B4FD2, 0x08B6996, 0x08B835B, 0x08B9D20, 0x08BB6E5, 0x08BD0AA, 0x08BEA6E, 0x08C0433,
	0x08C1DF8, 0x08C37BD, 0x08C5181, 0x08C6B46, 0x08C850B, 0x08C9ED0, 0x08CB895, 0x08CD259,
	0x08CEC1E, 0x08D05E3, 0x08D1FA8, 0x08D396C, 0x08D5331, 0x08D6CF6, 0x08D86BB, 0x08DA080,
	0x08DBA44, 0x08DD409, 0x08DEDCE, 0x08E0793, 0x08E2157, 0x08E3B1C, 0x08E54E1, 0x08E6EA6,
	0x08E886B, 0x08EA22F, 0x08EBBF4, 0x08ED5B9, 0x08EEF7E, 0x08F0942, 0x08F2307, 0x08F3CCC,
	0x08F5691, 0x08F7056, 0x08F8A1A, 0x08FA3DF, 0x08FBDA4, 0x08FD769, 0x08FF12D, 0x0900AF2,
	0x09024B7, 0x0903E7C, 0x0905840, 0x0907205, 0x0908BCA, 0x090A58F, 0x090BF54, 0x090D918,
	0x090F2DD, 0x0910CA2, 0x0912667, 0x091402B, 0x09159F0, 0x09173B5, 0x0918D7A, 0x091A73F,
	0x091C103, 0x091DAC8, 0x091F48D, 0x0920E52, 0x0922816, 0x09241DB, 0x0925BA0, 0x0927565,
	0x0928F2A, 0x092A8EE, 0x092C2B3, 0x092DC78, 0x092F63D, 0x0931001, 0x09329C6, 0x093438B,
	0x0935D50, 0x0937715, 0x09390D9, 0x093AA9E, 0x093C463, 0x093DE28, 0x093F7EC, 0x09411B1,
	0x0942B76, 0x094453B, 0x0945F00, 0x09478C4, 0x0949289, 0x094AC4E, 0x094C613, 0x094DFD7,
	0x094F99C, 0x0951361, 0x0952D26, 0x09546EB, 0x09560AF, 0x0957A74, 0x0959439, 0x095ADFE,
	0x095C7C2, 0x095E187, 0x095FB4C, 0x0961511, 0x0962ED6, 0x096489A, 0x096625F, 0x0967C24,
	0x09695E9, 0x096AFAD, 0x096C972, 0x096E337, 0x096FCFC, 0x09716C0, 0x0973085, 0x0974A4A,
	0x097640F, 0x0977DD4, 0x0979798, 0x097B15D, 0x097CB22, 0x097E4E7, 0x097FEAB, 0x0981870,
	0x0983235, 0x0984BFA, 0x09865BF, 0x0987F83, 0x0989948, 0x098B30D, 0x098CCD2, 0x098E696,
	0x099005B, 0x0991A20, 0x09933E5, 0x0994DAA, 0x099676E, 0x0998133, 0x0999AF8, 0x099B4BD,
	0x099CE81, 0x099E846, 0x09A020B, 0x09A1BD0, 0x09A3595, 0x09A4F59, 0x09A691E, 0x09A82E3,
	0x09A9CA8, 0x09AB66C, 0x09AD031, 0x09AE9F6, 0x09B03BB, 0x09B1D80, 0x09B3744, 0x09B5109,
	0x09B6ACE, 0x09B8493, 0x09B9E57, 0x09BB81C, 0x09BD1E1, 0x09BEBA6, 0x09C056B, 0x09C1F2F,
	0x09C38F4, 0x09C52B9, 0x09C6C7E, 0x09C8642, 0x09CA007, 0x09CB9CC, 0x09CD391, 0x09CED56,
	0x09D071A, 0x09D20DF, 0x09D3AA4, 0x09D5469, 0x09D6E2D, 0x09D87F2, 0x09DA1B7, 0x09DBB7C,
	0x09DD540, 0x09DEF05, 0x09E08CA, 0x09E228F, 0x09E3C54, 0x09E5618, 0x09E6FDD, 0x09E89A2,
	0x09EA367, 0x09EBD2B, 0x09ED6F0, 0x09EF0B5, 0x09F0A7A, 0x09F243F, 0x09F3E03, 0x09F57C8,
	0x09F718D, 0x09F8B52, 0x09FA516, 0x09FBEDB, 0x09FD8A0, 0x09FF265, 0x0A00C2A, 0x0A025EE,
	0x0A03FB3, 0x0A05978, 0x0A0733D, 0x0A08D01, 0x0A0A6C6, 0x0A0C08B, 0x0A0DA50, 0x0A0F415,
	0x0A10DD9, 0x0A1279E, 0x0A14163, 0x0A15B28, 0x0A174EC, 0x0A18EB1, 0x0A1A876, 0x0A1C23B,
	0x0A1DC00, 0x0A1F5C4, 0x0A20F89, 0x0A2294E, 0x0A24313, 0x0A25CD7, 0x0A2769C, 0x0A29061,
	0x0A2AA26, 0x0A2C3EB, 0x0A2DDAF, 0x0A2F774, 0x0A31139, 0x0A32AFE, 0x0A344C2, 0x0A35E87,
	0x0A3784C, 0x0A39211, 0x0A3ABD6, 0x0A3C59A, 0x0A3DF5F, 0x0A3F924, 0x0A412E9, 0x0A42CAD,
	0x0A44672, 0x0A46037, 0x0A479FC, 0x0A493C0, 0x0A4AD85, 0x0A4C74A, 0x0A4E10F, 0x0A4FAD4,
	0x0A51498, 0x0A52E5D, 0x0A54822, 0x0A561E7, 0x0A57BAB, 0x0A59570, 0x0A5AF35, 0x0A5C8FA,
	0x0A5E2BF, 0x0A5FC83, 0x0A61648, 0x0A6300D, 0x0A649D2, 0x0A66396, 0x0A67D5B, 0x0A69720,
	0x0A6B0E5, 0x0A6CAAA, 0x0A6E46E, 0x0A6FE33, 0x0A717F8, 0x0A731BD, 0x0A74B81, 0x0A76546,
	0x0A77F0B, 0x0A798D0, 0x0A7B295, 0x0A7CC59, 0x0A7E61E, 0x0A7FFE3, 0x0A819A8, 0x0A8336C,
	0x0A84D31, 0x0A866F6, 0x0A880BB, 0x0A89A80, 0x0A8B444, 0x0A8CE09, 0x0A8E7CE, 0x0A90193,
	0x0A91B57, 0x0A9351C, 0x0A94EE1, 0x0A968A6, 0x0A9826B, 0x0A99C2F, 0x0A9B5F4, 0x0A9CFB9,
	0x0A9E97E, 0x0AA0342, 0x0AA1D07, 0x0AA36CC, 0x0AA5091, 0x0AA6A56, 0x0AA841A, 0x0AA9DDF,
	0x0AAB7A4, 0x0AAD169, 0x0AAEB2D, 0x0AB04F2, 0x0AB1EB7, 0x0AB387C, 0x0AB5240, 0x0AB6C05,
	0x0AB85CA, 0x0AB9F8F, 0x0ABB954, 0x0ABD318, 0x0ABECDD, 0x0AC06A2, 0x0AC2067, 0x0AC3A2B,
	0x0AC53F0, 0x0AC6DB5, 0x0AC877A, 0x0ACA13F, 0x0ACBB03, 0x0ACD4C8, 0x0ACEE8D, 0x0AD0852,
	0x0AD2216, 0x0AD3BDB, 0x0AD55A0, 0x0AD6F65, 0x0AD892A, 0x0ADA2EE, 0x0ADBCB3, 0x0ADD678,
	0x0ADF03D, 0x0AE0A01, 0x0AE23C6, 0x0AE3D8B, 0x0AE5750, 0x0AE7115, 0x0AE8AD9, 0x0AEA49E,
	0x0AEBE63, 0x0AED828, 0x0AEF1EC, 0x0AF0BB1, 0x0AF2576, 0x0AF3F3B, 0x0AF5900, 0x0AF72C4,
	0x0AF8C89, 0x0AFA64E, 0x0AFC013, 0x0AFD9D7, 0x0AFF39C, 0x0B00D61, 0x0B02726, 0x0B040EB,
	0x0B05AAF, 0x0B07474, 0x0B08E39, 0x0B0A7FE, 0x0B0C1C2, 0x0B0DB87, 0x0B0F54C, 0x0B10F11,
	0x0B128D6, 0x0B1429A, 0x0B15C5F, 0x0B17624, 0x0B18FE9, 0x0B1A9AD, 0x0B1C372, 0x0B1DD37,
	0x0B1F6FC, 0x0B210C1, 0x0B22A85, 0x0B2444A, 0x0B25E0F, 0x0B277D4, 0x0B29198, 0x0B2AB5D,
	0x0B2C522, 0x0B2DEE7, 0x0B2F8AB, 0x0B31270, 0x0B32C35, 0x0B345FA, 0x0B35FBF, 0x0B37983,
	0x0B39348, 0x0B3AD0D, 0x0B3C6D2, 0x0B3E096, 0x0B3FA5B, 0x0B41420, 0x0B42DE5, 0x0B447AA,
	0x0B4616E, 0x0B47B33, 0x0B494F8, 0x0B4AEBD, 0x0B4C881, 0x0B4E246, 0x0B4FC0B, 0x0B515D0,
	0x0B52F95, 0x0B54959, 0x0B5631E, 0x0B57CE3, 0x0B596A8, 0x0B5B06C, 0x0B5CA31, 0x0B5E3F6,
	0x0B5FDBB, 0x0B61780, 0x0B63144, 0x0B64B09, 0x0B664CE, 0x0B67E93, 0x0B69857, 0x0B6B21C,
	0x0B6CBE1, 0x0B6E5A6, 0x0B6FF6B, 0x0B7192F, 0x0B732F4, 0x0B74CB9, 0x0B7667E, 0x0B78042,
	0x0B79A07, 0x0B7B3CC, 0x0B7CD91, 0x0B7E756, 0x0B8011A, 0x0B81ADF, 0x0B834A4, 0x0B84E69,
	0x0B8682D, 0x0B881F2, 0x0B89BB7, 0x0B8B57C, 0x0B8CF41, 0x0B8E905, 0x0B902CA, 0x0B91C8F,
	0x0B93654, 0x0B95018, 0x0B969DD, 0x0B983A2, 0x0B99D67, 0x0B9B72B, 0x0B9D0F0, 0x0B9EAB5,
	0x0BA047A, 0x0BA1E3F, 0x0BA3803, 0x0BA51C8, 0x0BA6B8D, 0x0BA8552, 0x0BA9F16, 0x0BAB8DB,
	0x0BAD2A0, 0x0BAEC65, 0x0BB062A, 0x0BB1FEE, 0x0BB39B3, 0x0BB5378, 0x0BB6D3D, 0x0BB8701,
	0x0BBA0C6, 0x0BBBA8B, 0x0BBD450, 0x0BBEE15, 0x0BC07D9, 0x0BC219E, 0x0BC3B63, 0x0BC5528,
	0x0BC6EEC, 0x0BC88B1, 0x0BCA276, 0x0BCBC3B, 0x0BCD600, 0x0BCEFC4, 0x0BD0989, 0x0BD234E,
	0x0BD3D13, 0x0BD56D7, 0x0BD709C, 0x0BD8A61, 0x0BDA426, 0x0BDBDEB, 0x0BDD7AF, 0x0BDF174,
	0x0BE0B39, 0x0BE24FE, 0x0BE3EC2, 0x0BE5887, 0x0BE724C, 0x0BE8C11, 0x0BEA5D6, 0x0BEBF9A,
	0x0BED95F, 0x0BEF324, 0x0BF0CE9, 0x0BF26AD, 0x0BF4072, 0x0BF5A37, 0x0BF73FC, 0x0BF8DC1,
	0x0BFA785, 0x0BFC14A, 0x0BFDB0F, 0x0BFF4D4, 0x0C00E98, 0x0C0285D, 0x0C04222, 0x0C05BE7,
	0x0C075AB, 0x0C08F70, 0x0C0A935, 0x0C0C2FA, 0x0C0DCBF, 0x0C0F683, 0x0C11048, 0x0C12A0D,
	0x0C143D2, 0x0C15D96, 0x0C1775B, 0x0C19120, 0x0C1AAE5, 0x0C1C4AA, 0x0C1DE6E, 0x0C1F833,
	0x0C211F8, 0x0C22BBD, 0x0C24581, 0x0C25F46, 0x0C2790B, 0x0C292D0, 0x0C2AC95, 0x0C2C659,
	0x0C2E01E, 0x0C2F9E3, 0x0C313A8, 0x0C32D6C, 0x0C34731, 0x0C360F6, 0x0C37ABB, 0x0C39480,
	0x0C3AE44, 0x0C3C809, 0x0C3E1CE, 0x0C3FB93, 0x0C41557, 0x0C42F1C, 0x0C448E1, 0x0C462A6,
	0x0C47C6B, 0x0C4962F, 0x0C4AFF4, 0x0C4C9B9, 0x0C4E37E, 0x0C4FD42, 0x0C51707, 0x0C530CC,
	0x0C54A91, 0x0C56456, 0x0C57E1A, 0x0C597DF, 0x0C5B1A4, 0x0C5CB69, 0x0C5E52D, 0x0C5FEF2,
	0x0C618B7, 0x0C6327C, 0x0C64C41, 0x0C66605, 0x0C67FCA, 0x0C6998F, 0x0C6B354, 0x0C6CD18,
	0x0C6E6DD, 0x0C700A2, 0x0C71A67, 0x0C7342B, 0x0C74DF0, 0x0C767B5, 0x0C7817A, 0x0C79B3F,
	0x0C7B503, 0x0C7CEC8, 0x0C7E88D, 0x0C80252, 0x0C81C16, 0x0C835DB, 0x0C84FA0, 0x0C86965,
	0x0C8832A, 0x0C89CEE, 0x0C8B6B3, 0x0C8D078, 0x0C8EA3D, 0x0C90401, 0x0C91DC6, 0x0C9378B,
	0x0C95150, 0x0C96B15, 0x0C984D9, 0x0C99E9E, 0x0C9B863, 0x0C9D228, 0x0C9EBEC, 0x0CA05B1,
	0x0CA1F76, 0x0CA393B, 0x0CA5300, 0x0CA6CC4, 0x0CA8689, 0x0CAA04E, 0x0CABA13, 0x0CAD3D7,
	0x0CAED9C, 0x0CB0761, 0x0CB2126, 0x0CB3AEB, 0x0CB54AF, 0x0CB6E74, 0x0CB8839, 0x0CBA1FE,
	0x0CBBBC2, 0x0CBD587, 0x0CBEF4C, 0x0CC0911, 0x0CC22D6, 0x0CC3C9A, 0x0CC565F, 0x0CC7024,
	0x0CC89E9, 0x0CCA3AD, 0x0CCBD72, 0x0CCD737, 0x0CCF0FC, 0x0CD0AC1, 0x0CD2485, 0x0CD3E4A,
	0x0CD580F, 0x0CD71D4, 0x0CD8B98, 0x0CDA55D, 0x0CDBF22, 0x0CDD8E7, 0x0CDF2AB, 0x0CE0C70,
	0x0CE2635, 0x0CE3FFA, 0x0CE59BF, 0x0CE7383, 0x0CE8D48, 0x0CEA70D, 0x0CEC0D2, 0x0CEDA96,
	0x0CEF45B, 0x0CF0E20, 0x0CF27E5, 0x0CF41AA, 0x0CF5B6E, 0x0CF7533, 0x0CF8EF8, 0x0CFA8BD,
	0x0CFC281, 0x0CFDC46, 0x0CFF60B, 0x0D00FD0, 0x0D02995, 0x0D04359, 0x0D05D1E, 0x0D076E3,
	0x0D090A8, 0x0D0AA6C, 0x0D0C431, 0x0D0DDF6, 0x0D0F7BB, 0x0D11180, 0x0D12B44, 0x0D14509,
	0x0D15ECE, 0x0D17893, 0x0D19257, 0x0D1AC1C, 0x0D1C5E1, 0x0D1DFA6, 0x0D1F96B, 0x0D2132F,
	0x0D22CF4, 0x0D246B9, 0x0D2607E, 0x0D27A42, 0x0D29407, 0x0D2ADCC, 0x0D2C791, 0x0D2E156,
	0x0D2FB1A, 0x0D314DF, 0x0D32EA4, 0x0D34869, 0x0D3622D, 0x0D37BF2, 0x0D395B7, 0x0D3AF7C,
	0x0D3C941, 0x0D3E305, 0x0D3FCCA, 0x0D4168F, 0x0D43054, 0x0D44A18, 0x0D463DD, 0x0D47DA2,
	0x0D49767, 0x0D4B12B, 0x0D4CAF0, 0x0D4E4B5, 0x0D4FE7A, 0x0D5183F, 0x0D53203, 0x0D54BC8,
	0x0D5658D, 0x0D57F52, 0x0D59916, 0x0D5B2DB, 0x0D5CCA0, 0x0D5E665, 0x0D6002A, 0x0D619EE,
	0x0D633B3, 0x0D64D78, 0x0D6673D, 0x0D68101, 0x0D69AC6, 0x0D6B48B, 0x0D6CE50, 0x0D6E815,
	0x0D701D9, 0x0D71B9E, 0x0D73563, 0x0D74F28, 0x0D768EC, 0x0D782B1, 0x0D79C76, 0x0D7B63B,
	0x0D7D000, 0x0D7E9C4, 0x0D80389, 0x0D81D4E, 0x0D83713, 0x0D850D7, 0x0D86A9C, 0x0D88461,
	0x0D89E26, 0x0D8B7EB, 0x0D8D1AF, 0x0D8EB74, 0x0D90539, 0x0D91EFE, 0x0D938C2, 0x0D95287,
	0x0D96C4C, 0x0D98611, 0x0D99FD6, 0x0D9B99A, 0x0D9D35F, 0x0D9ED24, 0x0DA06E9, 0x0DA20AD,
	0x0DA3A72, 0x0DA5437, 0x0DA6DFC, 0x0DA87C1, 0x0DAA185, 0x0DABB4A, 0x0DAD50F, 0x0DAEED4,
	0x0DB0898, 0x0DB225D, 0x0DB3C22, 0x0DB55E7, 0x0DB6FAB, 0x0DB8970, 0x0DBA335, 0x0DBBCFA,
	0x0DBD6BF, 0x0DBF083, 0x0DC0A48, 0x0DC240D, 0x0DC3DD2, 0x0DC5796, 0x0DC715B, 0x0DC8B20,
	0x0DCA4E5, 0x0DCBEAA, 0x0DCD86E, 0x0DCF233, 0x0DD0BF8, 0x0DD25BD, 0x0DD3F81, 0x0DD5946,
	0x0DD730B, 0x0DD8CD0, 0x0DDA695, 0x0DDC059, 0x0DDDA1E, 0x0DDF3E3, 0x0DE0DA8, 0x0DE276C,
	0x0DE4131, 0x0DE5AF6, 0x0DE74BB, 0x0DE8E80, 0x0DEA844, 0x0DEC209, 0x0DEDBCE, 0x0DEF593,
	0x0DF0F57, 0x0DF291C, 0x0DF42E1, 0x0DF5CA6, 0x0DF766B, 0x0DF902F, 0x0DFA9F4, 0x0DFC3B9,
	0x0DFDD7E, 0x0DFF742, 0x0E01107, 0x0E02ACC, 0x0E04491, 0x0E05E56, 0x0E0781A, 0x0E091DF,
	0x0E0ABA4, 0x0E0C569, 0x0E0DF2D, 0x0E0F8F2, 0x0E112B7, 0x0E12C7C, 0x0E14641, 0x0E16005,
	0x0E179CA, 0x0E1938F, 0x0E1AD54, 0x0E1C718, 0x0E1E0DD, 0x0E1FAA2, 0x0E21467, 0x0E22E2B,
	0x0E247F0, 0x0E261B5, 0x0E27B7A, 0x0E2953F, 0x0E2AF03, 0x0E2C8C8, 0x0E2E28D, 0x0E2FC52,
	0x0E31616, 0x0E32FDB, 0x0E349A0, 0x0E36365, 0x0E37D2A, 0x0E396EE, 0x0E3B0B3, 0x0E3CA78,
	0x0E3E43D, 0x0E3FE01, 0x0E417C6, 0x0E4318B, 0x0E44B50, 0x0E46515, 0x0E47ED9, 0x0E4989E,
	0x0E4B263, 0x0E4CC28, 0x0E4E5EC, 0x0E4FFB1, 0x0E51976, 0x0E5333B, 0x0E54D00, 0x0E566C4,
	0x0E58089, 0x0E59A4E, 0x0E5B413, 0x0E5CDD7, 0x0E5E79C, 0x0E60161, 0x0E61B26, 0x0E634EB,
	0x0E64EAF, 0x0E66874, 0x0E68239, 0x0E69BFE, 0x0E6B5C2, 0x0E6CF87, 0x0E6E94C, 0x0E70311,
	0x0E71CD6, 0x0E7369A, 0x0E7505F, 0x0E76A24, 0x0E783E9, 0x0E79DAD, 0x0E7B772, 0x0E7D137,
	0x0E7EAFC, 0x0E804C1, 0x0E81E85, 0x0E8384A, 0x0E8520F, 0x0E86BD4, 0x0E88598, 0x0E89F5D,
	0x0E8B922, 0x0E8D2E7, 0x0E8ECAB, 0x0E90670, 0x0E92035, 0x0E939FA, 0x0E953BF, 0x0E96D83,
	0x0E98748, 0x0E9A10D, 0x0E9BAD2, 0x0E9D496, 0x0E9EE5B, 0x0EA0820, 0x0EA21E5, 0x0EA3BAA,
	0x0EA556E, 0x0EA6F33, 0x0EA88F8, 0x0EAA2BD, 0x0EABC81, 0x0EAD646, 0x0EAF00B, 0x0EB09D0,
	0x0EB2395, 0x0EB3D59, 0x0EB571E, 0x0EB70E3, 0x0EB8AA8, 0x0EBA46C, 0x0EBBE31, 0x0EBD7F6,
	0x0EBF1BB, 0x0EC0B80, 0x0EC2544, 0x0EC3F09, 0x0EC58CE, 0x0EC7293, 0x0EC8C57, 0x0ECA61C,
	0x0ECBFE1, 0x0ECD9A6, 0x0ECF36B, 0x0ED0D2F, 0x0ED26F4, 0x0ED40B9, 0x0ED5A7E, 0x0ED7442,
	0x0ED8E07, 0x0EDA7CC, 0x0EDC191, 0x0EDDB56, 0x0EDF51A, 0x0EE0EDF, 0x0EE28A4, 0x0EE4269,
	0x0EE5C2D, 0x0EE75F2, 0x0EE8FB7, 0x0EEA97C, 0x0EEC341, 0x0EEDD05, 0x0EEF6CA, 0x0EF108F,
	0x0EF2A54, 0x0EF4418, 0x0EF5DDD, 0x0EF77A2, 0x0EF9167, 0x0EFAB2C, 0x0EFC4F0, 0x0EFDEB5,
	0x0EFF87A, 0x0F0123F, 0x0F02C03, 0x0F045C8, 0x0F05F8D, 0x0F07952, 0x0F09316, 0x0F0ACDB,
	0x0F0C6A0, 0x0F0E065, 0x0F0FA2A, 0x0F113EE, 0x0F12DB3, 0x0F14778, 0x0F1613D, 0x0F17B01,
	0x0F194C6, 0x0F1AE8B, 0x0F1C850, 0x0F1E215, 0x0F1FBD9, 0x0F2159E, 0x0F22F63, 0x0F24928,
	0x0F262EC, 0x0F27CB1, 0x0F29676, 0x0F2B03B, 0x0F2CA00, 0x0F2E3C4, 0x0F2FD89, 0x0F3174E,
	0x0F33113, 0x0F34AD7, 0x0F3649C, 0x0F37E61, 0x0F39826, 0x0F3B1EB, 0x0F3CBAF, 0x0F3E574,
	0x0F3FF39, 0x0F418FE, 0x0F432C2, 0x0F44C87, 0x0F4664C, 0x0F48011, 0x0F499D6, 0x0F4B39A,
	0x0F4CD5F, 0x0F4E724, 0x0F500E9, 0x0F51AAD, 0x0F53472, 0x0F54E37, 0x0F567FC, 0x0F581C1,
	0x0F59B85, 0x0F5B54A, 0x0F5CF0F, 0x0F5E8D4, 0x0F60298, 0x0F61C5D, 0x0F63622, 0x0F64FE7,
	0x0F669AC, 0x0F68370, 0x0F69D35, 0x0F6B6FA, 0x0F6D0BF, 0x0F6EA83, 0x0F70448, 0x0F71E0D,
	0x0F737D2, 0x0F75196, 0x0F76B5B, 0x0F78520, 0x0F79EE5, 0x0F7B8AA, 0x0F7D26E, 0x0F7EC33,
	0x0F805F8, 0x0F81FBD, 0x0F83981, 0x0F85346, 0x0F86D0B, 0x0F886D0, 0x0F8A095, 0x0F8BA59,
	0x0F8D41E, 0x0F8EDE3, 0x0F907A8, 0x0F9216C, 0x0F93B31, 0x0F954F6, 0x0F96EBB, 0x0F98880,
	0x0F9A244, 0x0F9BC09, 0x0F9D5CE, 0x0F9EF93, 0x0FA0957, 0x0FA231C, 0x0FA3CE1, 0x0FA56A6,
	0x0FA706B, 0x0FA8A2F, 0x0FAA3F4, 0x0FABDB9, 0x0FAD77E, 0x0FAF142, 0x0FB0B07, 0x0FB24CC,
	0x0FB3E91, 0x0FB5856, 0x0FB721A, 0x0FB8BDF, 0x0FBA5A4, 0x0FBBF69, 0x0FBD92D, 0x0FBF2F2,
	0x0FC0CB7, 0x0FC267C, 0x0FC4041, 0x0FC5A05, 0x0FC73CA, 0x0FC8D8F, 0x0FCA754, 0x0FCC118,
	0x0FCDADD, 0x0FCF4A2, 0x0FD0E67, 0x0FD282C, 0x0FD41F0, 0x0FD5BB5, 0x0FD757A, 0x0FD8F3F,
	0x0FDA903, 0x0FDC2C8, 0x0FDDC8D, 0x0FDF652, 0x0FE1016, 0x0FE29DB, 0x0FE43A0, 0x0FE5D65,
	0x0FE772A, 0x0FE90EE, 0x0FEAAB3, 0x0FEC478, 0x0FEDE3D, 0x0FEF801, 0x0FF11C6, 0x0FF2B8B,
	0x0FF4550, 0x0FF5F15, 0x0FF78D9, 0x0FF929E, 0x0FFAC63, 0x0FFC628, 0x0FFDFEC, 0x0FFF9B1,
	0x1001376, 0x1002D3B, 0x1004700, 0x10060C4, 0x1007A89, 0x100944E, 0x100AE13, 0x100C7D7,
	0x100E19C, 0x100FB61, 0x1011526, 0x1012EEB, 0x10148AF, 0x1016274, 0x1017C39, 0x10195FE,
	0x101AFC2, 0x101C987, 0x101E34C, 0x101FD11, 0x10216D6, 0x102309A, 0x1024A5F, 0x1026424,
	0x1027DE9, 0x10297AD, 0x102B172, 0x102CB37, 0x102E4FC, 0x102FEC1, 0x1031885, 0x103324A,
	0x1034C0F, 0x10365D4, 0x1037F98, 0x103995D, 0x103B322, 0x103CCE7, 0x103E6AC, 0x1040070,
	0x1041A35, 0x10433FA, 0x1044DBF, 0x1046783, 0x1048148, 0x1049B0D, 0x104B4D2, 0x104CE96,
	0x104E85B, 0x1050220, 0x1051BE5, 0x10535AA, 0x1054F6E, 0x1056933, 0x10582F8, 0x1059CBD,
	0x105B681, 0x105D046, 0x105EA0B, 0x10603D0, 0x1061D95, 0x1063759, 0x106511E, 0x1066AE3,
	0x10684A8, 0x1069E6C, 0x106B831, 0x106D1F6, 0x106EBBB, 0x1070580, 0x1071F44, 0x1073909,
	0x10752CE, 0x1076C93, 0x1078657, 0x107A01C, 0x107B9E1, 0x107D3A6, 0x107ED6B, 0x108072F,
	0x10820F4, 0x1083AB9, 0x108547E, 0x1086E42, 0x1088807, 0x108A1CC, 0x108BB91, 0x108D556,
	0x108EF1A, 0x10908DF, 0x10922A4, 0x1093C69, 0x109562D, 0x1096FF2, 0x10989B7, 0x109A37C,
	0x109BD41, 0x109D705, 0x109F0CA, 0x10A0A8F, 0x10A2454, 0x10A3E18, 0x10A57DD, 0x10A71A2,
	0x10A8B67, 0x10AA52C, 0x10ABEF0, 0x10AD8B5, 0x10AF27A, 0x10B0C3F, 0x10B2603, 0x10B3FC8,
	0x10B598D, 0x10B7352, 0x10B8D16, 0x10BA6DB, 0x10BC0A0, 0x10BDA65, 0x10BF42A, 0x10C0DEE,
	0x10C27B3, 0x10C4178, 0x10C5B3D, 0x10C7501, 0x10C8EC6, 0x10CA88B, 0x10CC250, 0x10CDC15,
	0x10CF5D9, 0x10D0F9E, 0x10D2963, 0x10D4328, 0x10D5CEC, 0x10D76B1, 0x10D9076, 0x10DAA3B,
	0x10DC400, 0x10DDDC4, 0x10DF789, 0x10E114E, 0x10E2B13, 0x10E44D7, 0x10E5E9C, 0x10E7861,
	0x10E9226, 0x10EABEB, 0x10EC5AF, 0x10EDF74, 0x10EF939, 0x10F12FE, 0x10F2CC2, 0x10F4687,
	0x10F604C, 0x10F7A11, 0x10F93D6, 0x10FAD9A, 0x10FC75F, 0x10FE124, 0x10FFAE9, 0x11014AD,
	0x1102E72, 0x1104837, 0x11061FC, 0x1107BC1, 0x1109585, 0x110AF4A, 0x110C90F, 0x110E2D4,
	0x110FC98, 0x111165D, 0x1113022, 0x11149E7, 0x11163AC, 0x1117D70, 0x1119735, 0x111B0FA,
	0x111CABF, 0x111E483, 0x111FE48, 0x112180D, 0x11231D2, 0x1124B96, 0x112655B, 0x1127F20,
	0x11298E5, 0x112B2AA, 0x112CC6E, 0x112E633, 0x112FFF8, 0x11319BD, 0x1133381, 0x1134D46,
	0x113670B, 0x11380D0, 0x1139A95, 0x113B459, 0x113CE1E, 0x113E7E3, 0x11401A8, 0x1141B6C,
	0x1143531, 0x1144EF6, 0x11468BB, 0x1148280, 0x1149C44, 0x114B609, 0x114CFCE, 0x114E993,
	0x1150357, 0x1151D1C, 0x11536E1, 0x11550A6, 0x1156A6B, 0x115842F, 0x1159DF4, 0x115B7B9,
	0x115D17E, 0x115EB42, 0x1160507, 0x1161ECC, 0x1163891, 0x1165256, 0x1166C1A, 0x11685DF,
	0x1169FA4, 0x116B969, 0x116D32D, 0x116ECF2, 0x11706B7, 0x117207C, 0x1173A41, 0x1175405,
	0x1176DCA, 0x117878F, 0x117A154, 0x117BB18, 0x117D4DD, 0x117EEA2, 0x1180867, 0x118222C,
	0x1183BF0, 0x11855B5, 0x1186F7A, 0x118893F, 0x118A303, 0x118BCC8, 0x118D68D, 0x118F052,
	0x1190A16, 0x11923DB, 0x1193DA0, 0x1195765, 0x119712A, 0x1198AEE, 0x119A4B3, 0x119BE78,
	0x119D83D, 0x119F201, 0x11A0BC6, 0x11A258B, 0x11A3F50, 0x11A5915, 0x11A72D9, 0x11A8C9E,
	0x11AA663, 0x11AC028, 0x11AD9EC, 0x11AF3B1, 0x11B0D76, 0x11B273B, 0x11B4100, 0x11B5AC4,
	0x11B7489, 0x11B8E4E, 0x11BA813, 0x11BC1D7, 0x11BDB9C, 0x11BF561, 0x11C0F26, 0x11C28EB,
	0x11C42AF, 0x11C5C74, 0x11C7639, 0x11C8FFE, 0x11CA9C2, 0x11CC387, 0x11CDD4C, 0x11CF711,
	0x11D10D6, 0x11D2A9A, 0x11D445F, 0x11D5E24, 0x11D77E9, 0x11D91AD, 0x11DAB72, 0x11DC537,
	0x11DDEFC, 0x11DF8C1, 0x11E1285, 0x11E2C4A, 0x11E460F, 0x11E5FD4, 0x11E7998, 0x11E935D,
	0x11EAD22, 0x11EC6E7, 0x11EE0AC, 0x11EFA70, 0x11F1435, 0x11F2DFA, 0x11F47BF, 0x11F6183,
	0x11F7B48, 0x11F950D, 0x11FAED2, 0x11FC896, 0x11FE25B, 0x11FFC20, 0x12015E5, 0x1202FAA,
	0x120496E, 0x1206333, 0x1207CF8, 0x12096BD, 0x120B081, 0x120CA46, 0x120E40B, 0x120FDD0,
	0x1211795, 0x1213159, 0x1214B1E, 0x12164E3, 0x1217EA8, 0x121986C, 0x121B231, 0x121CBF6,
	0x121E5BB, 0x121FF80, 0x1221944, 0x1223309, 0x1224CCE, 0x1226693, 0x1228057, 0x1229A1C,
	0x122B3E1, 0x122CDA6, 0x122E76B, 0x123012F, 0x1231AF4, 0x12334B9, 0x1234E7E, 0x1236842,
	0x1238207, 0x1239BCC, 0x123B591, 0x123CF56, 0x123E91A, 0x12402DF, 0x1241CA4, 0x1243669,
	0x124502D, 0x12469F2, 0x12483B7, 0x1249D7C, 0x124B741, 0x124D105, 0x124EACA, 0x125048F,
	0x1251E54, 0x1253818, 0x12551DD, 0x1256BA2, 0x1258567, 0x1259F2C, 0x125B8F0, 0x125D2B5,
	0x125EC7A, 0x126063F, 0x1262003, 0x12639C8, 0x126538D, 0x1266D52, 0x1268717, 0x126A0DB,
	0x126BAA0, 0x126D465, 0x126EE2A, 0x12707EE, 0x12721B3, 0x1273B78, 0x127553D, 0x1276F01,
	0x12788C6, 0x127A28B, 0x127BC50, 0x127D615, 0x127EFD9, 0x128099E, 0x1282363, 0x1283D28,
	0x12856EC, 0x12870B1, 0x1288A76, 0x128A43B, 0x128BE00, 0x128D7C4, 0x128F189, 0x1290B4E,
	0x1292513, 0x1293ED7, 0x129589C, 0x1297261, 0x1298C26, 0x129A5EB, 0x129BFAF, 0x129D974,
	0x129F339, 0x12A0CFE, 0x12A26C2, 0x12A4087, 0x12A5A4C, 0x12A7411, 0x12A8DD6, 0x12AA79A,
	0x12AC15F, 0x12ADB24, 0x12AF4E9, 0x12B0EAD, 0x12B2872, 0x12B4237, 0x12B5BFC, 0x12B75C1,
	0x12B8F85, 0x12BA94A, 0x12BC30F, 0x12BDCD4, 0x12BF698, 0x12C105D, 0x12C2A22, 0x12C43E7,
	0x12C5DAC, 0x12C7770, 0x12C9135, 0x12CAAFA, 0x12CC4BF, 0x12CDE83, 0x12CF848, 0x12D120D,
	0x12D2BD2, 0x12D4597, 0x12D5F5B, 0x12D7920, 0x12D92E5, 0x12DACAA, 0x12DC66E, 0x12DE033,
	0x12DF9F8, 0x12E13BD, 0x12E2D81, 0x12E4746, 0x12E610B, 0x12E7AD0, 0x12E9495, 0x12EAE59,
	0x12EC81E, 0x12EE1E3, 0x12EFBA8, 0x12F156C, 0x12F2F31, 0x12F48F6, 0x12F62BB, 0x12F7C80,
	0x12F9644, 0x12FB009, 0x12FC9CE, 0x12FE393, 0x12FFD57, 0x130171C, 0x13030E1, 0x1304AA6,
	0x130646B, 0x1307E2F, 0x13097F4, 0x130B1B9, 0x130CB7E, 0x130E542, 0x130FF07, 0x13118CC,
	0x1313291, 0x1314C56, 0x131661A, 0x1317FDF, 0x13199A4, 0x131B369, 0x131CD2D, 0x131E6F2,
	0x13200B7, 0x1321A7C, 0x1323441, 0x1324E05, 0x13267CA, 0x132818F, 0x1329B54, 0x132B518,
	0x132CEDD, 0x132E8A2, 0x1330267, 0x1331C2C, 0x13335F0, 0x1334FB5, 0x133697A, 0x133833F,
	0x1339D03, 0x133B6C8, 0x133D08D, 0x133EA52, 0x1340417, 0x1341DDB, 0x13437A0, 0x1345165,
	0x1346B2A, 0x13484EE, 0x1349EB3, 0x134B878, 0x134D23D, 0x134EC01, 0x13505C6, 0x1351F8B,
	0x1353950, 0x1355315, 0x1356CD9, 0x135869E, 0x135A063, 0x135BA28, 0x135D3EC, 0x135EDB1,
	0x1360776, 0x136213B, 0x1363B00, 0x13654C4, 0x1366E89, 0x136884E, 0x136A213, 0x136BBD7,
	0x136D59C, 0x136EF61, 0x1370926, 0x13722EB, 0x1373CAF, 0x1375674, 0x1377039, 0x13789FE,
	0x137A3C2, 0x137BD87, 0x137D74C, 0x137F111, 0x1380AD6, 0x138249A, 0x1383E5F, 0x1385824,
	0x13871E9, 0x1388BAD, 0x138A572, 0x138BF37, 0x138D8FC, 0x138F2C1, 0x1390C85, 0x139264A,
	0x139400F, 0x13959D4, 0x1397398, 0x1398D5D, 0x139A722, 0x139C0E7, 0x139DAAC, 0x139F470,
	0x13A0E35, 0x13A27FA, 0x13A41BF, 0x13A5B83, 0x13A7548, 0x13A8F0D, 0x13AA8D2, 0x13AC297,
	0x13ADC5B, 0x13AF620, 0x13B0FE5, 0x13B29AA, 0x13B436E, 0x13B5D33, 0x13B76F8, 0x13B90BD,
	0x13BAA81, 0x13BC446, 0x13BDE0B, 0x13BF7D0, 0x13C1195, 0x13C2B59, 0x13C451E, 0x13C5EE3,
	0x13C78A8, 0x13C926C, 0x13CAC31, 0x13CC5F6, 0x13CDFBB, 0x13CF980, 0x13D1344, 0x13D2D09,
	0x13D46CE, 0x13D6093, 0x13D7A57, 0x13D941C, 0x13DADE1, 0x13DC7A6, 0x13DE16B, 0x13DFB2F,
	0x13E14F4, 0x13E2EB9, 0x13E487E, 0x13E6242, 0x13E7C07, 0x13E95CC, 0x13EAF91, 0x13EC956,
	0x13EE31A, 0x13EFCDF, 0x13F16A4, 0x13F3069, 0x13F4A2D, 0x13F63F2, 0x13F7DB7, 0x13F977C,
	0x13FB141, 0x13FCB05, 0x13FE4CA, 0x13FFE8F, 0x1401854, 0x1403218, 0x1404BDD, 0x14065A2,
	0x1407F67, 0x140992C, 0x140B2F0, 0x140CCB5, 0x140E67A, 0x141003F, 0x1411A03, 0x14133C8,
	0x1414D8D, 0x1416752, 0x1418117, 0x1419ADB, 0x141B4A0, 0x141CE65, 0x141E82A, 0x14201EE,
	0x1421BB3, 0x1423578, 0x1424F3D, 0x1426901, 0x14282C6, 0x1429C8B, 0x142B650, 0x142D015,
	0x142E9D9, 0x143039E, 0x1431D63, 0x1433728, 0x14350EC, 0x1436AB1, 0x1438476, 0x1439E3B,
	0x143B800, 0x143D1C4, 0x143EB89, 0x144054E, 0x1441F13, 0x14438D7, 0x144529C, 0x1446C61,
	0x1448626, 0x1449FEB, 0x144B9AF, 0x144D374, 0x144ED39, 0x14506FE, 0x14520C2, 0x1453A87,
	0x145544C, 0x1456E11, 0x14587D6, 0x145A19A, 0x145BB5F, 0x145D524, 0x145EEE9, 0x14608AD,
	0x1462272, 0x1463C37, 0x14655FC, 0x1466FC1, 0x1468985, 0x146A34A, 0x146BD0F, 0x146D6D4,
	0x146F098, 0x1470A5D, 0x1472422, 0x1473DE7, 0x14757AC, 0x1477170, 0x1478B35, 0x147A4FA,
	0x147BEBF, 0x147D883, 0x147F248, 0x1480C0D, 0x14825D2, 0x1483F97, 0x148595B, 0x1487320,
	0x1488CE5, 0x148A6AA, 0x148C06E, 0x148DA33, 0x148F3F8, 0x1490DBD, 0x1492781, 0x1494146,
	0x1495B0B, 0x14974D0, 0x1498E95, 0x149A859, 0x149C21E, 0x149DBE3, 0x149F5A8, 0x14A0F6C,
	0x14A2931, 0x14A42F6, 0x14A5CBB, 0x14A7680, 0x14A9044, 0x14AAA09, 0x14AC3CE, 0x14ADD93,
	0x14AF757, 0x14B111C, 0x14B2AE1, 0x14B44A6, 0x14B5E6B, 0x14B782F, 0x14B91F4, 0x14BABB9,
	0x14BC57E, 0x14BDF42, 0x14BF907, 0x14C12CC, 0x14C2C91, 0x14C4656, 0x14C601A, 0x14C79DF,
	0x14C93A4, 0x14CAD69, 0x14CC72D, 0x14CE0F2, 0x14CFAB7, 0x14D147C, 0x14D2E41, 0x14D4805,
	0x14D61CA, 0x14D7B8F, 0x14D9554, 0x14DAF18, 0x14DC8DD, 0x14DE2A2, 0x14DFC67, 0x14E162C,
	0x14E2FF0, 0x14E49B5, 0x14E637A, 0x14E7D3F, 0x14E9703, 0x14EB0C8, 0x14ECA8D, 0x14EE452,
	0x14EFE17, 0x14F17DB, 0x14F31A0, 0x14F4B65, 0x14F652A, 0x14F7EEE, 0x14F98B3, 0x14FB278,
	0x14FCC3D, 0x14FE601, 0x14FFFC6, 0x150198B, 0x1503350, 0x1504D15, 0x15066D9, 0x150809E,
	0x1509A63, 0x150B428, 0x150CDEC, 0x150E7B1, 0x1510176, 0x1511B3B, 0x1513500, 0x1514EC4,
	0x1516889, 0x151824E, 0x1519C13, 0x151B5D7, 0x151CF9C, 0x151E961, 0x1520326, 0x1521CEB,
	0x15236AF, 0x1525074, 0x1526A39, 0x15283FE, 0x1529DC2, 0x152B787, 0x152D14C, 0x152EB11,
	0x15304D6, 0x1531E9A, 0x153385F, 0x1535224, 0x1536BE9, 0x15385AD, 0x1539F72, 0x153B937,
	0x153D2FC, 0x153ECC1, 0x1540685, 0x154204A, 0x1543A0F, 0x15453D4, 0x1546D98, 0x154875D,
	0x154A122, 0x154BAE7, 0x154D4AC, 0x154EE70, 0x1550835, 0x15521FA, 0x1553BBF, 0x1555583,
	0x1556F48, 0x155890D, 0x155A2D2, 0x155BC97, 0x155D65B, 0x155F020, 0x15609E5, 0x15623AA,
	0x1563D6E, 0x1565733, 0x15670F8, 0x1568ABD, 0x156A481, 0x156BE46, 0x156D80B, 0x156F1D0,
	0x1570B95, 0x1572559, 0x1573F1E, 0x15758E3, 0x15772A8, 0x1578C6C, 0x157A631, 0x157BFF6,
	0x157D9BB, 0x157F380, 0x1580D44, 0x1582709, 0x15840CE, 0x1585A93, 0x1587457, 0x1588E1C,
	0x158A7E1, 0x158C1A6, 0x158DB6B, 0x158F52F, 0x1590EF4, 0x15928B9, 0x159427E, 0x1595C42,
	0x1597607, 0x1598FCC, 0x159A991, 0x159C356, 0x159DD1A, 0x159F6DF, 0x15A10A4, 0x15A2A69,
	0x15A442D, 0x15A5DF2, 0x15A77B7, 0x15A917C, 0x15AAB41, 0x15AC505, 0x15ADECA, 0x15AF88F,
	0x15B1254, 0x15B2C18, 0x15B45DD, 0x15B5FA2, 0x15B7967, 0x15B932C, 0x15BACF0, 0x15BC6B5,
	0x15BE07A, 0x15BFA3F, 0x15C1403, 0x15C2DC8, 0x15C478D, 0x15C6152, 0x15C7B17, 0x15C94DB,
	0x15CAEA0, 0x15CC865, 0x15CE22A, 0x15CFBEE, 0x15D15B3, 0x15D2F78, 0x15D493D, 0x15D6301,
	0x15D7CC6, 0x15D968B, 0x15DB050, 0x15DCA15, 0x15DE3D9, 0x15DFD9E, 0x15E1763, 0x15E3128,
	0x15E4AEC, 0x15E64B1, 0x15E7E76, 0x15E983B, 0x15EB200, 0x15ECBC4, 0x15EE589, 0x15EFF4E,
	0x15F1913, 0x15F32D7, 0x15F4C9C, 0x15F6661, 0x15F8026, 0x15F99EB, 0x15FB3AF, 0x15FCD74,
	0x15FE739, 0x16000FE, 0x1601AC2, 0x1603487, 0x1604E4C, 0x1606811, 0x16081D6, 0x1609B9A,
	0x160B55F, 0x160CF24, 0x160E8E9, 0x16102AD, 0x1611C72, 0x1613637, 0x1614FFC, 0x16169C1,
	0x1618385, 0x1619D4A, 0x161B70F, 0x161D0D4, 0x161EA98, 0x162045D, 0x1621E22, 0x16237E7,
	0x16251AC, 0x1626B70, 0x1628535, 0x1629EFA, 0x162B8BF, 0x162D283, 0x162EC48, 0x163060D,
	0x1631FD2, 0x1633997, 0x163535B, 0x1636D20, 0x16386E5, 0x163A0AA, 0x163BA6E, 0x163D433,
	0x163EDF8, 0x16407BD, 0x1642182, 0x1643B46, 0x164550B, 0x1646ED0, 0x1648895, 0x164A259,
	0x164BC1E, 0x164D5E3, 0x164EFA8, 0x165096C, 0x1652331, 0x1653CF6, 0x16556BB, 0x1657080,
	0x1658A44, 0x165A409, 0x165BDCE, 0x165D793, 0x165F157, 0x1660B1C, 0x16624E1, 0x1663EA6,
	0x166586B, 0x166722F, 0x1668BF4, 0x166A5B9, 0x166BF7E, 0x166D942, 0x166F307, 0x1670CCC,
	0x1672691, 0x1674056, 0x1675A1A, 0x16773DF, 0x1678DA4, 0x167A769, 0x167C12D, 0x167DAF2,
	0x167F4B7, 0x1680E7C, 0x1682841, 0x1684205, 0x1685BCA, 0x168758F, 0x1688F54, 0x168A918,
	0x168C2DD, 0x168DCA2, 0x168F667, 0x169102C, 0x16929F0, 0x16943B5, 0x1695D7A, 0x169773F,
	0x1699103, 0x169AAC8, 0x169C48D, 0x169DE52, 0x169F817, 0x16A11DB, 0x16A2BA0, 0x16A4565,
	0x16A5F2A, 0x16A78EE, 0x16A92B3, 0x16AAC78, 0x16AC63D, 0x16AE002, 0x16AF9C6, 0x16B138B,
	0x16B2D50, 0x16B4715, 0x16B60D9, 0x16B7A9E, 0x16B9463, 0x16BAE28, 0x16BC7EC, 0x16BE1B1,
	0x16BFB76, 0x16C153B, 0x16C2F00, 0x16C48C4, 0x16C6289, 0x16C7C4E, 0x16C9613, 0x16CAFD7,
	0x16CC99C, 0x16CE361, 0x16CFD26, 0x16D16EB, 0x16D30AF, 0x16D4A74, 0x16D6439, 0x16D7DFE,
	0x16D97C2, 0x16DB187, 0x16DCB4C, 0x16DE511, 0x16DFED6, 0x16E189A, 0x16E325F, 0x16E4C24,
	0x16E65E9, 0x16E7FAD, 0x16E9972, 0x16EB337, 0x16ECCFC, 0x16EE6C1, 0x16F0085, 0x16F1A4A,
	0x16F340F, 0x16F4DD4, 0x16F6798, 0x16F815D, 0x16F9B22, 0x16FB4E7, 0x16FCEAC, 0x16FE870,
	0x1700235, 0x1701BFA, 0x17035BF, 0x1704F83, 0x1706948, 0x170830D, 0x1709CD2, 0x170B697,
	0x170D05B, 0x170EA20, 0x17103E5, 0x1711DAA, 0x171376E, 0x1715133, 0x1716AF8, 0x17184BD,
	0x1719E82, 0x171B846, 0x171D20B, 0x171EBD0, 0x1720595, 0x1721F59, 0x172391E, 0x17252E3,
	0x1726CA8, 0x172866C, 0x172A031, 0x172B9F6, 0x172D3BB, 0x172ED80, 0x1730744, 0x1732109,
	0x1733ACE, 0x1735493, 0x1736E57, 0x173881C, 0x173A1E1, 0x173BBA6, 0x173D56B, 0x173EF2F,
	0x17408F4, 0x17422B9, 0x1743C7E, 0x1745642, 0x1747007, 0x17489CC, 0x174A391, 0x174BD56,
	0x174D71A, 0x174F0DF, 0x1750AA4, 0x1752469, 0x1753E2D, 0x17557F2, 0x17571B7, 0x1758B7C,
	0x175A541, 0x175BF05, 0x175D8CA, 0x175F28F, 0x1760C54, 0x1762618, 0x1763FDD, 0x17659A2,
	0x1767367, 0x1768D2C, 0x176A6F0, 0x176C0B5, 0x176DA7A, 0x176F43F, 0x1770E03, 0x17727C8,
	0x177418D, 0x1775B52, 0x1777517, 0x1778EDB, 0x177A8A0, 0x177C265, 0x177DC2A, 0x177F5EE,
	0x1780FB3, 0x1782978, 0x178433D, 0x1785D02, 0x17876C6, 0x178908B, 0x178AA50, 0x178C415,
	0x178DDD9, 0x178F79E, 0x1791163, 0x1792B28, 0x17944EC, 0x1795EB1, 0x1797876, 0x179923B,
	0x179AC00, 0x179C5C4, 0x179DF89, 0x179F94E, 0x17A1313, 0x17A2CD7, 0x17A469C, 0x17A6061,
	0x17A7A26, 0x17A93EB, 0x17AADAF, 0x17AC774, 0x17AE139, 0x17AFAFE, 0x17B14C2, 0x17B2E87,
	0x17B484C, 0x17B6211, 0x17B7BD6, 0x17B959A, 0x17BAF5F, 0x17BC924, 0x17BE2E9, 0x17BFCAD,
	0x17C1672, 0x17C3037, 0x17C49FC, 0x17C63C1, 0x17C7D85, 0x17C974A, 0x17CB10F, 0x17CCAD4,
	0x17CE498, 0x17CFE5D, 0x17D1822, 0x17D31E7, 0x17D4BAC, 0x17D6570, 0x17D7F35, 0x17D98FA,
	0x17DB2BF, 0x17DCC83, 0x17DE648, 0x17E000D, 0x17E19D2, 0x17E3397, 0x17E4D5B, 0x17E6720,
	0x17E80E5, 0x17E9AAA, 0x17EB46E, 0x17ECE33, 0x17EE7F8, 0x17F01BD, 0x17F1B82, 0x17F3546,
	0x17F4F0B, 0x17F68D0, 0x17F8295, 0x17F9C59, 0x17FB61E, 0x17FCFE3, 0x17FE9A8, 0x180036C,
	0x1801D31, 0x18036F6, 0x18050BB, 0x1806A80, 0x1808444, 0x1809E09, 0x180B7CE, 0x180D193,
	0x180EB57, 0x181051C, 0x1811EE1, 0x18138A6, 0x181526B, 0x1816C2F, 0x18185F4, 0x1819FB9,
	0x181B97E, 0x181D342, 0x181ED07, 0x18206CC, 0x1822091, 0x1823A56, 0x182541A, 0x1826DDF,
	0x18287A4, 0x182A169, 0x182BB2D, 0x182D4F2, 0x182EEB7, 0x183087C, 0x1832241, 0x1833C05,
	0x18355CA, 0x1836F8F, 0x1838954, 0x183A318, 0x183BCDD, 0x183D6A2, 0x183F067, 0x1840A2C,
	0x18423F0, 0x1843DB5, 0x184577A, 0x184713F, 0x1848B03, 0x184A4C8, 0x184BE8D, 0x184D852,
	0x184F217, 0x1850BDB, 0x18525A0, 0x1853F65, 0x185592A, 0x18572EE, 0x1858CB3, 0x185A678,
	0x185C03D, 0x185DA02, 0x185F3C6, 0x1860D8B, 0x1862750, 0x1864115, 0x1865AD9, 0x186749E,
	0x1868E63, 0x186A828, 0x186C1EC, 0x186DBB1, 0x186F576, 0x1870F3B, 0x1872900, 0x18742C4,
	0x1875C89, 0x187764E, 0x1879013, 0x187A9D7, 0x187C39C, 0x187DD61, 0x187F726, 0x18810EB,
	0x1882AAF, 0x1884474, 0x1885E39, 0x18877FE, 0x18891C2, 0x188AB87, 0x188C54C, 0x188DF11,
	0x188F8D6, 0x189129A, 0x1892C5F, 0x1894624, 0x1895FE9, 0x18979AD, 0x1899372, 0x189AD37,
	0x189C6FC, 0x189E0C1, 0x189FA85, 0x18A144A, 0x18A2E0F, 0x18A47D4, 0x18A6198, 0x18A7B5D,
	0x18A9522, 0x18AAEE7, 0x18AC8AC, 0x18AE270, 0x18AFC35, 0x18B15FA, 0x18B2FBF, 0x18B4983,
	0x18B6348, 0x18B7D0D, 0x18B96D2, 0x18BB097, 0x18BCA5B, 0x18BE420, 0x18BFDE5, 0x18C17AA,
	0x18C316E, 0x18C4B33, 0x18C64F8, 0x18C7EBD, 0x18C9882, 0x18CB246, 0x18CCC0B, 0x18CE5D0,
	0x18CFF95, 0x18D1959, 0x18D331E, 0x18D4CE3, 0x18D66A8, 0x18D806C, 0x18D9A31, 0x18DB3F6,
	0x18DCDBB, 0x18DE780, 0x18E0144, 0x18E1B09, 0x18E34CE, 0x18E4E93, 0x18E6857, 0x18E821C,
	0x18E9BE1, 0x18EB5A6, 0x18ECF6B, 0x18EE92F, 0x18F02F4, 0x18F1CB9, 0x18F367E, 0x18F5042,
	0x18F6A07, 0x18F83CC, 0x18F9D91, 0x18FB756, 0x18FD11A, 0x18FEADF, 0x19004A4, 0x1901E69,
	0x190382D, 0x19051F2, 0x1906BB7, 0x190857C, 0x1909F41, 0x190B905, 0x190D2CA, 0x190EC8F,
	0x1910654, 0x1912018, 0x19139DD, 0x19153A2, 0x1916D67, 0x191872C, 0x191A0F0, 0x191BAB5,
	0x191D47A, 0x191EE3F, 0x1920803, 0x19221C8, 0x1923B8D, 0x1925552, 0x1926F17, 0x19288DB,
	0x192A2A0, 0x192BC65, 0x192D62A, 0x192EFEE, 0x19309B3, 0x1932378, 0x1933D3D, 0x1935702,
	0x19370C6, 0x1938A8B, 0x193A450, 0x193BE15, 0x193D7D9, 0x193F19E, 0x1940B63, 0x1942528,
	0x1943EEC, 0x19458B1, 0x1947276, 0x1948C3B, 0x194A600, 0x194BFC4, 0x194D989, 0x194F34E,
	0x1950D13, 0x19526D7, 0x195409C, 0x1955A61, 0x1957426, 0x1958DEB, 0x195A7AF, 0x195C174,
	0x195DB39, 0x195F4FE, 0x1960EC2, 0x1962887, 0x196424C, 0x1965C11, 0x19675D6, 0x1968F9A,
	0x196A95F, 0x196C324, 0x196DCE9, 0x196F6AD, 0x1971072, 0x1972A37, 0x19743FC, 0x1975DC1,
	0x1977785, 0x197914A, 0x197AB0F, 0x197C4D4, 0x197DE98, 0x197F85D, 0x1981222, 0x1982BE7,
	0x19845AC, 0x1985F70, 0x1987935, 0x19892FA, 0x198ACBF, 0x198C683, 0x198E048, 0x198FA0D,
	0x19913D2, 0x1992D97, 0x199475B, 0x1996120, 0x1997AE5, 0x19994AA, 0x199AE6E, 0x199C833,
	0x199E1F8, 0x199FBBD, 0x19A1582, 0x19A2F46, 0x19A490B, 0x19A62D0, 0x19A7C95, 0x19A9659,
	0x19AB01E, 0x19AC9E3, 0x19AE3A8, 0x19AFD6D, 0x19B1731, 0x19B30F6, 0x19B4ABB, 0x19B6480,
	0x19B7E44, 0x19B9809, 0x19BB1CE, 0x19BCB93, 0x19BE557, 0x19BFF1C, 0x19C18E1, 0x19C32A6
};

unsigned int OPN_lfo_freq [8]  = {
	0x0007A299, 0x000AB584, 0x000B9D31, 0x000C4EB2, 
	0x000D4CC9, 0x0012BDA7, 0x0067131A, 0x00A4EB5E
};

signed int YMOPN_ST_dt_tab [8][32]  = {
	{
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000
	},
	{
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x000000CE, 0x000000CE, 0x000000CE, 0x000000CE,
		0x000000CE, 0x000000CE, 0x000000CE, 0x000000CE, 0x0000019C, 0x0000019C, 0x0000019C, 0x0000019C,
		0x0000019C, 0x0000026A, 0x0000026A, 0x0000026A, 0x00000338, 0x00000338, 0x00000338, 0x00000406,
		0x00000406, 0x000004D4, 0x000004D4, 0x000005A3, 0x00000671, 0x00000671, 0x00000671, 0x00000671
	},
	{
		0x000000CE, 0x000000CE, 0x000000CE, 0x000000CE, 0x0000019C, 0x0000019C, 0x0000019C, 0x0000019C,
		0x0000019C, 0x0000026A, 0x0000026A, 0x0000026A, 0x00000338, 0x00000338, 0x00000338, 0x00000406,
		0x00000406, 0x000004D4, 0x000004D4, 0x000005A3, 0x00000671, 0x00000671, 0x0000073F, 0x0000080D,
		0x000008DB, 0x000009A9, 0x00000A77, 0x00000B46, 0x00000CE2, 0x00000CE2, 0x00000CE2, 0x00000CE2
	},
	{
		0x0000019C, 0x0000019C, 0x0000019C, 0x0000019C, 0x0000019C, 0x0000026A, 0x0000026A, 0x0000026A,
		0x00000338, 0x00000338, 0x00000338, 0x00000406, 0x00000406, 0x000004D4, 0x000004D4, 0x000005A3,
		0x00000671, 0x00000671, 0x0000073F, 0x0000080D, 0x000008DB, 0x000009A9, 0x00000A77, 0x00000B46,
		0x00000CE2, 0x00000DB0, 0x00000F4C, 0x0000101A, 0x000011B7, 0x000011B7, 0x000011B7, 0x000011B7
	},
	{
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000
	},
	{
		0x00000000, 0x00000000, 0x00000000, 0x00000000, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32,
		0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64,
		0xFFFFFE64, 0xFFFFFD96, 0xFFFFFD96, 0xFFFFFD96, 0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFBFA,
		0xFFFFFBFA, 0xFFFFFB2C, 0xFFFFFB2C, 0xFFFFFA5D, 0xFFFFF98F, 0xFFFFF98F, 0xFFFFF98F, 0xFFFFF98F
	},
	{
		0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFF32, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64,
		0xFFFFFE64, 0xFFFFFD96, 0xFFFFFD96, 0xFFFFFD96, 0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFBFA,
		0xFFFFFBFA, 0xFFFFFB2C, 0xFFFFFB2C, 0xFFFFFA5D, 0xFFFFF98F, 0xFFFFF98F, 0xFFFFF8C1, 0xFFFFF7F3,
		0xFFFFF725, 0xFFFFF657, 0xFFFFF589, 0xFFFFF4BA, 0xFFFFF31E, 0xFFFFF31E, 0xFFFFF31E, 0xFFFFF31E
	},
	{
		0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFE64, 0xFFFFFD96, 0xFFFFFD96, 0xFFFFFD96,
		0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFCC8, 0xFFFFFBFA, 0xFFFFFBFA, 0xFFFFFB2C, 0xFFFFFB2C, 0xFFFFFA5D,
		0xFFFFF98F, 0xFFFFF98F, 0xFFFFF8C1, 0xFFFFF7F3, 0xFFFFF725, 0xFFFFF657, 0xFFFFF589, 0xFFFFF4BA,
		0xFFFFF31E, 0xFFFFF250, 0xFFFFF0B4, 0xFFFFEFE6, 0xFFFFEE49, 0xFFFFEE49, 0xFFFFEE49, 0xFFFFEE49
	}
};

//int Algo0 (FM_CH *CH);
//int Algo1 (FM_CH *CH);
//int Algo2 (FM_CH *CH);
//int Algo3 (FM_CH *CH);
//int Algo4 (FM_CH *CH);
//int Algo5 (FM_CH *CH);
//int Algo6 (FM_CH *CH);
//int Algo7 (FM_CH *CH);

//const algo algofuncs[8]  = { Algo0, Algo1, Algo2, Algo3, Algo4, Algo5, Algo6, Algo7 };

FM_3SLOT SL3 ;			/* 3 slot mode state */

FM_ST	ST ;				/* general state */

UINT32 LFO_AM ;		/* runtime LFO calculations helper */
INT32 LFO_PM ;			/* runtime LFO calculations helper */

FM_OPN OPN ;				/* OPN state			*/
FM_CH CH[6] ;				/* channel state		*/

int dacout ;
int dacen ;
unsigned short dac_buffer[512];
unsigned int current_sample=0;
unsigned int last_sample=0;
unsigned int timer_base;
char OPN_pan[6*2] ;	/* fm channels output masks (non-zero = enable) */

static void YM2612UpdateReq(void)
{
  int currstage;
  int length;
  
  currstage=(int)sample_count_lookup[drmd.vdp_line];
  length=currstage-laststage;
  if(length>0)
  {
	  RefreshFm();
	  fm_update_dac();
	  YM2612UpdateOne(soundbuffer+(laststage<<1),length);
	  laststage=currstage;
  }
}

INLINE void advance_eg_channel(FM_SLOT *SLOT)
{
	unsigned int out;
	unsigned int swap_flag = 0;
        int temp;

	if (!(OPN.eg_cnt & SLOT->eg_sh_active_mask))
	{
		switch(SLOT->state)
		{
		case EG_ATT:
  		// attack phase 
			SLOT->volume += (~SLOT->volume * (eg_inc[SLOT->eg_sel_ar + ((OPN.eg_cnt>>SLOT->eg_sh_ar)&7)])) >>4;
			if (SLOT->volume <= MIN_ATT_INDEX)
			{
				SLOT->volume = MIN_ATT_INDEX;
				SLOT->state = EG_DEC;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d1r_mask;
			}
			break;
		case EG_DEC:	// decay phase
			{
				int temp;
				temp = eg_inc[SLOT->eg_sel_d1r + ((OPN.eg_cnt>>SLOT->eg_sh_d1r)&7)];
				if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
					SLOT->volume += temp << 2;
				else
					SLOT->volume += temp;
			}
			if ( SLOT->volume >= SLOT->sl )
			{
				SLOT->state = EG_SUS;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d2r_mask;
			}
			break;

		case EG_SUS:	// sustain phase 
			if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
			{
				SLOT->volume += (eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)]) << 2;
				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					if (SLOT->ssg&0x01)	// bit 0 = hold 
					{
						if (!(SLOT->ssgn&1))	// have we swapped once ??? 
							swap_flag = (SLOT->ssg&0x02) | 1 ; // bit 1 = alternate 
					}
					else
					{
						SLOT->state = EG_ATT;
						SLOT->eg_sh_active_mask = SLOT->eg_sh_ar_mask;
						swap_flag = (SLOT->ssg&0x02); // bit 1 = alternate 
					}
				}
			}
			else
			{
				SLOT->volume += eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)];

				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					// do not change SLOT->state (verified on real chip) 
				}
			}
			break;

		case EG_REL:	// release phase 
			SLOT->volume += eg_inc[SLOT->eg_sel_rr + ((OPN.eg_cnt>>SLOT->eg_sh_rr)&7)];
			if ( SLOT->volume >= MAX_ATT_INDEX )
			{
				SLOT->volume = MAX_ATT_INDEX;
				SLOT->state = EG_OFF;
				SLOT->eg_sh_active_mask = 0;
			}
			break;
		}
	}

	out = SLOT->tl + ((UINT32)SLOT->volume);

	if ((SLOT->ssg&0x08) && (SLOT->ssgn&2))	// negate output (changes come from alternate bit, init comes from attack bit) 
		out ^= ((1<<ENV_BITS)-1); // 1023 

	SLOT->vol_out = out;
	SLOT->ssgn ^= swap_flag;

	SLOT++;
	if (!(OPN.eg_cnt & SLOT->eg_sh_active_mask))
	{
		switch(SLOT->state)
		{
		case EG_ATT:		// attack phase 
			SLOT->volume += (~SLOT->volume * (eg_inc[SLOT->eg_sel_ar + ((OPN.eg_cnt>>SLOT->eg_sh_ar)&7)])) >>4;
			if (SLOT->volume <= MIN_ATT_INDEX)
			{
				SLOT->volume = MIN_ATT_INDEX;
				SLOT->state = EG_DEC;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d1r_mask;
			}
			break;
		case EG_DEC:	// decay phase 
			{
				int temp;
				temp = eg_inc[SLOT->eg_sel_d1r + ((OPN.eg_cnt>>SLOT->eg_sh_d1r)&7)];
				if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
					SLOT->volume += temp << 2;
				else
					SLOT->volume += temp;
			}
			if ( SLOT->volume >= SLOT->sl )
			{
				SLOT->state = EG_SUS;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d2r_mask;
			}
			break;

		case EG_SUS:	// sustain phase 
			if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
			{
				SLOT->volume += (eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)]) << 2;
				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					if (SLOT->ssg&0x01)	// bit 0 = hold 
					{
						if (!(SLOT->ssgn&1))	// have we swapped once ??? 
							swap_flag = (SLOT->ssg&0x02) | 1 ; // bit 1 = alternate 
					}
					else
					{
						SLOT->state = EG_ATT;
						SLOT->eg_sh_active_mask = SLOT->eg_sh_ar_mask;
						swap_flag = (SLOT->ssg&0x02); // bit 1 = alternate 
					}
				}
			}
			else
			{
				SLOT->volume += eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)];

				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					// do not change SLOT->state (verified on real chip) 
				}
			}
			break;

		case EG_REL:	// release phase 
			SLOT->volume += eg_inc[SLOT->eg_sel_rr + ((OPN.eg_cnt>>SLOT->eg_sh_rr)&7)];
			if ( SLOT->volume >= MAX_ATT_INDEX )
			{
				SLOT->volume = MAX_ATT_INDEX;
				SLOT->state = EG_OFF;
				SLOT->eg_sh_active_mask = 0;
			}
			break;
		}
	}

	out = SLOT->tl + ((UINT32)SLOT->volume);

	if ((SLOT->ssg&0x08) && (SLOT->ssgn&2))	// negate output (changes come from alternate bit, init comes from attack bit) 
		out ^= ((1<<ENV_BITS)-1); // 1023 

	SLOT->vol_out = out;
	SLOT->ssgn ^= swap_flag;

	SLOT++;
	if (!(OPN.eg_cnt & SLOT->eg_sh_active_mask))
	{
		switch(SLOT->state)
		{
		case EG_ATT:		// attack phase 
			SLOT->volume += (~SLOT->volume * (eg_inc[SLOT->eg_sel_ar + ((OPN.eg_cnt>>SLOT->eg_sh_ar)&7)])) >>4;
			if (SLOT->volume <= MIN_ATT_INDEX)
			{
				SLOT->volume = MIN_ATT_INDEX;
				SLOT->state = EG_DEC;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d1r_mask;
			}
			break;
		case EG_DEC:	// decay phase 
			{
				int temp;
				temp = eg_inc[SLOT->eg_sel_d1r + ((OPN.eg_cnt>>SLOT->eg_sh_d1r)&7)];
				if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
					SLOT->volume += temp << 2;
				else
					SLOT->volume += temp;
			}
			if ( SLOT->volume >= SLOT->sl )
			{
				SLOT->state = EG_SUS;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d2r_mask;
			}
			break;

		case EG_SUS:	// sustain phase 
			if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
			{
				SLOT->volume += (eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)]) << 2;
				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					if (SLOT->ssg&0x01)	// bit 0 = hold 
					{
						if (!(SLOT->ssgn&1))	// have we swapped once ??? 
							swap_flag = (SLOT->ssg&0x02) | 1 ; // bit 1 = alternate 
					}
					else
					{
						SLOT->state = EG_ATT;
						SLOT->eg_sh_active_mask = SLOT->eg_sh_ar_mask;
						swap_flag = (SLOT->ssg&0x02); // bit 1 = alternate 
					}
				}
			}
			else
			{
				SLOT->volume += eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)];

				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					// do not change SLOT->state (verified on real chip) 
				}
			}
			break;

		case EG_REL:	// release phase 
			SLOT->volume += eg_inc[SLOT->eg_sel_rr + ((OPN.eg_cnt>>SLOT->eg_sh_rr)&7)];
			if ( SLOT->volume >= MAX_ATT_INDEX )
			{
				SLOT->volume = MAX_ATT_INDEX;
				SLOT->state = EG_OFF;
				SLOT->eg_sh_active_mask = 0;
			}
			break;
		}
	}

	out = SLOT->tl + ((UINT32)SLOT->volume);

	if ((SLOT->ssg&0x08) && (SLOT->ssgn&2))	// negate output (changes come from alternate bit, init comes from attack bit) 
		out ^= ((1<<ENV_BITS)-1); // 1023 

	SLOT->vol_out = out;
	SLOT->ssgn ^= swap_flag;

	SLOT++;
	if (!(OPN.eg_cnt & SLOT->eg_sh_active_mask))
	{
		switch(SLOT->state)
		{
		case EG_ATT:		// attack phase 
			SLOT->volume += (~SLOT->volume * (eg_inc[SLOT->eg_sel_ar + ((OPN.eg_cnt>>SLOT->eg_sh_ar)&7)])) >>4;
			if (SLOT->volume <= MIN_ATT_INDEX)
			{
				SLOT->volume = MIN_ATT_INDEX;
				SLOT->state = EG_DEC;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d1r_mask;
			}
			break;
		case EG_DEC:	// decay phase 
			{
				int temp;
				temp = eg_inc[SLOT->eg_sel_d1r + ((OPN.eg_cnt>>SLOT->eg_sh_d1r)&7)];
				if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
					SLOT->volume += temp << 2;
				else
					SLOT->volume += temp;
			}
			if ( SLOT->volume >= SLOT->sl )
			{
				SLOT->state = EG_SUS;
				SLOT->eg_sh_active_mask = SLOT->eg_sh_d2r_mask;
			}
			break;

		case EG_SUS:	// sustain phase 
			if (SLOT->ssg&0x08)	// SSG EG type envelope selected 
			{
				SLOT->volume += (eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)]) << 2;
				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					if (SLOT->ssg&0x01)	// bit 0 = hold 
					{
						if (!(SLOT->ssgn&1))	// have we swapped once ??? 
							swap_flag = (SLOT->ssg&0x02) | 1 ; // bit 1 = alternate
					}
					else
					{
						SLOT->state = EG_ATT;
						SLOT->eg_sh_active_mask = SLOT->eg_sh_ar_mask;
						swap_flag = (SLOT->ssg&0x02); // bit 1 = alternate 
					}
				}
			}
			else
			{
				SLOT->volume += eg_inc[SLOT->eg_sel_d2r + ((OPN.eg_cnt>>SLOT->eg_sh_d2r)&7)];

				if ( SLOT->volume >= MAX_ATT_INDEX )
				{
					SLOT->volume = MAX_ATT_INDEX;
					// do not change SLOT->state (verified on real chip) 
				}
			}
			break;

		case EG_REL:	// release phase 
			SLOT->volume += eg_inc[SLOT->eg_sel_rr + ((OPN.eg_cnt>>SLOT->eg_sh_rr)&7)];
			if ( SLOT->volume >= MAX_ATT_INDEX )
			{
				SLOT->volume = MAX_ATT_INDEX;
				SLOT->state = EG_OFF;
				SLOT->eg_sh_active_mask = 0;
			}
			break;
		}
	}

	out = SLOT->tl + ((UINT32)SLOT->volume);

	if ((SLOT->ssg&0x08) && (SLOT->ssgn&2))	// negate output (changes come from alternate bit, init comes from attack bit) 
		out ^= ((1<<ENV_BITS)-1); // 1023 


	SLOT->vol_out = out;
	SLOT->ssgn ^= swap_flag;
} 

INLINE int op_calc(UINT32 phase, unsigned int env, signed int pm)
{
	UINT32 p;
	UINT32 x;
	int z;

	x = (phase >> FREQ_SH) + pm;
	if (x&256)
		x^=0xff;

	p = (env<<3) + ((unsigned int)sin_tab[ x & 255 ]);

	if (p < TL_TAB_LEN)
	{
		z=(int)tl_tab[(p&510)>>1];
		z>>=(p>>9);
		if (x&512)
			return -z;
		else
			return z;
	}
	return 0;
}

INLINE int op_calc2(UINT32 phase, unsigned int env)
{
	UINT32 p;
	UINT32 x;
	int z;

	x = phase >> FREQ_SH;
	if (x&256)
		x^=0xff;

	p = (env<<3) + ((unsigned int)sin_tab[ x & 255 ]);

	if (p < TL_TAB_LEN)
	{
		z=(int)tl_tab[(p&510)>>1];
		z>>=(p>>9);
		if (x&512)
			return -z;
		else
			return z;
	}
	return 0;
}

INLINE int Algo0 (FM_CH *CH)
{
	INT32	m2,c1;		// Phase Modulation input for operators 2,3,4 
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	m2 = CH->mem_value;

	c1 = CH->op1_out[0] = CH->op1_out[1];

	CH->op1_out[1] = 0;
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	// SLOT 1 
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		CH->mem_value = op_calc(p->phase, eg, c1>>1);
	else
		CH->mem_value = 0;

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
	{
		FM_SLOT *p2 = &(CH->SLOT[SLOT3]);
		unsigned int eg2 = p2->vol_out + (ams & p2->AMmask);
		int c2;

		if( eg2 < ENV_QUIET )		/* SLOT 3 */
			c2 = op_calc(p2->phase, eg2, m2>>1);
		else
			c2=0;

		return op_calc(p->phase, eg, c2>>1);
	}

	return 0;
}
INLINE int Algo1 (FM_CH *CH)
{
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	m2;		/* Phase Modulation input for operators 2,3,4 */
	INT32	mem;			/* one sample delay memory */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	m2 = CH->mem_value;

	mem = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		mem += op_calc2(p->phase, eg>>1);

	CH->mem_value = mem;

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
	{
		FM_SLOT *p2 = &(CH->SLOT[SLOT3]);
		unsigned int eg2 = p2->vol_out + (ams & p2->AMmask);
		int c2;

		if( eg2 < ENV_QUIET )		/* SLOT 3 */
			c2 = op_calc(p2->phase, eg2, m2>>1);
		else
			c2=0;

		return op_calc(p->phase, eg, c2>>1);
	}

	return 0;
}

INLINE int Algo2 (FM_CH *CH)
{
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	m2;		/* Phase Modulation input for operators 2,3,4 */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	m2 = CH->mem_value;
	CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		CH->mem_value = op_calc2(p->phase, eg);
	else
		CH->mem_value = 0;

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
	{
		FM_SLOT *p2 = &(CH->SLOT[SLOT3]);
		unsigned int eg2 = p2->vol_out + (ams & p2->AMmask);
		int c2 = CH->op1_out[0];

		if( eg2 < ENV_QUIET )		/* SLOT 3 */
			c2 += op_calc(p2->phase, eg2, m2>>1);

		return op_calc(p->phase, eg, c2>>1);
	}

	return 0;
}

INLINE int Algo3 (FM_CH *CH)
{
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	c1,c2;		/* Phase Modulation input for operators 2,3,4 */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	c2 = CH->mem_value;

	c1 = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT3]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 3 */
		c2 += op_calc2(p->phase, eg);

	CH->mem_value = 0;
	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		CH->mem_value = op_calc(p->phase, eg, c1>>1);

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
	{
		FM_SLOT *p2 = &(CH->SLOT[SLOT3]);
		unsigned int eg2 = p2->vol_out + (ams & p2->AMmask);

		if( eg2 < ENV_QUIET )		/* SLOT 3 */
			c2 += op_calc2(p2->phase, eg2);

		return op_calc(p->phase, eg, c2>>1);
	}

	return 0;
}

INLINE int Algo4 (FM_CH *CH)
{
	int chan_output = 0;
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	c1;		/* Phase Modulation input for operators 2,3,4 */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	c1 = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		chan_output = op_calc(p->phase, eg, c1>>1);

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
	{
		FM_SLOT *p2 = &(CH->SLOT[SLOT3]);
		unsigned int eg2 = p2->vol_out + (ams & p2->AMmask);
		int c2;

		if( eg2 < ENV_QUIET )		/* SLOT 3 */
			c2 = op_calc2(p2->phase, eg2);
		else
			c2=0;

		chan_output += op_calc(p->phase, eg, c2>>1);
	}

	return chan_output;
}

INLINE int Algo5 (FM_CH *CH)
{
	int chan_output = 0;
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	m2,c1;		/* Phase Modulation input for operators 2,3,4 */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	m2 = CH->mem_value;
	CH->mem_value = c1 = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT3]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 3 */
		chan_output += op_calc(p->phase, eg, m2>>1);

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		chan_output += op_calc(p->phase, eg, c1>>1);

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
		chan_output += op_calc(p->phase, eg, c1>>1);

	return chan_output;
}

INLINE int Algo6 (FM_CH *CH)
{
	int chan_output = 0;
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	INT32	c1;		/* Phase Modulation input for operators 2,3,4 */
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	c1 = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT3]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 3 */
		chan_output += op_calc2(p->phase, eg);

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		chan_output += op_calc(p->phase, eg, c1>>1);

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
		chan_output += op_calc2(p->phase, eg);

	return chan_output;
}

INLINE int Algo7 (FM_CH *CH)
{
	int chan_output;
	INT32 out = CH->op1_out[0] + CH->op1_out[1];
	FM_SLOT *p = CH->SLOT;
	unsigned int eg;
	unsigned int ams = LFO_AM >> CH->ams;

	chan_output = CH->op1_out[0] = CH->op1_out[1];

	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )	/* SLOT 1 */
	{
		if (!CH->FB)
			out=0;
		CH->op1_out[1] = op_calc2(p->phase + (out<<CH->FB), eg );
	}
	else
		CH->op1_out[1] = 0;

	p=&(CH->SLOT[SLOT3]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 3 */
		chan_output += op_calc2(p->phase, eg);

	p=&(CH->SLOT[SLOT2]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 2 */
		chan_output += op_calc2(p->phase, eg);

	p=&(CH->SLOT[SLOT4]);
	eg = p->vol_out + (ams & p->AMmask);
	if( eg < ENV_QUIET )		/* SLOT 4 */
		chan_output += op_calc2(p->phase, eg);

	return chan_output;
}

INLINE void chan_calc2 (FM_CH *CH)
{
	// update phase counters AFTER output calculations 
	if(CH->pms)
	{
		// add support for 3 slot mode 
		UINT32 block_fnum = CH->block_fnum;

		UINT32 fnum_lfo   = ((block_fnum & 0x7f0) << 4);
		INT32  lfo_fn_table_index_offset = (int)lfo_pm_table[ fnum_lfo + CH->pms + LFO_PM ];

		if (lfo_fn_table_index_offset)	// LFO phase modulation active 
		{
			UINT8  blk;
			UINT32 fn;
			int kc,fc;

			block_fnum = (block_fnum<<1) + lfo_fn_table_index_offset;

			blk = (block_fnum&0x7000) >> 12;
			fn  = block_fnum & 0xfff;

			// keyscale code 
			kc = (blk<<2) | opn_fktable[fn >> 8];
			// phase increment counter 
			fc = OPN_fn_table[fn]>>(7-blk);

			CH->SLOT[SLOT1].phase += ((fc+CH->SLOT[SLOT1].DT[kc])*CH->SLOT[SLOT1].mul) >> 1;
			CH->SLOT[SLOT2].phase += ((fc+CH->SLOT[SLOT2].DT[kc])*CH->SLOT[SLOT2].mul) >> 1;
			CH->SLOT[SLOT3].phase += ((fc+CH->SLOT[SLOT3].DT[kc])*CH->SLOT[SLOT3].mul) >> 1;
			CH->SLOT[SLOT4].phase += ((fc+CH->SLOT[SLOT4].DT[kc])*CH->SLOT[SLOT4].mul) >> 1;

			return;
		}
	}

	CH->SLOT[SLOT1].phase += CH->SLOT[SLOT1].Incr;
	CH->SLOT[SLOT2].phase += CH->SLOT[SLOT2].Incr;
	CH->SLOT[SLOT3].phase += CH->SLOT[SLOT3].Incr;
	CH->SLOT[SLOT4].phase += CH->SLOT[SLOT4].Incr;
}

int dac_sample=0;

void YM2612UpdateOne(short *buffer, unsigned int length)
{
	int i;
	register opn_pan;
	opn_pan=0;
	if(OPN_pan[0]) opn_pan|=(1<<0);
	if(OPN_pan[1]) opn_pan|=(1<<1);
	if(OPN_pan[2]) opn_pan|=(1<<2);
	if(OPN_pan[3]) opn_pan|=(1<<3);
	if(OPN_pan[4]) opn_pan|=(1<<4);
	if(OPN_pan[5]) opn_pan|=(1<<5);
	if(OPN_pan[6]) opn_pan|=(1<<6);
	if(OPN_pan[7]) opn_pan|=(1<<7);
	if(OPN_pan[8]) opn_pan|=(1<<8);
	if(OPN_pan[9]) opn_pan|=(1<<9);
	if(OPN_pan[10]) opn_pan|=(1<<10);
	if(OPN_pan[11]) opn_pan|=(1<<11);

        // buffering 
	for(i=0; i < length ; i++)
	{
		//if((menu_options.sound_on==2)||(menu_options.sound_on==3))
		{
			if (OPN.lfo_inc)	// LFO enabled ? 
			{
				UINT8 pos;
				OPN.lfo_cnt += OPN.lfo_inc;
				pos = (OPN.lfo_cnt >> LFO_SH) & 127;
				if (pos<64)
					LFO_AM = pos << 1;
				else
					LFO_AM = 126 - ((pos&63) << 1);

				LFO_PM = pos >> 2;
			}
			else
			{
				LFO_AM = 0;
				LFO_PM = 0;
			}

		}

		// advance envelope generator 
		//if((menu_options.sound_on==2)||(menu_options.sound_on==3))
		{
			OPN.eg_timer += OPN.eg_timer_add;
			while (OPN.eg_timer >= OPN.eg_timer_overflow)
			{
				OPN.eg_timer -= OPN.eg_timer_overflow;
				OPN.eg_cnt++;

				advance_eg_channel(CH[0].SLOT);
				advance_eg_channel(CH[1].SLOT);
				advance_eg_channel(CH[2].SLOT);
				advance_eg_channel(CH[3].SLOT);
				advance_eg_channel(CH[4].SLOT);
				advance_eg_channel(CH[5].SLOT);
			}
		}
		{
			unsigned int psg_out=0;
			//if((menu_options.sound_on==1)||(menu_options.sound_on==3))
			{
				int vol;

				vol=0;
				if (PSG.sn_Output[0])
					vol = PSG.sn_Count[0];
				PSG.sn_Count[0] -= STEP;

				while (PSG.sn_Count[0] <= 0)
				{
					PSG.sn_Count[0] += PSG.sn_Period[0];
					if (PSG.sn_Count[0] > 0)
					{
						PSG.sn_Output[0] ^= 1;
						if (PSG.sn_Output[0])
							vol += PSG.sn_Period[0];
						break;
					}
					PSG.sn_Count[0] += PSG.sn_Period[0];
					vol += PSG.sn_Period[0];
				}
				if (PSG.sn_Output[0])
					vol -= PSG.sn_Count[0];

				psg_out = vol * PSG.sn_Volume[0];

				vol=0;
				if (PSG.sn_Output[1])
					vol = PSG.sn_Count[1];
				PSG.sn_Count[1] -= STEP;

				while (PSG.sn_Count[1] <= 0)
				{
					PSG.sn_Count[1] += PSG.sn_Period[1];
					if (PSG.sn_Count[1] > 0)
					{
						PSG.sn_Output[1] ^= 1;
						if (PSG.sn_Output[1])
							vol += PSG.sn_Period[1];
						break;
					}
					PSG.sn_Count[1] += PSG.sn_Period[1];
					vol += PSG.sn_Period[1];
				}
				if (PSG.sn_Output[1])
					vol -= PSG.sn_Count[1];

				psg_out += vol * PSG.sn_Volume[1];

				vol=0;
				if (PSG.sn_Output[2])
					vol = PSG.sn_Count[2];
				PSG.sn_Count[2] -= STEP;
				
				while (PSG.sn_Count[2] <= 0)
				{
					PSG.sn_Count[2] += PSG.sn_Period[2];
					if (PSG.sn_Count[2] > 0)
					{
						PSG.sn_Output[2] ^= 1;
						if (PSG.sn_Output[2])
							vol += PSG.sn_Period[2];
						break;
					}
					PSG.sn_Count[2] += PSG.sn_Period[2];
					vol += PSG.sn_Period[2];
				}
				if (PSG.sn_Output[2])
					vol -= PSG.sn_Count[2];

				psg_out += vol * PSG.sn_Volume[2];
				
				{
					int left = STEP;

					vol=0;
					do
					{
						int nextevent;

						if (PSG.sn_Count[3] < left)
							nextevent = PSG.sn_Count[3];
						else
							nextevent = left;

						if (PSG.sn_Output[3])
							vol += PSG.sn_Count[3];

						PSG.sn_Count[3] -= nextevent;

						if (PSG.sn_Count[3] <= 0)
						{
							if (PSG.sn_RNG & 1)
								PSG.sn_RNG ^= PSG.sn_NoiseFB;
							PSG.sn_RNG >>= 1;
							PSG.sn_Output[3] = PSG.sn_RNG & 1;
							PSG.sn_Count[3] += PSG.sn_Period[3];
							if (PSG.sn_Output[3])
								vol += PSG.sn_Period[3];
						}
						if (PSG.sn_Output[3])
							vol -= PSG.sn_Count[3];

						left -= nextevent;
					}
					while (left > 0);

					psg_out += vol * PSG.sn_Volume[3];
				}


			}
			
			//if((menu_options.sound_on==2)||(menu_options.sound_on==3))
			{
				int lt = (psg_out>>17);
				int rt = lt;
                                unsigned short dac_val;
				unsigned short pad3;
				// calculate FM 
				{
					FM_CH *pCH = CH;
					int temp;
					
						switch(pCH->ALGO)
						{
						   case 0:
						     temp = Algo0(pCH);
						     break;
						   case 1:
						     temp = Algo1(pCH);
						     break;
						   case 2:
						     temp = Algo2(pCH);
						     break;
						   case 3:
						     temp = Algo3(pCH);
						     break;
						   case 4:
						     temp = Algo4(pCH);
						     break;
						   case 5:
						     temp = Algo5(pCH);
						     break;
						   case 6:
						     temp = Algo6(pCH);
						     break;
						   case 7:
						     temp = Algo7(pCH);
						     break;
						}
						if (opn_pan&(1<<0))
							lt+=temp;
						if (opn_pan&(1<<1))
							rt+=temp;
					
					chan_calc2( pCH++ );

					
						switch(pCH->ALGO)
						{
						   case 0:
						     temp = Algo0(pCH);
						     break;
						   case 1:
						     temp = Algo1(pCH);
						     break;
						   case 2:
						     temp = Algo2(pCH);
						     break;
						   case 3:
						     temp = Algo3(pCH);
						     break;
						   case 4:
						     temp = Algo4(pCH);
						     break;
						   case 5:
						     temp = Algo5(pCH);
						     break;
						   case 6:
						     temp = Algo6(pCH);
						     break;
						   case 7:
						     temp = Algo7(pCH);
						     break;
						}
						if (opn_pan&(1<<2))
							lt+=temp;
						if (opn_pan&(1<<3))
							rt+=temp;
					
					chan_calc2( pCH++ );

					
						switch(pCH->ALGO)
						{
						   case 0:
						     temp = Algo0(pCH);
						     break;
						   case 1:
						     temp = Algo1(pCH);
						     break;
						   case 2:
						     temp = Algo2(pCH);
						     break;
						   case 3:
						     temp = Algo3(pCH);
						     break;
						   case 4:
						     temp = Algo4(pCH);
						     break;
						   case 5:
						     temp = Algo5(pCH);
						     break;
						   case 6:
						     temp = Algo6(pCH);
						     break;
						   case 7:
						     temp = Algo7(pCH);
						     break;
						}
						if (opn_pan&(1<<4))
							lt+=temp;
						if (opn_pan&(1<<5))
							rt+=temp;
					
					chan_calc2( pCH++ );

					
						switch(pCH->ALGO)
						{
						   case 0:
						     temp = Algo0(pCH);
						     break;
						   case 1:
						     temp = Algo1(pCH);
						     break;
						   case 2:
						     temp = Algo2(pCH);
						     break;
						   case 3:
						     temp = Algo3(pCH);
						     break;
						   case 4:
						     temp = Algo4(pCH);
						     break;
						   case 5:
						     temp = Algo5(pCH);
						     break;
						   case 6:
						     temp = Algo6(pCH);
						     break;
						   case 7:
						     temp = Algo7(pCH);
						     break;
						}
						if (opn_pan&(1<<6))
							lt+=temp;
						if (opn_pan&(1<<7))
							rt+=temp;
					
					chan_calc2( pCH++ );

					
						switch(pCH->ALGO)
						{
						   case 0:
						     temp = Algo0(pCH);
						     break;
						   case 1:
						     temp = Algo1(pCH);
						     break;
						   case 2:
						     temp = Algo2(pCH);
						     break;
						   case 3:
						     temp = Algo3(pCH);
						     break;
						   case 4:
						     temp = Algo4(pCH);
						     break;
						   case 5:
						     temp = Algo5(pCH);
						     break;
						   case 6:
						     temp = Algo6(pCH);
						     break;
						   case 7:
						     temp = Algo7(pCH);
						     break;
						}
						if (opn_pan&(1<<8))
							lt+=temp;
						if (opn_pan&(1<<9))
							rt+=temp;
					
					chan_calc2( pCH++ );
					
					dac_val=dac_buffer[dac_sample];
					if( dac_val&0xFF00 )
					{
					    if (opn_pan&(1<<10))
					        lt += ((signed char)(dac_val&0xFF))<<6;
					    if (opn_pan&(1<<11))
					        rt += ((signed char)(dac_val&0xFF))<<6;
					}
					else
					{
							switch(pCH->ALGO)
							{
							   case 0:
							     temp = Algo0(pCH);
							     break;
							   case 1:
							     temp = Algo1(pCH);
							     break;
							   case 2:
							     temp = Algo2(pCH);
							     break;
							   case 3:
							     temp = Algo3(pCH);
							     break;
							   case 4:
							     temp = Algo4(pCH);
							     break;
							   case 5:
							     temp = Algo5(pCH);
							     break;
							   case 6:
							     temp = Algo6(pCH);
							     break;
							   case 7:
							     temp = Algo7(pCH);
							     break;
							}
							if (opn_pan&(1<<10))
								lt+=temp;
							if (opn_pan&(1<<11))
								rt+=temp;
						
						chan_calc2( pCH++ );
					}

					dac_sample++;
	                                //dac_sample&=(2048-1);
					lt <<= 1;
					rt <<= 1;
					
					

					if (lt > (32767L))
						lt = 32767L;
					if (rt > (32767L))
						rt = 32767L;
					if (lt < (-32768L))
						lt = -32768L;
					if (rt < (-32768L))
						rt = -32768L;

					// buffering
					//lt=(int)(((short)rt<<16)|((short)lt));
					//if(menu_options.stereo)
					//{
#if defined(__GP2X__) || defined(__GP32__)
					   *buffer++ = (short)lt;
					   *buffer++ = (short)rt;
#endif
#if defined(__GIZ__)
						*buffer++ = (short)rt;
						*buffer++ = (short)lt;
#endif
					   
					   //*buffer++ = (short)rt;
					//}
					//else
					//{
					//   *buffer++ = (short)rt;
					//   *buffer++ = (short)lt;
					//}
				}
			}
			//else
			//{
			        //psg_out>>=16;
				//psg_out|=psg_out<<16;
			//	*buffer++ = (short)(psg_out>>16);
			//	*buffer++ = (short)(psg_out>>16);
			//}
		}

	}
}

UINT8 addr_A1 ;			/* address line A1		*/
static unsigned char pad[3];
/* OPN Mode Register Write */
INLINE void set_timers(int v )
{
	/* b7 = CSM MODE */
	/* b6 = 3 slot mode */
	/* b5 = reset b */
	/* b4 = reset a */
	/* b3 = timer enable b */
	/* b2 = timer enable a */
	/* b1 = load b */
	/* b0 = load a */
	ST.mode = v;

	/* reset Timer b flag */
	if( v & 0x20 )
	{
		/* reset status flag */
		ST.status &=~0x02;
		if (!(ST.status & ST.irqmask) )
			ST.irq = 0;
	}
	/* reset Timer a flag */
	if( v & 0x10 )
	{
		/* reset status flag */
		ST.status &=~0x01;
		if (!(ST.status & ST.irqmask) )
			ST.irq = 0;
	}


}

INLINE void FM_KEYON(FM_CH *CH , int s )
{
	FM_SLOT *SLOT = &CH->SLOT[s];
	if( !SLOT->key )
	{
		SLOT->key = 1;
		SLOT->phase = 0;		/* restart Phase Generator */
		SLOT->state = EG_ATT;	/* phase -> Attack */
		SLOT->eg_sh_active_mask = SLOT->eg_sh_ar_mask;
	}
}

INLINE void FM_KEYOFF(FM_CH *CH , int s )
{
	FM_SLOT *SLOT = &CH->SLOT[s];
	if( SLOT->key )
	{
		SLOT->key = 0;
		if (SLOT->state>EG_REL)
		{
			SLOT->state = EG_REL;/* phase -> Release */
			SLOT->eg_sh_active_mask = SLOT->eg_sh_rr_mask;
		}
	}
}

/* set detune & multiple */
INLINE void set_det_mul(FM_CH *CH,FM_SLOT *SLOT,int v)
{
	SLOT->mul = (v&0x0f)? (v&0x0f)*2 : 1;
	SLOT->DT  = (INT32 *)YMOPN_ST_dt_tab[(v>>4)&7];
	CH->SLOT[SLOT1].Incr=-1;
}

/* set total level */
INLINE void set_tl(FM_SLOT *SLOT , int v)
{
	SLOT->tl = (v&0x7f)<<(ENV_BITS-7); /* 7bit TL */
}

/* set attack rate & key scale  */
INLINE void set_ar_ksr(FM_CH *CH,FM_SLOT *SLOT,int v)
{
	UINT8 old_KSR = SLOT->KSR;

	SLOT->ar = (v&0x1f) ? 32 + ((v&0x1f)<<1) : 0;

	SLOT->KSR = 3-(v>>6);
	if (SLOT->KSR != old_KSR)
	{
		CH->SLOT[SLOT1].Incr=-1;
	}
	else
	{
		/* refresh Attack rate */
		if ((SLOT->ar + SLOT->ksr) < 32+62)
		{
			SLOT->eg_sh_ar  = eg_rate_shift [SLOT->ar  + SLOT->ksr ];
			SLOT->eg_sel_ar = eg_rate_select[SLOT->ar  + SLOT->ksr ];
		}
		else
		{
			SLOT->eg_sh_ar  = 0;
			SLOT->eg_sel_ar = 17*RATE_STEPS;
		}
	}
}

/* set decay rate */
INLINE void set_dr(FM_SLOT *SLOT,int v)
{
	SLOT->d1r = (v&0x1f) ? 32 + ((v&0x1f)<<1) : 0;

	SLOT->eg_sh_d1r = eg_rate_shift [SLOT->d1r + SLOT->ksr];
	SLOT->eg_sel_d1r= eg_rate_select[SLOT->d1r + SLOT->ksr];

}

/* set sustain rate */
INLINE void set_sr(FM_SLOT *SLOT,int v)
{
	SLOT->d2r = (v&0x1f) ? 32 + ((v&0x1f)<<1) : 0;

	SLOT->eg_sh_d2r = eg_rate_shift [SLOT->d2r + SLOT->ksr];
	SLOT->eg_sel_d2r= eg_rate_select[SLOT->d2r + SLOT->ksr];
}

/* set release rate */
INLINE void set_sl_rr(FM_SLOT *SLOT,int v)
{
	SLOT->sl = sl_table[ v>>4 ];

	SLOT->rr  = 34 + ((v&0x0f)<<2);

	SLOT->eg_sh_rr  = eg_rate_shift [SLOT->rr  + SLOT->ksr];
	SLOT->eg_sel_rr = eg_rate_select[SLOT->rr  + SLOT->ksr];
}

/* write a OPN mode register 0x20-0x2f */
static void OPNWriteMode(int r, int v)
{
	UINT8 c;
	FM_CH *pCH;

	switch(r){
	case 0x21:	/* Test */
		break;
	case 0x22:	/* LFO FREQ (YM2608/YM2610/YM2610B/YM2612) */
		if (v&0x08) /* LFO enabled ? */
			OPN.lfo_inc = OPN_lfo_freq[v&7];
		else
			OPN.lfo_inc = 0;
		break;
	case 0x24:	/* timer A High 8*/
		ST.TA = (ST.TA & 0x03)|(((int)v)<<2);
		if(ST.TA_Base != (1024 - ST.TA) << (12))
		{
		   ST.TA_Count = ST.TA_Base = (1024 - ST.TA) << (12);
		}
		break;
	case 0x25:	/* timer A Low 2*/
		ST.TA = (ST.TA & 0x3fc)|(v&3);
		if(ST.TA_Base != (1024 - ST.TA) << (12))
		{
		   ST.TA_Count = ST.TA_Base = (1024 - ST.TA) << (12);
		}
		break;
	case 0x26:	/* timer B */
		ST.TB = v;
		if(ST.TB_Base != (256 - ST.TB) << (4 + 12))
		{
		   ST.TB_Count = ST.TB_Base = (256 - ST.TB) << (4 + 12);
		}
		break;
	case 0x27:	/* mode, timer control */
			if (md_menu_options.sound_on>=3)
			{
				if ((v ^ ST.mode) & 0x40) YM2612UpdateReq();
			}
		set_timers( v );
		break;
	case 0x28:	/* key on / off */
		c = v & 0x03;
		if( c == 3 ) break;
		if(v&0x04) c+=3;
		pCH = &CH[c];
		if (md_menu_options.sound_on>=3)
		{
			YM2612UpdateReq();
		}
		if(v&0x10) FM_KEYON(pCH,SLOT1); else FM_KEYOFF(pCH,SLOT1);
		if(v&0x20) FM_KEYON(pCH,SLOT2); else FM_KEYOFF(pCH,SLOT2);
		if(v&0x40) FM_KEYON(pCH,SLOT3); else FM_KEYOFF(pCH,SLOT3);
		if(v&0x80) FM_KEYON(pCH,SLOT4); else FM_KEYOFF(pCH,SLOT4);
		break;
	}
}

/* write a OPN register (0x30-0xff) */
static void OPNWriteReg(int r, int v)
{
	FM_CH *pCH;
	FM_SLOT *SLOT;

	UINT8 c = r&3;

	if (c == 3) return; /* 0xX3,0xX7,0xXB,0xXF */

	if (r >= 0x100) c+=3;

	pCH = &CH[c];

	SLOT = &(pCH->SLOT[OPN_SLOT(r)]);

	switch( r & 0xf0 ) {
	case 0x30:	/* DET , MUL */
		set_det_mul(pCH,SLOT,v);
		break;

	case 0x40:	/* TL */
		set_tl(SLOT,v);
		break;

	case 0x50:	/* KS, AR */
		set_ar_ksr(pCH,SLOT,v);
		break;

	case 0x60:	/* bit7 = AM ENABLE, DR */
		set_dr(SLOT,v);

		SLOT->AMmask = (v&0x80) ? ~0 : 0;
		break;

	case 0x70:	/*     SR */
		set_sr(SLOT,v);
		break;

	case 0x80:	/* SL, RR */
		set_sl_rr(SLOT,v);
		break;

	case 0x90:	/* SSG-EG */

		SLOT->ssg  =  v&0x0f;
		SLOT->ssgn = (v&0x04)>>1; /* bit 1 in ssgn = attack */
		break;

	case 0xa0:
		switch( OPN_SLOT(r) ){
		case 0:		/* 0xa0-0xa2 : FNUM1 */
			{
				UINT32 fn = (((UINT32)( (ST.fn_h)&7))<<8) + v;
				UINT8 blk = ST.fn_h>>3;
				/* keyscale code */
				pCH->kcode = (blk<<2) | opn_fktable[fn >> 7];
				/* phase increment counter */
				pCH->fc = OPN_fn_table[fn*2]>>(7-blk);

				/* store fnum in clear form for LFO PM calculations */
				pCH->block_fnum = (blk<<11) | fn;

				pCH->SLOT[SLOT1].Incr=-1;
			}
			break;
		case 1:		/* 0xa4-0xa6 : FNUM2,BLK */
			ST.fn_h = v&0x3f;
			break;
		case 2:		/* 0xa8-0xaa : 3CH FNUM1 */
			if(r < 0x100)
			{
				UINT32 fn = (((UINT32)(SL3.fn_h&7))<<8) + v;
				UINT8 blk = SL3.fn_h>>3;
				/* keyscale code */
				SL3.kcode[c]= (blk<<2) | opn_fktable[fn >> 7];
				/* phase increment counter */
				SL3.fc[c] = OPN_fn_table[fn*2]>>(7-blk);
				SL3.block_fnum[c] = fn;
				pCH[2].SLOT[SLOT1].Incr=-1;
			}
			break;
		case 3:		/* 0xac-0xae : 3CH FNUM2,BLK */
			if(r < 0x100)
				SL3.fn_h = v&0x3f;
			break;
		}
		break;

	case 0xb0:
		switch( OPN_SLOT(r) ){
		case 0:		/* 0xb0-0xb2 : FB,ALGO */
			{
				int feedback = (v>>3)&7;
				if(pCH->ALGO != v&7)
				{
					if (md_menu_options.sound_on>=3)
					{
						YM2612UpdateReq();
					}
				   pCH->ALGO = v&7;
				}
				   
				pCH->FB   = feedback ? feedback+6 : 0;
				//pCH->FB   = 9 - ((v >> 3) & 7); // dave
			}
			break;
		case 1:		/* 0xb4-0xb6 : L , R , AMS , PMS (YM2612/YM2610B/YM2610/YM2608) */
			/* b0-2 PMS */
			pCH->pms = (v & 7) << 5; /* CH->pms = PM depth * 32 (index in lfo_pm_table) */

			/* b4-5 AMS */
			pCH->ams = lfo_ams_depth_shift[(v>>4) & 0x03];
			/* PAN :  b7 = L, b6 = R */
			OPN_pan[ c*2   ] = (v & 0x80);
			OPN_pan[ c*2+1 ] = (v & 0x40);
			break;
		}
		break;
	}
}

void update_tables(void)
{

        int i,d;
	double rate;
	
	// DeTune table
	for (d = 0;d <= 3;d++){
		for (i = 0;i <= 31;i++){
			rate = ((double)dt_tab[d*32 + i]) * SIN_LEN  * drmd.cpl_fm  * (1<<FREQ_SH) / ((double)(1<<20));
			YMOPN_ST_dt_tab[d][i]   = (INT32) rate;
			YMOPN_ST_dt_tab[d+4][i] = -YMOPN_ST_dt_tab[d][i];
		}
	}
	
        for(i = 0; i < 4096; i++)
	{
		// freq table for octave 7 
		// OPN phase increment counter = 20bit 
		OPN_fn_table[i] = (UINT32)( (double)i * 32.0 * drmd.cpl_fm * (double)(1<<(FREQ_SH-10)) ); // -10 because chip works with 10.10 fixed point, while we use 16.16 
	}
	// LFO freq. table 
	for(i = 0; i < 8; i++)
	{
		// Amplitude modulation: 64 output levels (triangle waveform); 1 level lasts for one of "lfo_samples_per_step" samples 
		// Phase modulation: one entry from lfo_pm_output lasts for one of 4 * "lfo_samples_per_step" samples  
		OPN_lfo_freq[i] = (unsigned int)(1.0 / (float) lfo_samples_per_step[i]) * (double)(1<<LFO_SH) * drmd.cpl_fm;
	}

	// clear DAC sample tables
	memset(&dac_buffer,0,sizeof(dac_buffer));

}


/* initialize YM2612 emulator(s) */
int YM2612Init()
{
        //update_tables - now done by menu code as it depends on a menu option
	// to select the sound rate
	YM2612ResetChip();
	return 0;
}

void YM2612ResetChip(void)
{
	int i;
	int c,s;



	/* status clear */
	ST.irqmask = 0x03;
	/* reset status flag */
	ST.status = 0;
	ST.irq = 0;

	OPNWriteMode(0x27,0x30); /* mode 0 , timer reset */

	OPN.eg_timer = 0;
	OPN.eg_cnt   = 0;

	ST.mode   = 0;	/* normal mode */
	ST.TA     = 0;
	ST.TA_Count = 0;
	ST.TA_Base = 0;
	ST.TB     = 0;
	ST.TB_Count = 0;
	ST.TB_Base = 0;


	for( c = 0 ; c < 6 ; c++ )
	{
		CH[c].fc = 0;
		for(s = 0 ; s < 4 ; s++ )
		{
			CH[c].SLOT[s].ssg = 0;
			CH[c].SLOT[s].ssgn = 0;
			CH[c].SLOT[s].state= EG_OFF;
			CH[c].SLOT[s].eg_sh_active_mask = 0;
			CH[c].SLOT[s].volume = MAX_ATT_INDEX;
			CH[c].SLOT[s].vol_out= MAX_ATT_INDEX;
		}
	}

	for(i = 0xb6 ; i >= 0xb4 ; i-- )
	{
		OPNWriteReg(i      ,0xc0);
		OPNWriteReg(i|0x100,0xc0);
	}
	for(i = 0xb2 ; i >= 0x30 ; i-- )
	{
		OPNWriteReg(i      ,0);
		OPNWriteReg(i|0x100,0);
	}
	for(i = 0x26 ; i >= 0x20 ; i-- ) OPNWriteReg(i,0);
	/* DAC mode clear */
	dacen = 0;
	dacout = 0;
}

/* YM2612 write */
/* n = number  */
/* a = address */
/* v = value   */




void fm_write(int a, UINT8 v)
{
	int addr;

	v &= 0xff;	/* adjust to 8 bit bus */

	switch( a&3){
	case 0:	/* address port 0 */
		ST.address = v;
		addr_A1 = 0;
		break;

	case 1:	/* data port 0    */
		if (addr_A1 != 0)
			break;	/* verified on real YM2608 */

		addr = ST.address;
		switch( addr & 0xf0 )
		{
		case 0x20:	/* 0x20-0x2f Mode */
			switch( addr )
			{
			case 0x2a:	/* DAC data (YM2612) */
				dacout = v-0x80;// - 0x80);// << 6;	/* level unknown */
				break;
			case 0x2b:	/* DAC Sel  (YM2612) */
				/* b7 = dac enable */
				if (dacen ^ (v & 0x80)) 
				{
					if (md_menu_options.sound_on>=3)
					{
						YM2612UpdateReq();
					}
				}
				dacen = v & 0x80;
				break;
			default:	/* OPN section */
				/* write register */
				OPNWriteMode(addr,v);
			}
			break;
		default:	/* 0x30-0xff OPN section */
			/* write register */
			OPNWriteReg(addr,v);
		}
		break;

	case 2:	/* address port 1 */
		ST.address = v;
		addr_A1 = 1;
		break;

	case 3:	/* data port 1    */
		if (addr_A1 != 1)
			break;	/* verified on real YM2608 */

		addr = ST.address;
		if (md_menu_options.sound_on>=3)
		{
			YM2612UpdateReq();
		}
		OPNWriteReg(addr | 0x100,v);
		break;
	}
}

int fm_read(int address)
{
	return ST.status;
}

/* update phase increment and envelope generator */
INLINE void refresh_fc_eg_slot(FM_SLOT *SLOT , int fc , int kc )
{
	int ksr;

	/* (frequency) phase increment counter */
	SLOT->Incr = ((fc+SLOT->DT[kc])*SLOT->mul) >> 1;

	ksr = kc >> SLOT->KSR;
	if( SLOT->ksr != ksr )
	{
		SLOT->ksr = ksr;

		/* calculate envelope generator rates */
		if ((SLOT->ar + SLOT->ksr) < 32+62)
		{
			SLOT->eg_sh_ar  = eg_rate_shift [SLOT->ar  + SLOT->ksr ];
			SLOT->eg_sel_ar = eg_rate_select[SLOT->ar  + SLOT->ksr ];
		}
		else
		{
			SLOT->eg_sh_ar  = 0;
			SLOT->eg_sel_ar = 17*RATE_STEPS;
		}

		SLOT->eg_sh_d1r = eg_rate_shift [SLOT->d1r + SLOT->ksr];
		SLOT->eg_sel_d1r= eg_rate_select[SLOT->d1r + SLOT->ksr];

		SLOT->eg_sh_d2r = eg_rate_shift [SLOT->d2r + SLOT->ksr];
		SLOT->eg_sel_d2r= eg_rate_select[SLOT->d2r + SLOT->ksr];

		SLOT->eg_sh_rr  = eg_rate_shift [SLOT->rr  + SLOT->ksr];
		SLOT->eg_sel_rr = eg_rate_select[SLOT->rr  + SLOT->ksr];
	}

	SLOT->eg_sh_ar_mask = (1<<SLOT->eg_sh_ar)-1;
	SLOT->eg_sh_d1r_mask = (1<<SLOT->eg_sh_d1r)-1;
	SLOT->eg_sh_d2r_mask = (1<<SLOT->eg_sh_d2r)-1;
	SLOT->eg_sh_rr_mask = (1<<SLOT->eg_sh_rr)-1;
}

/* update phase increment counters */
INLINE void refresh_fc_eg_chan(FM_CH *CH )
{
	if( CH->SLOT[SLOT1].Incr==-1){
		int fc = CH->fc;
		int kc = CH->kcode;
		refresh_fc_eg_slot(&CH->SLOT[SLOT1] , fc , kc );
		refresh_fc_eg_slot(&CH->SLOT[SLOT2] , fc , kc );
		refresh_fc_eg_slot(&CH->SLOT[SLOT3] , fc , kc );
		refresh_fc_eg_slot(&CH->SLOT[SLOT4] , fc , kc );
	}
}

void RefreshFm ( void )
{
	int i;

	/* If the volume is 0, increase the counter */
	for (i = 0;i < 4;i++)
	{
		if (PSG.sn_Volume[i] == 0)
		{
			/* note that I do count += length, NOT count = length + 1. You might think */
			/* it's the same since the volume is 0, but doing the latter could cause */
			/* interferencies when the program is rapidly modulating the volume. */
			if (PSG.sn_Count[i] <= (sound_buffer_size<<16))
				PSG.sn_Count[i] += (sound_buffer_size<<16);
		}
	}

	/* refresh PG and EG */
	refresh_fc_eg_chan( &CH[0] );
	refresh_fc_eg_chan( &CH[1] );
	if( (ST.mode & 0xc0) )
	{
		/* 3SLOT MODE */
		if( CH[2].SLOT[SLOT1].Incr==-1)
		{
			refresh_fc_eg_slot(&CH[2].SLOT[SLOT1] , SL3.fc[1] , SL3.kcode[1] );
			refresh_fc_eg_slot(&CH[2].SLOT[SLOT2] , SL3.fc[2] , SL3.kcode[2] );
			refresh_fc_eg_slot(&CH[2].SLOT[SLOT3] , SL3.fc[0] , SL3.kcode[0] );
			refresh_fc_eg_slot(&CH[2].SLOT[SLOT4] , CH[2].fc , CH[2].kcode );
		}
	}else refresh_fc_eg_chan( &CH[2] );
	refresh_fc_eg_chan( &CH[3] );
	refresh_fc_eg_chan( &CH[4] );
	refresh_fc_eg_chan( &CH[5] );
}

void fm_update_dac(void)
{
	int i;
    int curr_sample;
    int length;
    curr_sample= sample_count_lookup[drmd.vdp_line];
    
    length=curr_sample-last_sample; 
    
    if(length>0)
    {
	    for(i=0;i<length;i++)
	    {
		dac_buffer[current_sample++]=(unsigned short)(dacen<<8)|(dacout&0xFF);
		//current_sample&=(2048-1);
	    }
	    last_sample=curr_sample;
    }
    else
    {
		//dac_buffer[current_sample-1]=(unsigned short)(dacen<<8)|(dacout&0xFF);
    }

}

void fm_update_timers(void)
{
	/* Process YM2612 timers */
	//Timer A
	
	//i=(timer_base*length);
	if(ST.mode&1)
	{
	if((ST.TA_Count -= timer_base)<= 0)
	{
		if (md_menu_options.sound_on>=2)
		{
			YM2612UpdateReq();
		}
	   ST.TA_Count+=ST.TA_Base;
	   //ST.mode&= ~1;  // clear bit 1
	   ST.status |= (ST.mode & 0x04) >> 2;	   
	}
	}
	//Timer B
	if(ST.mode&2)
	{
	if((ST.TB_Count -= timer_base)<= 0)
	{
		if (md_menu_options.sound_on>=2)
		{
			YM2612UpdateReq();
		}
	   //ST.mode&= ~2; // clear bit 1
	   ST.TB_Count+=ST.TB_Base;
	   ST.status |= (ST.mode & 0x08) >> 2;	   
	}
	}
}

void fm_vdp_line_update(void)
{
	fm_update_dac();
	fm_update_timers();
	
	if (md_menu_options.sound_on>=4)
	{
		YM2612UpdateReq();
	}
	
}





